
package body ACM_RPoly is

  function ULP1 return Scalar is
    Eps: Scalar := ScalHalf;
    S0: Scalar := ScalOne+Eps;
    S1: Scalar;
  begin
    loop
      Eps := Half*Eps;
      S1 := ScalOne+Eps;
      exit when S1=ScalOne or else S1=S0;
      S0 := S1;
    end loop;
    return Two*Eps;
  end ULP1;

  function Floor(S: in Scalar) return Integer is
  begin
    return Integer(Flt'Floor(Approx(S)));
  end Floor;

  ---------------------------------------------
  ---------- translated from Fortran ----------
  ---------------------------------------------

  procedure RPoly(OP: in Coeffs; DEGREE: in out Integer; ZEROR,ZEROI: in out Scalars; FAIL: out Boolean) is
    --- FINDS THE ZEROS OF A REAL POLYNOMIAL
    --- OP  - DOUBLE PRECISION VECTOR OF COEFFICIENTS IN ORDER OF DECREASING POWERS
    --- DEGREE   - INTEGER DEGREE OF POLYNOMIAL
    --- ZEROR, ZEROI - OUTPUT DOUBLE PRECISION VECTORS OF REAL AND IMAGINARY PARTS OF THE ZEROS
    --- FAIL  - OUTPUT LOGICAL PARAMETER, TRUE ONLY IF LEADING COEFFICIENT IS ZERO
    ---         OR IF RPOLY HAS FOUND FEWER THAN DEGREE ZEROS
    ---         IN THE LATTER CASE DEGREE IS RESET TO THE NUMBER OF ZEROS FOUND
    LO: constant Scalar := SMALNO/ETA;
    COSR: constant Scalar := Scal(-0.069756474);
    SINR: constant Scalar := Scal(0.99756405);
    ZEROK: Boolean;
    NZ,J,L,NM1: Integer;
    XX: Scalar := Scal(0.70710678);
    YY: Scalar := -XX;
    T,AA,BB,CC,FACTOR,MAX,MIN,XXX,X,SC,BND,XM,FF,DF,DX: Scalar;
    PT,TEMP: Coeffs;
  begin
    N := DEGREE;
    NN := N+1;
    FAIL := False;
    --- ALGORITHM FAILS IF THE LEADING COEFFICIENT IS ZERO
    if IsZero(OP(1)) then
      FAIL := True;
      DEGREE := 0;
      return;
    end if;
    while IsZero(OP(NN)) loop
      J := DEGREE - N + 1;
      ZEROR(J) := ScalZero;
      ZEROI(J) := ScalZero;
      NN := NN - 1;
      N := N - 1;
    end loop;
    --- MAKE A COPY OF THE COEFFICIENTS
    for I in 1 .. NN loop P(I) := OP(I); end loop;
    --- START THE ALGORITHM FOR ONE ZERO
<<L40>>
    if N <= 2 then
      if N<1 then return; end if;
      --- CALCULATE THE FINAL ZERO OR PAIR OF ZEROS
      if N /= 2 then
        ZEROR(DEGREE) := -P(2)/P(1);
        ZEROI(DEGREE) := ScalZero;
        return;
      end if;
      QUAD(P(1),P(2),P(3),ZEROR(DEGREE-1),ZEROI(DEGREE-1),ZEROR(DEGREE),ZEROI(DEGREE));
      return;
    end if;
    --- FIND LARGEST AND SMALLEST MODULI OF COEFFICIENTS
    MAX := ScalZero;
    MIN := INFIN;
    for I in 1 .. NN loop
      X := abs(P(I));
      if X>MAX then MAX := X; end if;
      if (not IsZero(X)) and then X<MIN then MIN := X; end if;
    end loop;
    --- SCALE IF THERE ARE LARGE OR VERY SMALL COEFFICIENTS
    --- COMPUTES A SCALE FACTOR TO MULTIPLY THE COEFFICIENTS OF THE POLYNOMIAL.
    --- THE SCALING IS DONE TO AVOID OVERFLOW AND TO AVOID UNDETECTED UNDERFLOW
    --- INTERFERING WITH THE CONVERGENCE CRITERION. THE FACTOR IS A POWER OF THE BASE
    SC := LO/MIN;
    if IsZero(SC) then SC := SMALNO; end if;
    if (SC > ScalOne or else MAX >= Scal(Ten)) and then (INFIN/SC) >= MAX then
      L := Floor(Log(SC)/LogBASE + ScalHalf);
      FACTOR := BASE**L;
      if FACTOR /= ScalOne then
        for I in 1 .. NN loop P(I) := FACTOR*P(I); end loop;
      end if;
    end if;
    --- COMPUTE LOWER BOUND ON MODULI OF ZEROS
    for I in 1 .. NN loop PT(I) := abs(P(I)); end loop;
    PT(NN) := -PT(NN);
    --- COMPUTE UPPER ESTIMATE OF BOUND
    X := Exp((Log(-PT(NN))-Log(PT(1)))/Scal(N));
    if not IsZero(PT(N)) then
      --- IF NEWTON STEP AT THE ORIGIN IS BETTER, USE IT
      XM := -PT(NN)/PT(N);
      if XM<X then X := XM; end if;
    end if;
    --- CHOP THE INTERVAL (0,X) UNTIL FF .LE. 0
    loop
      XM := Flt(0.1)*X;
      FF := PT(1);
      for I in 2 .. NN loop FF := FF*XM + PT(I); end loop;
      exit when FF <= ScalZero;
      X := XM;
    end loop;
    DX := X;
    --- DO NEWTON ITERATION UNTIL X CONVERGES TO TWO DECIMAL PLACES
    while abs(DX/X) > Scal(0.005) loop
      FF := PT(1);
      DF := FF;
      for I in 2 .. N loop
        FF := FF*X + PT(I);
        DF := DF*X + FF;
      end loop;
      FF := FF*X + PT(NN);
      DX := FF/DF;
      X := X - DX;
    end loop;
    BND := X;
    --- COMPUTE THE DERIVATIVE AS THE INTIAL K POLYNOMIAL AND DO 5 STEPS WITH NO SHIFT
    NM1 := N - 1;
    for I in 2 .. N loop K(I) := Flt(NN-I)*P(I)/Scal(N); end loop;
    K(1) := P(1);
    AA := P(NN);
    BB := P(N);
    ZEROK := IsZero(K(N));
    for JJ in 1 .. 5 loop
      CC := K(N);
      if not ZEROK then
        --- USE SCALED FORM OF RECURRENCE IF VALUE OF K AT 0 IS NONZERO
        T := -AA/CC;
        for I in 1 .. NM1 loop
          J := NN - I;
          K(J) := T*K(J-1) + P(J);
        end loop;
        K(1) := P(1);
        ZEROK := abs(K(N)) <= Ten*abs(BB)*ETA;
      else
        --- USE UNSCALED FORM OF RECURRENCE
        for I in 1 .. NM1 loop
          J := NN - I;
          K(J) := K(J-1);
        end loop;
        K(1) := ScalZero;
        ZEROK := IsZero(K(N));
      end if;
    end loop;
    --- SAVE K FOR RESTARTS WITH NEW SHIFTS
    for I in 1 .. N loop TEMP(I) := K(I); end loop;
    --- LOOP TO SELECT THE QUADRATIC  CORRESPONDING TO EACH NEW SHIFT
    for CNT in 1 .. 20 loop
      --- QUADRATIC CORRESPONDS TO A DOUBLE SHIFT TO A
      --- NON-REAL POINT AND ITS COMPLEX CONJUGATE. THE POINT
      --- HAS MODULUS BND AND AMPLITUDE ROTATED BY 94 DEGREES
      --- FROM THE PREVIOUS SHIFT
      XXX := COSR*XX - SINR*YY;
      YY := SINR*XX + COSR*YY;
      XX := XXX;
      SR := BND*XX;
      SI := BND*YY;
      U := Flt(-2)*SR;
      V := BND;
      --- SECOND STAGE CALCULATION, FIXED QUADRATIC
      FXSHFR(20*CNT,NZ);
      if NZ /= 0 then
        --- THE SECOND STAGE JUMPS DIRECTLY TO
        --- ONE OF THE THIRD STAGE ITERATIONS AND RETURNS HERE IF SUCCESSFUL.
        --- DEFLATE THE POLYNOMIAL, STORE THE ZERO OR ZEROS AND RETURN TO THE MAIN ALGORITHM
        J := DEGREE - N + 1;
        ZEROR(J) := SZR;
        ZEROI(J) := SZI;
        NN := NN - NZ;
        N := NN - 1;
        for I in 1 .. NN loop P(I) := QP(I); end loop;
        if NZ /= 1 then
          ZEROR(J+1) := LZR;
          ZEROI(J+1) := LZI;
        end if;
        goto L40;
      end if;
      --- IF THE ITERATION IS UNSUCCESSFUL ANOTHER QUADRATIC IS CHOSEN AFTER RESTORING K
      for I in 1 .. N loop K(I) := TEMP(I); end loop;
    end loop;
    --- RETURN WITH FAILURE IF NO CONVERGENCE WITH 20 SHIFTS
    FAIL := True;
    DEGREE := DEGREE - N;
  end RPoly;

  procedure FXSHFR(L2: in Integer; NZ: out Integer) is
    --- COMPUTES UP TO  L2  FIXED SHIFT K-POLYNOMIALS,
    --- TESTING FOR CONVERGENCE IN THE LINEAR OR QUADRATIC CASE
    --- INITIATES ONE OF THE VARIABLE SHIFTITERATIONS
    --- AND RETURNS WITH THE NUMBER OF ZEROS FOUND.
    --- L2 - LIMIT OF FIXED SHIFT STEPS
    --- NZ - NUMBER OF ZEROS FOUND
    VPASS,SPASS,VTRY,STRY: Boolean;
    TYP,IFLAG: Integer;
    BETAS,BETAV: Scalar := Scal(Quarter);
    OSS: Scalar := SR;
    OVV: Scalar := V;
    OTS,OTV: Scalar := ScalZero; --- avoid compiler warning
    SVU,SVV,UI,VI,S,SS,VV,TS,TV,TVV,TSS: Scalar;
  begin
    NZ := 0;
    --- EVALUATE POLYNOMIAL BY SYNTHETIC DIVISION
    QUADSD(NN,U,V,P,QP,A,B);
    CALCSC(TYP);
    for J in 1 .. L2 loop
      --- CALCULATE NEXT K POLYNOMIAL AND ESTIMATE V
      NEXTK(TYP);
      CALCSC(TYP);
      NEWEST(TYP,UI,VI);
      VV := VI;
      --- ESTIMATE S
      SS := ScalZero;
      if not IsZero(K(N)) then SS := -P(NN)/K(N); end if;
      TV := ScalOne;
      TS := ScalOne;
      if J /= 1 and then TYP /= 3 then
        --- COMPUTE RELATIVE MEASURES OF CONVERGENCE OF S AND V SEQUENCES
        if not IsZero(VV) then TV := abs((VV-OVV)/VV); end if;
        if not IsZero(SS) then TS := abs((SS-OSS)/SS); end if;
        --- IF DECREASING, MULTIPLY TWO MOST RECENT CONVERGENCE MEASURES
        TVV := ScalOne;
        if TV<OTV then TVV := TV*OTV; end if;
        TSS := ScalOne;
        if TS<OTS then TSS := TS*OTS; end if;
        --- COMPARE WITH CONVERGENCE CRITERIA
        VPASS := TVV<BETAV;
        SPASS := TSS<BETAS;
        if (SPASS or else VPASS) then
          --- AT LEAST ONE SEQUENCE HAS PASSED THE CONVERGENCE TEST
          --- STORE VARIABLES BEFORE ITERATING
          SVU := U;
          SVV := V;
          for I in 1 .. N loop SVK(I) := K(I); end loop;
          S := SS;
          VTRY := False;
          STRY := False;
          if (SPASS and then ((not VPASS) or else TSS<TVV)) then goto L40; end if;
      <<L20>>
          QUADIT(UI,VI,NZ);
          if NZ>0 then return; end if;
          --- QUADRATIC ITERATION HAS FAILED.
          --- FLAG THAT IT HAS BEEN TRIED AND DECREASE THE CONVERGENCE CRITERION.
          VTRY := True;
          BETAV := Quarter*BETAV;
          --- TRY LINEAR ITERATION IF IT HAS NOT BEEN TRIED
          --- AND THE S SEQUENCE IS CONVERGING
          if STRY or else (not SPASS) then goto L50; end if;
          for I in 1 .. N loop K(I) := SVK(I); end loop;
      <<L40>>
          REALIT(S,NZ,IFLAG);
          if NZ>0 then return; end if;
          --- LINEAR ITERATION HAS FAILED.
          --- FLAG THAT IT HAS BEEN TRIED AND DECREASE THE CONVERGENCE CRITERION
          STRY := True;
          BETAS := Quarter*BETAS;
          if IFLAG /= 0 then
            --- IF LINEAR ITERATION SIGNALS AN ALMOST DOUBLE REAL ZERO ATTEMPT QUADRATIC INTERATION
            UI := -(S+S);
            VI := S*S;
            goto L20;
          end if;
          --- RESTORE VARIABLES
      <<L50>>
          U := SVU;
          V := SVV;
          for I in 1 .. N loop K(I) := SVK(I); end loop;
          --- TRY QUADRATIC ITERATION IF IT HAS NOT BEEN TRIED AND THE V SEQUENCE IS CONVERGING
          if VPASS and then (not VTRY) then goto L20; end if;
          --- RECOMPUTE QP AND SCALAR VALUES TO CONTINUE THE SECOND STAGE
          QUADSD(NN,U,V,P,QP,A,B);
          CALCSC(TYP);
        end if;
      end if;
      OVV := VV;
      OSS := SS;
      OTV := TV;
      OTS := TS;
    end loop;
  end FXSHFR;

  procedure QUADIT(UU, VV: in Scalar; NZ: out Integer) is
    --- VARIABLE-SHIFT K-POLYNOMIAL ITERATION FOR A QUADRATIC FACTOR
    --- CONVERGES ONLY IF THE ZEROS ARE EQUIMODULAR OR NEARLY SO.
    --- UU,VV - COEFFICIENTS OF STARTING QUADRATIC
    --- NZ - NUMBER OF ZERO FOUND
    TRIED: Boolean := False;
    J: Integer := 0;
    TYP: Integer;
    RELSTP,OMP: Scalar := ScalZero; --- avoid compiler warning
    UI,VI,MP,EE,T,ZM: Scalar;
  begin
    NZ := 0;
    U := UU;
    V := VV;
    --- MAIN LOOP
    loop
      QUAD(ScalOne,U,V,SZR,SZI,LZR,LZI);
      --- RETURN IF ROOTS OF THE QUADRATIC ARE REAL
      --- AND NOT CLOSE TO MULTIPLE OR NEARLY EQUAL AND  OF OPPOSITE SIGN
      if (abs(abs(SZR)-abs(LZR))>Flt(0.01)*abs(LZR)) then return; end if;
      --- EVALUATE POLYNOMIAL BY QUADRATIC SYNTHETIC DIVISION
      QUADSD(NN,U,V,P,QP,A,B);
      MP := abs(A-SZR*B) + abs(SZI*B);
      --- COMPUTE A RIGOROUS  BOUND ON THE ROUNDING ERROR IN EVALUTING P
      ZM := Sqrt(abs(V));
      EE := Two*abs(QP(1));
      T := -SZR*B;
      for I in 2 .. N loop EE := EE*ZM + abs(QP(I)); end loop;
      EE := EE*ZM + abs(A+T);
      EE := (Five*MRE+Four*ARE)*EE
        - (Five*MRE+Two*ARE)*(abs(A+T)+abs(B)*ZM) + Two*ARE*abs(T);
      --- ITERATION HAS CONVERGED SUFFICIENTLY
      --- IF THE POLYNOMIAL VALUE IS LESS THAN 20 TIMES THIS BOUND
      if MP <= (Twenty*EE) then
        NZ := 2;
        return;
      end if;
      J := J+1;
      --- STOP ITERATION AFTER 20 STEPS
      if J>20 then return; end if;
      if J >= 2 and then RELSTP <= Scal(0.01) and then MP >= OMP and then (not TRIED) then
        --- A CLUSTER APPEARS TO BE STALLING THE CONVERGENCE.
        --- FIVE FIXED SHIFT STEPS ARE TAKEN WITH A U,V CLOSE TO THE CLUSTER
        if RELSTP<ETA then RELSTP := ETA; end if;
        RELSTP := Sqrt(RELSTP);
        U := U - U*RELSTP;
        V := V + V*RELSTP;
        QUADSD(NN,U,V,P,QP,A,B);
        for I in 1 .. 5 loop
          CALCSC(TYP);
          NEXTK(TYP);
        end loop;
        TRIED := True;
        J := 0;
      end if;
      OMP := MP;
      ---  CALCULATE NEXT K POLYNOMIAL AND NEW U AND V
      CALCSC(TYP);
      NEXTK(TYP);
      CALCSC(TYP);
      NEWEST(TYP,UI,VI);
      --- IF VI IS ZERO THE ITERATION IS NOT CONVERGING
      if IsZero(VI) then return; end if;
      RELSTP := abs((VI-V)/VI);
      U := UI;
      V := VI;
    end loop;
  end QUADIT;

  procedure REALIT(SSS: in out Scalar; NZ,IFLAG: out Integer) is
    --- VARIABLE-SHIFT H POLYNOMIAL ITERATION FOR A REAL ZERO.
    --- SSS   - STARTING ITERATE
    --- NZ    - NUMBER OF ZERO FOUND
    --- IFLAG - FLAG TO INDICATE A PAIR OF ZEROS NEAR REAL AXIS.
    J: Integer := 0;
    T,OMP: Scalar := ScalZero; --- avoid compiler warning
    PV,KV,S,MS,MP,EE: Scalar;
  begin
    NZ := 0;
    S := SSS;
    IFLAG := 0;
    --- MAIN LOOP
    loop
      PV := P(1);
      --- EVALUATE P AT S
      QP(1) := PV;
      for I in 2 .. NN loop
        PV := PV*S + P(I);
        QP(I) := PV;
      end loop;
      MP := abs(PV);
      --- COMPUTE A RIGOROUS BOUND ON THE ERROR IN EVALUATING P
      MS := abs(S);
      EE := (MRE/(ARE+MRE))*abs(QP(1));
      for I in 2 .. NN loop EE := EE*MS + abs(QP(I)); end loop;
      --- ITERATION HAS CONVERGED SUFFICIENTLY
      --- IF THE POLYNOMIAL VALUE IS LESS THAN 20 TIMES THIS BOUND
      if (MP <= Twenty*((ARE+MRE)*EE-MRE*MP)) then
        NZ := 1;
        SZR := S;
        SZI := ScalZero;
        return;
      end if;
      J := J + 1;
      --- STOP ITERATION AFTER 10 STEPS
      if J>10 then return; end if;
      if J >= 2 and then abs(T) <= Flt(0.001)*abs(S-T) and then MP>OMP then
        --- A CLUSTER OF ZEROS NEAR THE REAL AXIS HAS BEEN ENCOUNTERED
        --- RETURN WITH IFLAG SET TO INITIATE A QUADRATIC ITERATION
        IFLAG := 1;
        SSS := S;
        return;
      end if;
      --- RETURN IF THE POLYNOMIAL VALUE HAS INCREASED IGNIFICANTLY
      OMP := MP;
      --- COMPUTE T, THE NEXT POLYNOMIAL, AND THE NEW ITERATE
      KV := K(1);
      QK(1) := KV;
      for I in 2 .. N loop
        KV := KV*S + K(I);
        QK(I) := KV;
      end loop;
      if abs(KV) > Ten*abs(K(N))*ETA then
        --- USE THE SCALED FORM OF THE RECURRENCE IF THE VALUE OF K AT S IS NONZERO
        T := -PV/KV;
        K(1) := QP(1);
        for I in 2 .. N loop K(I) := T*QK(I-1) + QP(I); end loop;
      else
        --- USE UNSCALED FORM
        K(1) := ScalZero;
        for I in 2 .. N loop K(I) := QK(I-1); end loop;
      end if;
      KV := K(1);
      for I in 2 .. N loop KV := KV*S + K(I); end loop;
      T := ScalZero;
      IF (abs(KV) > Ten*abs(K(N))*ETA) then T := -PV/KV; end if;
      S := S + T;
    end loop;
  end REALIT;

  procedure CALCSC(TYP: out Integer) is
    --- THIS ROUTINE CALCULATES SCALAR QUANTITIES USED TO COMPUTE
    --- THE NEXT K POLYNOMIAL AND NEW ESTIMATES OF THE QUADRATIC COEFFICIENTS.
    --- TYPE - INTEGER VARIABLE SET HERE INDICATING
    ---        HOW THE CALCULATIONS ARE NORMALIZED TO AVOID OVERFLOW
  begin
    QUADSD(N,U,V,K,QK,C,D);
    if abs(C) <= Flt(100)*abs(K(N))*ETA and then abs(D) <= Flt(100)*abs(K(N-1))*ETA then
      TYP := 3;
      --- TYPE=3 INDICATES THE QUADRATIC IS ALMOST A FACTOR OF K
      return;
    end if;
    if abs(D) >= abs(C) then
      TYP := 2;
      --- TYPE=2 INDICATES THAT ALL FORMULAS ARE DIVIDED BY D
      E := A/D;
      F := C/D;
      G := U*B;
      H := V*B;
      A3 := (A+G)*E + H*(B/D);
      A1 := B*F - A;
      A7 := (F+U)*A + H;
      return;
    end if;
    TYP := 1;
    --- TYPE=1 INDICATES THAT ALL FORMULAS ARE DIVIDED BY C
    E := A/C;
    F := D/C;
    G := U*E;
    H := V*B;
    A3 := A*E + (H/C+G)*B;
    A1 := B - A*(D/C);
    A7 := A + G*D + H*F;
  end CALCSC;

  procedure NEXTK(TYP: in Integer) is
    TEMP: Scalar;
  begin
    if TYP /= 3 then
      TEMP := A;
      if TYP=1 then TEMP := B; end if;
      if (abs(A1) <= Ten*abs(TEMP)*ETA) then
        --- IF A1 IS NEARLY ZERO THEN USE A SPECIAL FORM OF THE RECURRENCE
        K(1) := ScalZero;
        K(2) := -A7*QP(1);
        for I in 3 .. N loop K(I) := A3*QK(I-2) - A7*QP(I-1); end loop;
        return;
      end if;
      --- USE SCALED FORM OF THE RECURRENCE
      A7 := A7/A1;
      A3 := A3/A1;
      K(1) := QP(1);
      K(2) := QP(2) - A7*QP(1);
      for I in 3 .. N loop K(I) := A3*QK(I-2) - A7*QP(I-1) + QP(I); end loop;
      return;
    end if;
    --- USE UNSCALED FORM OF THE RECURRENCE IF TYPE IS 3
    K(1) := ScalZero;
    K(2) := ScalZero;
    for I in 3 .. N loop
      K(I) := QK(I-2);
    end loop;
  end NEXTK;

  procedure NEWEST(TYP: in Integer; UU,VV: in out Scalar) is
    --- COMPUTE NEW ESTIMATES OF THE QUADRATIC COEFFICIENTS
    --- USING THE SCALARS COMPUTED IN CALCSC.
    A4,A5,B1,B2,C1,C2,C3,C4,TEMP: Scalar;
  begin
    --- USE FORMULAS APPROPRIATE TO SETTING OF TYPE.
    if TYP /= 3 then
      if TYP /= 2 then
        A4 := A + U*B + H*F;
        A5 := C + (U+V*F)*D;
      else
        A4 := (A+G)*F + H;
        A5 := (F+U)*C + V*D;
      end if;
      --- EVALUATE NEW QUADRATIC COEFFICIENTS
      B1 := -K(N)/P(NN);
      B2 := -(K(N-1)+B1*P(N))/P(NN);
      C1 := V*B2*A1;
      C2 := B1*A7;
      C3 := B1*B1*A3;
      C4 := C1 - C2 - C3;
      TEMP := A5 + B1*A4 - C4;
      if not IsZero(TEMP) then
        UU := U - (U*(C3+C2)+V*(B1*A1+B2*A7))/TEMP;
        VV := V*(ScalOne+C4/TEMP);
        return;
      end if;
    end if;
    --- IF TYPE=3 THE QUADRATIC IS ZEROED
    UU := ScalZero;
    VV := ScalZero;
  end NEWEST;

  procedure QUADSD(NN0: in Integer; U0,V0: in Scalar; P0: in Coeffs; Q0: in out Coeffs; A0,B0: in out Scalar) is
    --- DIVIDES P BY THE QUADRATIC  1,U,V
    --- PLACING THE QUOTIENT IN Q AND THE REMAINDER IN A,B
    C0: Scalar;
  begin
    B0 := P0(1);
    Q0(1) := B0;
    A0 := P0(2) - U0*B0;
    Q0(2) := A0;
    for I in 3 .. NN0 loop
      C0 := P0(I) - U0*A0 - V0*B0;
      Q0(I) := C0;
      B0 := A0;
      A0 := C0;
    end loop;
  end QUADSD;

  procedure QUAD(A0,B1,C0: in Scalar; SR0,SI0,LR,LI: in out Scalar) is
    --- CALCULATE THE ZEROS OF THE QUADRATIC A*Z**2+B1*Z+C.
    --- THE QUADRATIC FORMULA,MODIFIED TO AVOID OVERFLOW,IS USED TO FIND
    --- THE LARGER ZERO IF THE ZEROS ARE REAL AND BOTH ZEROS ARE COMPLEX.
    --- THE SMALLER REAL ZERO IS FOUND DIRECTLY FROM THE ZEROS C/A.
    B0,D0,E0: Scalar;
  begin
    if not IsZero(A0) then goto L20; end if;
    SR0 := ScalZero;
    if not IsZero(B1) then SR0 := -C0/B1; end if;
    LR := ScalZero;
<<L10>>
    SI0 := ScalZero;
    LI := ScalZero;
    return;
<<L20>>
    if IsZero(C0) then
      SR0 := ScalZero;
      LR := -B1/A0;
      goto L10;
    end if;
    --- COMPUTE DISCRIMINANT AVOIDING OVERFLOW
    B0 := Half*B1;
    IF (abs(B0) >= abs(C0)) then
      E0 := ScalOne - (A0/B0)*(C0/B0);
      D0 := Sqrt(abs(E0))*abs(B0);
    else
      E0 := A0;
      if (C0 < ScalZero) then E0 := -A0; end if;
      E0 := B0*(B0/abs(C0)) - E0;
      D0 := Sqrt(abs(C0*E0));
    end if;
    if (E0 >= ScalZero) then
      --- REAL ZEROS
      if (B0 >=  ScalZero) then D0 := -D0; end if;
      LR := (D0-B0)/A0;
      SR0 := ScalZero;
      if (LR /= ScalZero) then SR0 := (C0/LR)/A0; end if;
      goto L10;
    end if;
    --- COMPLEX CONJUGATE ZEROS
    SR0 := -B0/A0;
    LR := SR0;
    SI0 := abs(D0/A0);
    LI := -SI0;
  end QUAD;

begin

  ETA := ULP1;

end ACM_RPoly;
