with Globals;
use Globals;

pragma Elaborate_All (Globals);

package body ArcTangent is

  procedure TanQuot(X,Y: in Scalar; Q,Tmp: in out Scalar) is
    --- Q = X/Sqrt(X*X+Y*Y)
  begin
    Prod(X,X,Q);
    Prod(Y,Y,Tmp);
    Add(Tmp,Q);
    Sqrt(Q,Tmp);
    Quot(X,Tmp,Q);
  end TanQuot;

  procedure ArcTan_Via_ArcSin(Num,Den: in Scalar; S: in out Scalar) is
    --- S=ArcTan(Num/Den) computed via ArcSin
    Q,Tmp: Scalar;
  begin
    TanQuot(Num,Den,Q,Tmp);             -- Q = Num/Sqrt(Num*Num+Den*Den)
    if Sign(Den)<0 then Neg(Q); end if; -- Cos is non-neg on [-Pi/2,Pi/2]
    ArcSin(Q,S);
  end ArcTan_Via_ArcSin;

  procedure ArcTan_Via_ArcCos(Num,Den: in Scalar; S: in out Scalar; Enforce_Range: Boolean) is
    --- S=ArcTan(Num/Den) computed via ArcCos
    Q,Tmp: Scalar;
  begin
    TanQuot(Den,Num,Q,Tmp);             -- Q = Den/Sqrt(Num*Num+Den*Den)
    if Sign(Num)<0 then Neg(Q); end if; -- Sin is non-neg on [0,Pi]
    ArcCos(Q,S);
    if Enforce_Range then
      Diff(S,PiHalf,Tmp);
      if Sign(Tmp)>=0 then Diff(Tmp,PiHalf,S); end if;
    end if;
  end ArcTan_Via_ArcCos;

  procedure ArcTan(Num,Den: in Scalar; S: in out Scalar; Enforce_Range: Boolean := True) is
    --- S=ArcTan(Num/Den) computed via ArcSin or ArcCos
    --- Enforce_Range is only for Den0
    Num0: constant Boolean := Contains0(Num)/=False;  -- Num could be 0
    Den0: constant Boolean := Contains0(Den)/=False;  -- Den could be 0
    NiceNum: constant Boolean := abs(Approx(Num))<Two*abs(Approx(Den));
  begin
    if Num0 then
      if Den0 then
        Show2("Num Den:",Num,Den);
        raise Undefined;
      end if;
      if IsZero(Num) then
        SetZero(S);
      else
        ArcTan_Via_ArcSin(Num,Den,S);
      end if;
    elsif Den0 then
      if IsZero(Den) then
        Copy(PiHalf,S);
      elsif Enforce_Range then
        raise Range_Error;                  -- disallow values on both sides of Pi/2
      else
        ArcTan_Via_ArcCos(Num,Den,S,False); -- allow values on both sides of Pi/2
      end if;
    elsif NiceNum then
      ArcTan_Via_ArcSin(Num,Den,S);
    else
      ArcTan_Via_ArcCos(Num,Den,S,True);
    end if;
  end ArcTan;

end Arctangent;
