with Strings;
use Strings;

package body Bisect.PSign is

  procedure FindSigns(X1,X2: in out Flt; Y1,Y2: out Flt) is
    Done: exception;
    pragma Warnings (Off);
  begin
    Y1 := FltF(X1);
    Y2 := FltF(X2);
    if Y1*Y2 <= Zero then return; end if;
    declare
      Eps: constant Flt := Half;
      Neg: constant Boolean := (Y1<Zero);
      X,Xt,Y,Yt: Flt;

      function G(S: Flt) return Flt is
      begin
        if Neg then return -FltF(S); else return FltF(S); end if;
      end G;

      procedure FindSmaller is
        --- expecting 0<G(X1)<G(X2) and G'(X1)*DX<0
        DX: Flt := Eps*(X2-X1);
      begin
        X := X1;
        loop
          Xt := X;
          X := X1+DX;
          exit when X=Xt;
          Y := G(X);
          if Y<Y1 then return; end if;
          DX := Half*DX;
        end loop;
        Show0("FindSigns: no smaller G-value found");
        raise Done;
      end FindSmaller;

      procedure FindNeg is
        --- expecting G(X)<G(X1)<G(X2)
        Fac1: constant Flt := Flt(5063)/Flt(8192);
        Fac2: constant Flt := Flt(3129)/Flt(8192);
      begin
        loop
          if abs(X-X1)<abs(X-X2) then -- X closer to X1
            Xt := Fac2*X+Fac1*X2;
            exit when Xt=X;
            Yt := G(Xt);
            if Yt<Y then
              if Yt<Zero then
                X1 := X; Y1 := Y;
                X2 := Xt; Y2 := Yt;
                raise Done;          --- creates warning
              end if;
              X1 := X; Y1 := Y;
              X := Xt; Y := Yt;
            else
              X2 := Xt; Y2 := Yt;
            end if;
          else                        -- X closer to X2
            Xt := Fac2*X+Fac1*X1;
            exit when Xt=X;
            Yt := G(Xt);
            if Yt<Y then
              if Yt<Zero then
                X1 := X; Y1 := Y;
                X2 := Xt; Y2 := Yt;
                raise Done;
              end if;
              X2 := X; Y2 := Y;
              X := Xt; Y := Yt;
            else
              X1 := Xt; Y1 := Yt;
            end if;
          end if;
        end loop;
        Show0("FindSigns: no negative G-value found");
        raise Done;
      end FindNeg;

    begin
      if Neg then Y1 := -Y1; Y2 := -Y2; end if;
      if Y2<Y1 then
        Xt := X2; X2 := X1; X1 := Xt;
        Yt := Y2; Y2 := Y1; Y1 := Yt;
      end if;
      FindSmaller; --- now G(X)<G(X1)<G(X2)
      if Y<Zero then
        X2 := X; Y2 := Y;
        raise Done;
      end if;
      FindNeg;
    exception
      when Done => if Neg then Y1 := -Y1; Y2 := -Y2; end if;
    end;
  end FindSigns;

  procedure RoughRoot(WMin,WMax: in Flt; W: in out Scalar; Err: in out Radius) is
    procedure Find_Signs is new FindSigns (FltF => FltF);
    procedure Find_Zero is new FindZero (F => F);
    X1: Flt := WMin;
    X2: Flt := WMax;
    Y1,Y2: Flt;
    W1,W2: Scalar;
  begin
    Find_Signs(X1,X2,Y1,Y2);
    Assign(X1,W1);
    Assign(X2,W2);
    if Y1*Y2<Zero then
      if Y2<Zero then Swap(W1,W2); end if;
      Find_Zero(W1,W2,Err);
      Sum(W1,W2,W);
      Mult(Half,W);
    elsif Y1=Zero then
      Copy(W1,W);
      Sub(W,W1);
      Err := Half*MaxNorm(W1);
    elsif Y2=Zero then
      Copy(W2,W);
      Sub(W,W2);
      Err := Half*MaxNorm(W2);
    else
      raise Not_Found with "no sign change found";
    end if;
  end RoughRoot;

end Bisect.PSign;
