with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls, Protected_Counters;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls, Protected_Counters;
with Ada.Numerics, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps,Protected_Counters);
pragma Elaborate_All (Ada.Numerics,Rattleback.Flows);

procedure Het is --- R symmetric heteroclinic orbit

  PDeg:         constant Positive := 767;     -- Taylor degree
  Prec:         constant Positive := 1000;    -- target precision for Integrate
  NegLogMinRad: constant Positive := 10;      -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;      -- MaxRad := MaxRadFac*MinRad
  ABFlag:       constant Flag     := 0;       -- no need to compute Alpha or Beta
  FRP:          constant FRParam  := Param13; -- rattleback parameters

  NT:           constant Positive := 8;       -- max number of concurrent tasks
  NDigits:      constant Positive := 4096;    -- number of binary digits
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  use RB;
  package RBFlow is new RB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use SMV,PMore,RBFlow;

  procedure Flow is
    R_G1: constant Flt := -348680.0*Half**20;  --- -0.332527
    R_G2: constant Flt := -144635.0*Half**20;  --- -0.137935
    ---
    TotalTime: constant Flt := 100.0;
    Stem: constant String := "P13-hetR";
    T,DT: Flt;
    G1,G2: Scalar;
    XT: Point;
    X,DotX: Curve;
  begin
    Show0("running Het.FlowR with ...");
    Assign(R_G1,G1);
    Assign(R_G2,G2);
    MakeZeroM(G1,G2,XT);
    Write(Output_Dir & Stem & ".init",XT);
    ShowApprox1("initial ",XT);
    Show_Approx_Constraints(XT);
    Plain_Flow(TotalTime,T,DT,XT,X,DotX);
    Write(Output_Dir & Stem & ".final",XT);
    ShowApprox1("final ",XT);
    Show_Approx_Constraints(XT);
    --  for I in 1..100 loop
    --     Plain_Flow(TotalTime,T,DT,XT,X);
    --     Write(Output_Dir & Stem & "." & Strng0(I,3),XT);
    --     ShowApprox1("final ",XT);
    --     Show_Approx_Constraints(XT);
    --  end loop;
  end Flow;

begin
  Verbosity := 3;
  NTasks := NT;
  Show_Rattle_Param;

  Flow;

end Het;
