with Ints, Globals, Rounding, Rounding.FPU, Strings, Bisect, Bisect.PSign;
use Ints, Globals, Rounding, Strings;

pragma Elaborate_All (Ints,Globals,Rounding.FPU,Strings,Bisect.PSign);

package body ObO is

  package Bis is new Bisect(Scalar => Scalar);
  package Bis_PSign is new Bis.PSign;

  procedure Show_ObO_Param is
    use Rounding.FPU;
    M: constant Rounding_Mode := Get_Rounding_Mode;
    Prec: constant Positive := Get_Precision(SOne/Three);
  begin
    Show0("      RepType =  " & SInfo.RepType);
    Show1("    Precision = ",Prec);         --- set precision before elaboration
    Show1("       STrunc =  ",STrunc);
    Show0("Rounding_Mode =  " & Strng(M));
    Show1("       NTasks = ",NTasks);
    Show1( "   Verbosity = ",Verbosity);
    Show1("         PDeg = ",PDeg);
  end Show_ObO_Param;

  ---------- all degrees

  function EffLast(P: Taylor) return Power is
  begin
    for I in reverse 1 .. PDeg loop
      if not IsZero(P(I)) then return I; end if;
    end loop;
    return 0;
  end EffLast;

  function IsZero(P: Taylor) return Boolean is
  begin
    for I in Power loop
      if not IsZero(P(I)) then return False; end if;
    end loop;
    return True;
  end IsZero;

  function IsPoly(P: Taylor) return Boolean is
  begin
    return STrunc or else IsSharp(P(PDeg));
  end IsPoly;

  procedure SetZero(P: in out Taylor) is
  begin
    for I in Power loop SetZero(P(I)); end loop;
  end SetZero;

  procedure Assign(S: in Scalar; P: in out Taylor) is
  begin
    Copy(S,P(0));
    for I in 1 .. PDeg loop SetZero(P(I)); end loop;
  end Assign;

  procedure Copy(P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop Copy(P(I),Q(I)); end loop;
  end Copy;

  procedure Neg(P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop Neg(P(I),Q(I)); end loop;
  end Neg;

  procedure Add(P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop Add(P(I),Q(I)); end loop;
  end Add;

  procedure Sum(P,Q: in Taylor; H: in out Taylor) is
  begin
    for I in Power loop Sum(P(I),Q(I),H(I)); end loop;
  end Sum;

  procedure Sub(P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop Sub(P(I),Q(I)); end loop;
  end Sub;

  procedure Diff(P,Q: in Taylor; H: in out Taylor) is
  begin
    for I in Power loop Diff(P(I),Q(I),H(I)); end loop;
  end Diff;

  procedure Prod(S: in Scalar; P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop Prod(S,P(I),Q(I)); end loop;
  end Prod;

  procedure AddProd(S: in Scalar; P: in Taylor; Q: in out Taylor) is
  begin
    for I in Power loop AddProd(S,P(I),Q(I)); end loop;
  end AddProd;

  function Prod_Err(R: in Radius; P1,P2: in Taylor; D1,D2: Natural) return Radius is
    use FBalls;
    E: Radius := Zero;
    S: FBall;
    F1,F2: FTaylor;
  begin
    for I in 0 .. D1 loop             --- P1 --> F1
      FltBall(P1(I),F1(I).C,F1(I).R);
    end loop;
    for I in 0 .. D2 loop             --- P2 --> F2
      FltBall(P2(I),F2(I).C,F2(I).R);
    end loop;
    for Deg in reverse PDegPlus1 .. D1+D2 loop
      S := (Zero,Zero);
      for I in Deg-D2 .. D1 loop
        AddProd(F1(I),F2(Deg-I),S);
      end loop;
      E := R*(E+MaxNorm(S));
    end loop;
    return E;
  end Prod_Err;

  procedure AddHiProd(R: in Radius; P,Q: in Taylor; HS: in out Scalar) is
    DP: constant Integer := EffLast(P);
    DQ: constant Integer := EffLast(Q);
    E: Scalar;
  begin
    if Use_FBalls then
      BallAt0(Prod_Err(R,P,Q,DP,DQ),E);
    else
      SetZero(E);
      declare
        S: Scalar;
      begin
        for Deg in reverse PDegPlus1 .. DP+DQ loop
          SetZero(S);
          for I in Deg-DQ .. DP loop
            AddProd(P(I),Q(Deg-I),S);
          end loop;
          ToErr(S);
          Add(S,E);
          Mult(R,E);
        end loop;
      end;
    end if;
    Add(E,HS);
  end AddHiProd;

  procedure Prod(R: in Radius; P,Q: in Taylor; H: in out Taylor) is
  begin
    for Deg in Power loop           -- degrees 0 .. PDeg
      declare
        HS: Scalar renames H(Deg);
      begin
        Prod(P(0),Q(Deg),HS);
        for I in 1 .. Deg loop
          AddProd(P(I),Q(Deg-I),HS);
        end loop;
      end;
    end loop;
    if Not_STrunc then
      AddHiProd(R,P,Q,H(PDeg));     -- higher degrees
    end if;
  end Prod;

  procedure Show1(N: in String; P: in Taylor; Hide0: in Boolean := True) is
    W: constant Positive := 5;
    AllZero: Boolean := True;
  begin
    for I in Power loop
      if not IsZero(P(I)) then
        Show1(N & Strng(I,W) & " ",P(I));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0"); end if;
      Show0;
    end if;
  end Show1;

  procedure Show2(N: in String; P,Q: in Taylor; Hide0: in Boolean := True) is
    W: constant Positive := 5;
    AllZero: Boolean := True;
  begin
    for I in Power loop
      if not (IsZero(P(I)) and then IsZero(Q(I))) then
        Show2(N & Strng(I,W) & " ",P(I),Q(I));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0 0"); end if;
      Show0;
    end if;
  end Show2;

  procedure NumVal(P: in Taylor; X: in Scalar; Y: in out Scalar) is
  begin
    Copy(P(PDeg),Y);
    for I in reverse 0 .. PDeg-1 loop
      Mult(X,Y);
      Add(P(I),Y);
    end loop;
  end NumVal;

  procedure Val(R: in Radius; P: in Taylor; X: in Scalar; Y: in out Scalar) is
  begin
    if IsZero(X) then
      Copy(P(0),Y);
    else
      if Not_STrunc and then MaxNorm(X)>R then raise Undefined; end if;
      Copy(P(PDeg),Y);
      for I in reverse 0 .. PDeg-1 loop
        Mult(X,Y);
        Add(P(I),Y);
      end loop;
    end if;
  end Val;

  function MaxNorm(R: Radius; P: Taylor) return Radius is
    N: Radius := Zero;
  begin
    for I in reverse 0 .. EffLast(P) loop
      N := R*N+MaxNorm(P(I));
    end loop;
    return N;
  end MaxNorm;

  procedure Norm1(R: in Radius; P: in Taylor; S: in out Scalar) is
    D: constant Integer := EffLast(P);
    Tmp: Scalar;
  begin
    Norm(P(D),S);
    for I in reverse 0 .. D-1 loop
      Mult(R,S,Tmp);
      Norm(P(I),Tmp);
      Add(Tmp,S);
    end loop;
  end Norm1;

  procedure AntiDer(P: in Taylor; Q: in out Taylor) is
    E: Scalar renames Q(0);
  begin
    for I in reverse 2 .. PDeg loop Quot(P(I-1),Flt(I),Q(I)); end loop;
    Copy(P(0),Q(1));
    if Not_STrunc then
      ToErr(P(PDeg),E);
      Div(Flt(PDegPlus1),E);
      Add(E,Q(PDeg));
    end if;
    SetZero(Q(0));
  end AntiDer;

  procedure NumValDer(P: in Taylor; X: in Scalar; Y,Tmp: in out Scalar) is
  begin
    Prod(Flt(PDeg),P(PDeg),Y);
    for I in reverse 1 .. PDeg-1 loop
      Mult(X,Y);
      AddProd(Flt(I),P(I),Y,Tmp);
    end loop;
  end NumValDer;

  procedure ValDer(R: in Radius; P: in Taylor; X: in Scalar; Y,Tmp: in out Scalar) is
    E,A,Fac: Radius;
  begin
    if IsZero(X) then
      Copy(P(1),Y);
    else
      NumValDer(P,X,Y,Tmp);
      if Not_STrunc then
        ModCenter(P(PDeg),Tmp);
        E := MaxNorm(Tmp);
        if E/=Zero then
          A := MaxNorm(X);
          Fac := (A/R)*Flt(PDegPlus2)/Flt(PDegPlus1);
          if Fac>One then raise Sorry with "X too large"; end if;
          Fac := NPower(A,PDeg)*Flt(PDegPlus1);
          BallAt0(Fac*E/R,Tmp);
          Add(Tmp,Y);
        end if;
      end if;
    end if;
  end ValDer;

  procedure SetZero(U: in out TVector) is
  begin
    for J in U'Range loop SetZero(U(J)); end loop;
  end SetZero;

  procedure Copy(U: in TVector; V: in out TVector) is
  begin
    for J in U'Range loop Copy(U(J),V(J)); end loop;
  end Copy;

  procedure Copy(A: in TMatrix; B: in out TMatrix) is
  begin
    for M in A'Range(1) loop
      for N in A'Range(2) loop Copy(A(M,N),B(M,N)); end loop;
    end loop;
  end Copy;

  procedure MaxCoeffPDegMinus1(V: in TVector; C: in out Radius) is
  begin
    for J in V'Range loop
      C := RMax(C,MaxNorm(V(J)(PDegMinus1)));
    end loop;
  end MaxCoeffPDegMinus1;

  ---------- order by order

  procedure Copy(Deg: in Power; U: in TVector; V: in out TVector) is
  begin
    for J in V'Range loop
      Copy(U(J)(Deg),V(J)(Deg));
    end loop;
  end Copy;

  procedure Sum(Deg: in Power; U,V: in TVector; W: in out TVector) is
  begin
    for J in W'Range loop
      Sum(U(J)(Deg),V(J)(Deg),W(J)(Deg));
    end loop;
  end Sum;

  procedure Diff(Deg: in Power; U,V: in TVector; W: in out TVector) is
  begin
    for J in W'Range loop
      Diff(U(J)(Deg),V(J)(Deg),W(J)(Deg));
    end loop;
  end Diff;

  procedure AddProd(Deg: in Power; S: in Scalar; U: in TVector; V: in out TVector) is
  begin
    for J in V'Range loop
      AddProd(S,U(J)(Deg),V(J)(Deg));
    end loop;
  end AddProd;

  procedure Prod(R: in Radius; Deg: in Power; P,Q: in Taylor; H: in out Taylor) is
    HS: Scalar renames H(Deg);
  begin
    Prod(P(0),Q(Deg),HS);
    for I in 1 .. Deg loop
      AddProd(P(I),Q(Deg-I),HS);
    end loop;
    if Deg=PDeg and then Not_STrunc then
      AddHiProd(R,P,Q,HS);
    end if;
  end Prod;

  procedure AddHiSqr(R: in Radius; P: in Taylor; HS: in out Scalar) is
    DP: constant Integer := EffLast(P);
    DegHalf: Integer;
    S,E: Scalar;
  begin
    SetZero(E);
    for Deg in reverse PDegPlus1 .. 2*DP loop
      SetZero(S);
      DegHalf := Deg/2;
      if 2*DegHalf=Deg then                        -- even Deg
        for I in Deg-DP .. IMin(DP,DegHalf-1) loop -- I<Deg-I
          AddProd(P(I),P(Deg-I),S);
        end loop;
        Mult(STwo,S);
        AddProd(P(DegHalf),P(DegHalf),S);
      else                                         -- odd Deg
        for I in Deg-DP .. IMin(DP,DegHalf) loop   -- I<Deg-I
          AddProd(P(I),P(Deg-I),S);
        end loop;
        Mult(STwo,S);
      end if;
      ToErr(S);
      Add(S,E);
      Mult(R,E);
    end loop;
    Add(E,HS);
  end AddHiSqr;

  procedure Old_Sqr(R: in Radius; Deg: in Power; P: in Taylor; H: in out Taylor) is
  begin
    Prod(R,Deg,P,P,H);
  end Old_Sqr;

  procedure New_Sqr(R: in Radius; Deg: in Power; P: in Taylor; H: in out Taylor) is
    HS: Scalar renames H(Deg);
    DegHalf: constant Integer := Deg/2;
  begin
    SetZero(HS);
    if 2*DegHalf=Deg then                -- even Deg
      for I in 0 .. DegHalf-1 loop       -- I<Deg-I
        AddProd(P(I),P(Deg-I),HS);
      end loop;
      Mult(STwo,HS);
      AddProd(P(DegHalf),P(DegHalf),HS);
    else                                 -- odd Deg
      for I in 0 .. DegHalf loop         -- I<Deg-I
        AddProd(P(I),P(Deg-I),HS);
      end loop;
      Mult(STwo,HS);
    end if;
    if Deg=PDeg and then Not_STrunc then
      AddHiSqr(R,P,HS);
    end if;
  end New_Sqr;

  procedure Low_Antider(Deg: in Power; DotP: in Taylor; P0: in Scalar; P: in out Taylor) is
  begin
    if Deg>1 then
      Quot(DotP(Deg-1),Flt(Deg),P(Deg));
    elsif Deg=1 then
      Copy(DotP(0),P(1));
    else
      Copy(P0,P(0));
    end if;
  end Low_Antider;

  procedure Hi_Antider(R: in Radius; DotP: in Taylor; P: in out Taylor; Tmp1,Tmp2,Tmp3: in out Scalar; Contr: in out Boolean) is
    DotS: Scalar renames DotP(PDeg);
    S:    Scalar renames P(PDeg);            -- old P(PDeg)
    Low:  Scalar renames Tmp1;
    Hi:   Scalar renames Tmp2;
    SNew: Scalar renames Tmp3;
  begin
    Quot(DotP(PDeg-1),Flt(PDeg),Low);        -- P(PDeg) without h.o. errors
    ToErr(DotS,Tmp3);
    Prod(R/Flt(PDegPlus1),Tmp3,Hi);          -- new h.o. error
    Sum(Low,Hi,SNew);                        -- new P(PDeg)
    Contr := Contr and Contains(S,SNew)=True;
    Copy(SNew,S);
  end Hi_Antider;

  procedure Inv(R: in Radius; Deg: in Power; O: in out Data_Inv; P: in Taylor; Q: in out Taylor) is
    P1: Taylor renames O(1);
    P2: Taylor renames O(2);
    S1: Scalar renames P1(Deg);
    S2: Scalar renames P2(Deg);
    Tmp: LT_Pointer;
  begin
    if Deg=0 then
      Copy(SOne,P1(0));
      Copy(SOne,P2(0));
      Inv(P(0),Q(0));
    else
      --- update P1
      Prod(Q(0),P(Deg),S1);
      --- update P2
      Copy(S1,S2);
      for I in 1 .. Deg-1 loop
        AddProd(P1(I),P2(Deg-I),S2);
      end loop;
      Neg(S2);
      if Deg=PDeg and then Not_STrunc then -- higher order
        TPool.Allocate(Tmp);
        declare
          EFac: constant Radius := NPower(One/R,PDeg);
          Del,Eps: Radius;
          H: Taylor renames Tmp.all.Data;
          H0: Scalar renames H(0);
        begin
          Copy(P1,H);
          Sub(SOne,H0);
          Del := MaxNorm(R,H);
          Del := NegOne/(Del-One);
          Prod(R,P1,P2,H);
          Sub(SOne,H0);
          Eps := Del*MaxNorm(R,H)*EFac;
          BallAt0(Eps,H0);
          Add(H0,S2);
        end;
        TPool.Recycle(Tmp);
      end if;
      --- update Q
      Prod(Q(0),S2,Q(Deg));
    end if;
  end Inv;

  procedure Sqrt(R: in Radius; Deg: in Power; O: in out Data_Sqrt; P: in Taylor; Q: in out Taylor) is
    P1: Taylor renames O(1);
    P2: Taylor renames O(2);
    S1: Scalar renames P1(Deg);
    S2: Scalar renames P2(Deg);
    Tmp: LT_Pointer;
  begin
    if Deg=0 then
      Copy(SOne,P1(0));
      Copy(SOne,P2(0));
      Sqrt(P(0),Q(0));
    else
      --- update P1
      Quot(P(Deg),P(0),S1);
        --- update P2
      Neg(S1,S2);
      for I in 1 .. Deg-1 loop
        AddProd(P2(I),P2(Deg-I),S2);
      end loop;
      Mult(SNegHalf,S2);
      if Deg=PDeg and then Not_STrunc then -- higher order
        TPool.Allocate(Tmp);
        declare
          EFac: constant Radius := NPower(One/R,PDeg);
          Del,Eps: Radius;
          H: Taylor renames Tmp.all.Data;
          H0: Scalar renames H(0);
        begin
          Copy(P1,H);
          Sub(SOne,H0);
          Del := MaxNorm(R,H);
          if Del>Half then raise Sorry; end if;
          Prod(R,P2,P2,H);
          Sub(P1,H);
          Eps := MaxNorm(R,H);
          if Eps>Quarter then raise Sorry; end if;
          Eps := (Flt(8)/Flt(5))*Eps*EFac;
          BallAt0(Eps,H0);
          Add(H0,S2);
        end;
        TPool.Recycle(Tmp);
      end if;
      --- update Q
      Prod(Q(0),S2,Q(Deg));
    end if;
  end Sqrt;

  ---------- zeros

  procedure Rough_Zero(P: in Taylor; X1,X2: in Flt; W: in out Scalar) is
    --- bisection
    Y: Scalar;
    function SP(X: Scalar) return Flt is
    begin
      NumVal(P,X,Y);
      return Approx(Y);
    end SP;
    X: Scalar;
    function RP(R: Flt) return Flt is
    begin
      Assign(R,X);
      NumVal(P,X,Y);
      return Approx(Y);
    end RP;
    procedure Rough_Root is new Bis_PSign.RoughRoot (FltF => RP, F => SP);
    Err: Flt := Zero;
  begin
    Rough_Root(X1,X2,W,Err);
    if Verbosity>3 then
      Show1("Rough_Zero Err =",Err);
    end if;
  end Rough_Zero;

  procedure Improve_Zero(P: in Taylor; X: in out Scalar) is
    --- Newton
    Steps: constant Positive := 64;
    Eps0,Eps1: Flt := Flt'Last;
    PX,DPX,DX,Tmp: Scalar;
  begin
    for I in 1 .. Steps loop -- X -> X-P(X)/DP(X)
      Center(X);
      Eps0 := Eps1;
      NumVal(P,X,PX);
      NumValDer(P,X,DPX,Tmp);
      Quot(PX,DPX,DX);
      Eps1 := MaxNorm(DX);
      if Eps1=Zero or else (Eps1 >= Eps0) then return; end if;
      Sub(DX,X);
    end loop;
    Show1("Improve_Zero: Warning, Eps only",Eps1);
  end Improve_Zero;

  procedure FindZero(R: in Radius; P: in Taylor; X1,X2: in Flt; X: in out Scalar) is
  begin
    Rough_Zero(P,X1,X2,X);
    Improve_Zero(P,X);
    if Not_STrunc then
      declare
        N: constant Positive := 8;     -- want contraction factor 1/(N+1)
        E,K: Flt;
        PX,DPX,DX,Tmp: Scalar;
        DPXB: Scalar renames DX;
      begin
        Center(X);
        Val(R,P,X,PX);
        ValDer(R,P,X,DPX,Tmp);
        Quot(PX,DPX,DX);               -- DX := P(X)/DP(X)
        E := MaxNorm(DX)*Flt(N+1)/Flt(N);
        BallAt0(E,Tmp);                -- Ball of radius ||DS||*(N+1)/N
        Add(Tmp,X);                    -- X+Ball
        ValDer(R,P,X,DPXB,Tmp);
        Quot(DPXB,DPX,Tmp);
        Sub(SOne,Tmp);
        K := MaxNorm(Tmp)*Flt(N+1);  -- K := |DP(X+Ball)/DP(X)-1|*(N+1)
        if K>One then
          Show1("K =",K);
          raise Sorry;
        end if;
      end;
    end if;
  end FindZero;

  procedure Put(F: in File_Type; P: in Taylor; Decimal: in Boolean := False) is
  begin
    Put(F,PDeg);
    for I in 0 .. PDeg loop
      Put(F,P(I),Decimal);
    end loop;
  end Put;

  procedure Get(F: in File_Type; P: in out Taylor; Decimal: in Boolean := False) is
     I: Natural := 0;
  begin
     SetZero(P);
     Get(F,I);
     if I/=PDeg then raise Not_Implemented; end if;
    for I in 0 .. PDeg loop
      Get(F,P(I),Decimal);
    end loop;
  end Get;

  procedure Write(FileName: in String; P: in Taylor; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Taylor; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

end ObO;
