with Ada.Text_IO, Logicals, Rationals, Flts, ScalInfo;
use Ada.Text_IO, Logicals, Rationals, Flts, ScalInfo;
with Pooling, FBalls;

pragma Elaborate_All (Ada.Text_IO,Logicals,Rationals,Flts,ScalInfo);
pragma Elaborate_All (Pooling);

generic

  PDeg: in Positive; --- max degree

  -------------------- begin_include scalars Scalar

  type Scalar is private;
  --with procedure Proper_Rounding is <>;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  --with function Center0(S: Scalar) return Boolean is <>;
  --with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  --with function ToErr(S: Scalar) return Scalar is <>;
  with procedure FltBall(S: in Scalar; C,R: out Flt) is <>;
  with procedure Center(S: in out Scalar) is <>;
  --with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  --with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  --with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  --with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  --with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  --with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  --with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  --with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  --with function Inv(S: Scalar) return Scalar is <>;
  --with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  --with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;

  --- fun
  --with function Sign(S: Scalar) return Integer is <>;
  --with procedure Eval0(S: in out Scalar) is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  --with function Scal(I: Integer) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  --with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  --with procedure Reduce_Precision(Binary_Digits: in Positive; S: in out Scalar) is <>;

  -------------------- end_include scalars

package ObO is  --- Taylor series for order-by-order ops

  subtype Power is Integer range 0 .. PDeg;
  type Taylor is array(Power) of Scalar; -- last coeff includes general error

  Use_FBalls: Boolean := False;  -- useful when Scalar = MPFR.Floats.Balls.Ball

  procedure Show_ObO_Param;

  ---------- all degrees
  function EffLast(P: Taylor) return Power;
  function IsZero(P: Taylor) return Boolean;
  function IsPoly(P: Taylor) return Boolean;
  procedure SetZero(P: in out Taylor);
  procedure Assign(S: in Scalar; P: in out Taylor);
  procedure Copy(P: in Taylor; Q: in out Taylor);
  procedure Neg(P: in Taylor; Q: in out Taylor);
  procedure Add(P: in Taylor; Q: in out Taylor);
  procedure Sum(P,Q: in Taylor; H: in out Taylor);
  procedure Sub(P: in Taylor; Q: in out Taylor);
  procedure Diff(P,Q: in Taylor; H: in out Taylor);
  procedure Prod(S: in Scalar; P: in Taylor; Q: in out Taylor);
  procedure AddProd(S: in Scalar; P: in Taylor; Q: in out Taylor);
  procedure Prod(R: in Radius; P,Q: in Taylor; H: in out Taylor);
  procedure Show1(N: in String; P: in Taylor; Hide0: in Boolean := True);
  procedure Show2(N: in String; P,Q: in Taylor; Hide0: in Boolean := True);
  procedure NumVal(P: in Taylor; X: in Scalar; Y: in out Scalar);
  procedure Val(R: in Radius; P: in Taylor; X: in Scalar; Y: in out Scalar);
  function MaxNorm(R: Radius; P: Taylor) return Radius;
  procedure Norm1(R: in Radius; P: in Taylor; S: in out Scalar);
  procedure AntiDer(P: in Taylor; Q: in out Taylor);
  procedure NumValDer(P: in Taylor; X: in Scalar; Y,Tmp: in out Scalar);
  procedure ValDer(R: in Radius; P: in Taylor; X: in Scalar; Y,Tmp: in out Scalar);

  PDegMinus1: constant Integer := PDeg-1;
  PDegPlus1:  constant Integer := PDeg+1;
  PDegPlus2:  constant Integer := PDeg+2;

  type TVector is array(Integer range <>) of Taylor;
  type TMatrix is array(Integer range <>, Integer range <>) of Taylor;

  procedure SetZero(U: in out TVector);
  procedure Copy(U: in TVector; V: in out TVector);
  procedure Copy(A: in TMatrix; B: in out TMatrix);
  procedure MaxCoeffPDegMinus1(V: in TVector; C: in out Radius);

  ---------- order by order
  procedure Copy(Deg: in Power; U: in TVector; V: in out TVector);
  procedure Sum(Deg: in Power; U,V: in TVector; W: in out TVector);
  procedure Diff(Deg: in Power; U,V: in TVector; W: in out TVector);
  procedure AddProd(Deg: in Power; S: in Scalar; U: in TVector; V: in out TVector);

  procedure Prod(R: in Radius; Deg: in Power; P,Q: in Taylor; H: in out Taylor);
  procedure Old_Sqr(R: in Radius; Deg: in Power; P: in Taylor; H: in out Taylor);
  procedure New_Sqr(R: in Radius; Deg: in Power; P: in Taylor; H: in out Taylor);
  procedure Sqr(R: in Radius; Deg: in Power; P: in Taylor; H: in out Taylor) renames New_Sqr;
  procedure Low_Antider(Deg: in Power; DotP: in Taylor; P0: in Scalar; P: in out Taylor);
  procedure Hi_Antider(R: in Radius; DotP: in Taylor; P: in out Taylor; Tmp1,Tmp2,Tmp3: in out Scalar; Contr: in out Boolean);

  subtype Data_Inv is TVector(1..2);
  procedure Inv(R: in Radius; Deg: in Power; O: in out Data_Inv; P: in Taylor; Q: in out Taylor);

  subtype Data_Sqrt is TVector(1..2);
  procedure Sqrt(R: in Radius; Deg: in Power; O: in out Data_Sqrt; P: in Taylor; Q: in out Taylor);

  ---------- zeros
  procedure FindZero(R: in Radius; P: in Taylor; X1,X2: in Flt; X: in out Scalar);

  ---------- input output
  procedure Put(F: in File_Type; P: in Taylor; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Taylor; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Taylor; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Taylor; Decimal: in Boolean := False);

  ---------- pooling
  procedure CP(P1: in Taylor; P2: in out Taylor) renames Copy;
  package TP is new Pooling (Item => Taylor);
  subtype Pooled_Taylor is TP.Controlled_LD_Pointer; --- not used here

  package TAlloc renames TP.Alloc;
  subtype LT_Pointer is TAlloc.LD_Pointer;
  TAC: TAlloc.Controlled_Protected_Data;
  TPool: TAlloc.Protected_Data renames TAC.P;

private

  function Prod_Err(R: in Radius; P1,P2: in Taylor; D1,D2: Natural) return Radius;
  procedure AddHiProd(R: in Radius; P,Q: in Taylor; HS: in out Scalar);
  procedure AddHiSqr(R: in Radius; P: in Taylor; HS: in out Scalar);
  procedure Rough_Zero(P: in Taylor; X1,X2: in Flt; W: in out Scalar); -- bisection
  procedure Improve_Zero(P: in Taylor; X: in out Scalar); -- Newton

  SZero:      constant Scalar      := Scal(Zero);
  SOne:       constant Scalar      := Scal(One);
  STwo:       constant Scalar      := Scal(Two);
  SNegHalf:   constant Scalar      := Scal(NegHalf);
  SInfo:      constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean     := SInfo.IsNumeric;
  Not_STrunc: constant Boolean     := not STrunc;

  type FTaylor is array(Power) of FBalls.FBall;

end ObO;
