with Ada.Text_IO, Interfaces, Flts;
use Ada.Text_IO, Interfaces, Flts;

pragma Elaborate_All(Ada.Text_IO,Interfaces,Flts);

package Params is

  GAccel: constant Flt := Flt(40141)/Flt(4096); --- 9.800048828125

  type FRParam is
    record
      I11,I12,I22,I33: Flt;   --- inertia tensor
      BB1,BB2,BB3: Flt;       --- geometry
      Mass: Flt;              --- mass
    end record;

  --- conversions
  procedure Approx_Inertia(I1,I2,Del: in Flt; I11,I12,I22: out Flt);
  procedure Approx_Delta(I11,I12,I22: in Flt; I1,I2,Del: out Flt);
  function Make_Approx_Param(I1,I2,I3,Del,BB1,BB2,BB3,Mass: Flt) return FRParam; -- no Reduce_Precision
  procedure Reduce_Precision(Binary_Digits: in Positive; P: in out FRParam);
  function Approx_Homog_Param(Binary_Digits: in Positive; Mass,B1,B2,B3,Del: in Flt) return FRParam;

  function Approx_EMin(P: FRParam) return Flt;
  procedure Show(P: in FRParam);
  procedure Put(F: in File_Type; P: in FRParam; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out FRParam; Decimal: in Boolean := False);

  Param0: constant FRParam --- 9 x 3 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>  5.0,
        BB1 =>  81.0,
        BB2 =>   9.0,
        BB3 =>   4.0,
        --- I1=BB2+BB3=13, I2=BB3+BB1=85, I3=BB1+BB2=90
        --- for Delta = Pi/20 = 0.157079... this yields roughly
        I11 =>  14.75,
        I12 => -11.125,
        I22 =>  83.25,
        I33 =>  90.0
       );   --- no critical Mo3

  Param1: constant FRParam --- near-original choice, 10 x 1.15 x 1
    := (
        Mass => 10.0,
        BB1 => 100.0,
        BB2 => Flt( 710011781)/Flt(536870912), --  1.32249999...
        BB3 => 1.0,
        I11 => Flt( 655114409)/Flt(536870912), --  1.22024567...
        I12 => Flt(-186640015)/Flt(134217728), -- -1.39057647...
        I22 => Flt( 656308203)/Flt( 67108864), --  9.77975432...
        I33 => 10.0
       );   --- critical Mo3=26.25245032784939214303726879434...

  Param2: constant FRParam --- new choice, 10 x 1.118... x 1
    := (
        Mass => 10.0,
        BB1 => 100.0,
        BB2 =>   1.25,
        BB3 =>   1.0,
        I11 =>   1.25,
        I12 =>  -1.375,
        I22 =>  10.0,
        I33 =>  10.0
       );   --- critical Mo3=19.75419420354122466418987878883884...

  Param3: constant FRParam --- 12 x 3 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>  5.0,
        BB1 => 144.0,
        BB2 =>   9.0,
        BB3 =>   4.0,
        --- I1=BB2+BB3=13, I2=BB3+BB1=85, I3=BB1+BB2=90
        --- for Delta = Pi/20 = 0.157079... this yields roughly
        I11 =>  16.3046875,
        I12 => -20.859375,
        I22 => 144.6953125,
        I33 => 153.0
       );   --- no critical Mo3

  Param4: constant FRParam --- 10 x 2.5 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>  5.0,
        BB1 => 100.0,
        BB2 =>   6.25,
        BB3 =>   4.0,
        --- I1=BB2+BB3=10.25, I2=BB3+BB1=104.0, I3=BB1+BB2=106.25
        --- for Delta = Pi/20 = 0.157079... this yields roughly
        I11 =>  12.5439453125,
        I12 => -14.4853515625,
        I22 => 101.70703125,
        I33 => 106.25
       );  --- critical Mo3=317.91494735950286872377357291930297106...

  Param5: constant FRParam --- 9 x 17/8 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>  5.0,
        BB1 =>  81.0,
        BB2 =>   4.515625,
        BB3 =>   4.0,
        --- I1=BB2+BB3=8.515625, I2=BB3+BB1=85.0, I3=BB1+BB2=85.515625
        --- for Delta = Pi/20 = 0.157079... this yields roughly
        I11 =>  10.38720703125,
        I12 => -11.8173828125,
        I22 =>  83.1279296875,
        I33 =>  85.515625
       );   --- critical Mo3=63.33480839374123097097132259376316647929...

  Param6: constant FRParam --- 10 x 3 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>   5.0,
        BB1 =>  100.0,
        BB2 =>    9.0,
        BB3 =>    4.0,
        --- I1=BB2+BB3=13, I2=BB3+BB1=104, I3=BB1+BB2=109
        --- for Delta = Pi/12=0.261799... this yields roughly    (notice Pi/12)
        I11 =>  19.095703125,
        I12 => -22.75,
        I22 =>  97.900390625,
        I33 => 109.0
       );   --- no critical Mo3

  Param7: constant FRParam --- 11 x 3 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>   5.0,
        BB1 =>  121.0,
        BB2 =>    9.0,
        BB3 =>    4.0,
        --- I1=BB2+BB3=13, I2=BB3+BB1=125, I3=BB1+BB2=130
        --- for Delta = Pi/24=0.1308996... this yields roughly    (notice Pi/24)
        I11 =>  14.908203125,
        I12 => -14.494140625,
        I22 => 123.091796875,
        I33 => 130.0
       );   --- no critical Mo3

  Param8: constant FRParam --- 4 x 5/4 x 1 radii ellipsoid with mass 10 has
    := (
        Mass =>  10.0,
        BB1 =>   16.0,
        BB2 =>    1.5625,
        BB3 =>    1.0,
        --- for Delta = Pi/20 this yields roughly
        I11 =>    5.83203125,
        I12 =>   -4.4609375,
        I22 =>   33.296875,
        I33 =>   35.125
       ); -- critical Mo3=144.0182453929197633272581155686050731...

  Param9: constant FRParam --- 7 x 9/4 x 2 radii ellipsoid with mass 5 has
    := (
        Mass =>  5.0,
        BB1 =>  49.0,
        BB2 =>   5.0625,
        BB3 =>   4.0,
        --- for Delta = Pi/20 this yields roughly
        I11 =>  10.13671875,
        I12 =>  -6.7890625,
        I22 =>  51.921875,
        I33 =>  54.0625
       ); -- critical Mo3=64.95631999312413175880936267037743150...

  Param10: constant FRParam --- 10 x 3 x 2 radii ellipsoid with mass 1 has
    := (
        Mass =>  1.0,
        BB1 => 100.0,
        BB2 =>   9.0,
        BB3 =>   4.0,
        --- for Delta = Pi/20 this yields roughly
        I11 =>   3.044921875,
        I12 =>  -2.8125,
        I22 =>  20.3515625,
        I33 =>  21.796875
       );   --- no (useful) critical Mo3

  Param11: constant FRParam --- 5 x 5/4 x 1 radii ellipsoid with mass 1 has
    --- bad for hetero with stable limit, rolls over
    := (
        Mass =>  1.0,
        BB1 =>  25.0,
        BB2 =>   1.5625,
        BB3 =>   1.0,
        --- for Delta = Pi/20 this yields roughly
        I11 =>  0.627197265625,
        I12 => -0.724365234375,
        I22 =>  5.0859375,
        I33 =>  5.3125
       );   -- critical Mo3=22.45811961047807699268658253949903...

  Param12: constant FRParam --- 5 x 9/8 x 1 radii ellipsoid with mass 10 has
    := (
        Mass => 10.0,
        BB1 =>  25.0,
        BB2 =>   1.265625,
        BB3 =>   1.0,
        --- for Delta = Pi/20 this yields roughly
        I11 =>  5.693359375,
        I12 => -7.333984375,
        I22 => 50.84375,
        I33 => 52.53125
       );   -- critical Mo3=90.102043063904285813189171088...
            -- M3=99: -1.27904850053822279E-002 +i* 1.79987411667616803E+000

  Param13: constant FRParam --- 5 x 17/16 x 1 radii ellipsoid with mass 4 has
    := (
        Mass => Flt(4),
        BB1 =>  Flt(25),
        BB2 =>  Flt(289)*Half**8,    ---  1.12890625
        BB3 =>  One,
        --- for Delta = Pi/20 this yields roughly
        I11 =>  Flt(2223)*Half**10,  ---  2.1708984375
        I12 => -Flt(3021)*Half**10,  --- -2.9501953125
        I22 =>  Flt(2603)*Half**7,   --- 20.3359375
        I33 =>  Flt(669)*Half**5     --- 20.90625
       );   -- E_min=39.2001953125
            -- critical Mo3=21.91320185202643773473397531442913436945997...
            -- M3=27: -1.25190862582119224E-002 +i* 1.12038683667665040E+000
            -- M3=25: -6.81593997827483892E-003 +i* 1.09037393781061147E+000

  ------------------------------------------------------------------
  ------------------------------------------------------------------

  package Mod_IO is new Ada.Text_IO.Modular_IO (Num => Unsigned_16);
  subtype Flag is Unsigned_16;

  procedure Show(N: in String; P: in Flag);

  Alpha_Defined:      constant Flag := 2#0000_0000_0000_0001#;
  Beta_Defined:       constant Flag := 2#0000_0000_0000_0100#;
  Yaw_Defined:        constant Flag := 2#0000_0000_0001_0101#; -- also includes Alpha_Defined and Beta_Defined
  Roll_Defined:       constant Flag := 2#0000_0000_0010_0000#;

  --- "or" sets both flags
  AlphaBeta_Defined:  constant Flag := Alpha_Defined or Beta_Defined;
  YawRoll_Defined:    constant Flag := Yaw_Defined or Roll_Defined;

  function Choose(ChooseF0: Boolean; F0,F1: Flag) return Flag;

end Params;
