with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps;
with Ada.Numerics, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps);
pragma Elaborate_All (Ada.Numerics,Rattleback.Flows);

procedure Plot_Orbit is --- using Param13 as default

  NT:           constant Positive := 8;     -- number of concurrent tasks
  PDeg:         constant Positive := 255;   -- Taylor degree
  Prec:         constant Positive := 800;   -- target precision for Integrate
  NegLogMinRad: constant Positive := 8;     -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;    -- MaxRad := MaxRadFac*MinRad

  FRP: constant FRParam := Param13;         -- CHANGE if incorrect
  ABFlag: constant Flag := YawRoll_Defined; -- CHANGE if incorrect

  NDigits:      constant Positive := 2048;
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Rep;
  package RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  package RBFlow is new RB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use RB,RB.PMore,RBFlow;

  procedure Plot_Flow is
    FileName: constant String := GetArg; -- specify input file
    TotalTime: constant Flt   := GetArg; -- specify integration time
    T,DT: Flt;
    XT: Point;
    X,DotX: Curve;
  begin
    Show1(FileName & " up to time ",TotalTime);
    Read(Input_Dir & FileName,XT);
    ShowApprox1("initial ",XT);
    Show_Approx_Constraints(XT);
    ---
    Prepare_Plot(Zero,TotalTime,FileName & ".plot");
    Plain_Flow(TotalTime,T,DT,XT,X,DotX);
    ShowApprox1("  final ",XT);
    Show_Approx_Constraints(XT);
  end Plot_Flow;

begin
  Verbosity := 3;
  NTasks := NT;
  Show_Rattle_Param;

  Plot_Flow;

end Plot_Orbit;
