with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
with Tangents, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps);
pragma Elaborate_All (Tangents,Rattleback.Flows);

procedure R_Der is --- derivative at symmetric periodic orbit

  NT:           constant Positive := 8;       -- number of concurrent tasks
  PDeg:         constant Positive := 511;     -- Taylor degree
  Prec:         constant Positive := 500;     -- target precision for Integrate
  NegLogMinRad: constant Positive := 8;       -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;      -- MaxRad := MaxRadFac*MinRad
  Rho:          constant Radius   := One;     -- for Tangents
  LogFac:       constant Positive := 64;      -- Half**LogFac is the size of the initial tangent vectors
  FRP:          constant FRParam  := Param13; -- rattleback parameters
  ABFlag:       constant Flag     := 0;       -- no need to compute Alpha and Beta

  NDigits:      constant Positive := 4096;    -- number of binary digits
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package Tangs is new Tangents (Dim => 2, Rho => Rho, Scalar => Scalar);
  package SRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  use Tangs;
  package TRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Tangent);
  package TPMore renames TRB.PMore;
  package TMS renames TRB.MS;
  package TSMV renames TRB.SMV;
  use TRB, TPMore,TSMV;
  package TRBFlows is new TRB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  package TRObO renames TRBFlows.RObO;
  use TRBFlows;

  File_Name: constant String := "P13-R.0006"; -- specifies initial Scalar Point
  TNeed: constant Flt := 113.0;
  F: constant Flt_Affine := (Zero,One,Zero,Zero);

  procedure Make_Initial(FileName: in String; P: in out Point) is
    Decimal: constant Boolean := False;
    TSize: constant Radius := TwoPower(-LogFac);
    Eps: constant Radius := TwoPower(-2000);
    E: Scalar;
    SP: SRB.Point;
    G1,G2: Tangent;
  begin
    SRB.Read(FileName,SP,Decimal);
    Copy(SP.Gamma(1),G1(0));
    DiskAt0(Eps,E);
    Add(E,G1(0));
    Assign(TSize,G1(1));
    SetZero(G1(2));
    Copy(SP.Gamma(2),G2(0));
    Add(E,G2(0));
    SetZero(G2(1));
    Assign(TSize,G2(2));
    Enlarge(G1);             -- for open domain
    Enlarge(G2);             -- for open domain
    MakeZeroM(G1,G2,P);
    Enlarge(P.M);            -- for open domain
  end Make_Initial;

  procedure Save_Der(FileName: in String; P: in Point) is
    Decimal: constant Boolean := False;
    InvTSize: constant Radius := TwoPower(LogFac);
    SDer: SRB.SMatrix2;
  begin
    Prod(InvTSize,P.M(2)(1),SDer(1,1));
    Prod(InvTSize,P.M(2)(2),SDer(1,2));
    Prod(InvTSize,P.M(3)(1),SDer(2,1));
    Prod(InvTSize,P.M(3)(2),SDer(2,2));
    SRB.Write(FileName,SDer,Decimal);
  end Save_Der;

  P0,P1: Point;

begin
  NTasks := NT;
  Verbosity := 3;
  Show_Rattle_Param;

  Make_Initial(Input_Dir & File_Name,P0);
  Neg2Pos_Poincare(TNeed,F,P0,P1);
  Write(Output_Dir & File_Name & ".tan.final",P1);   -- Affine Point
  Save_Der(Output_Dir & File_Name & ".Der",P1);      -- Scalar SMatrix2

end R_Der;
