with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls, Protected_Counters;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls, Protected_Counters;
with Ada.Numerics, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps,Protected_Counters);
pragma Elaborate_All (Ada.Numerics,Rattleback.Flows);

procedure R_Point is --- R-symmetric periodic orbit, Poincare with F=(0,1,0,0)

  NT:           constant Positive := 8;       -- number of concurrent tasks
  PDeg:         constant Positive := 767;     -- Taylor degree
  Prec:         constant Positive := 1000;    -- target precision for Integrate
  NegLogMinRad: constant Positive := 6;       -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;      -- MaxRad := MaxRadFac*MinRad
  FRP:          constant FRParam  := Param13; -- rattleback parameters
  Run:          constant Natural  := GetArg;  -- 0 for the center, 1 for the interval, 2 for Check_Final
  ABFlag:       constant Flag     := Choose(Run=0,0,YawRoll_Defined);

  NDigits:      constant Positive := 4096;    -- number of binary digits
  Dummy:        constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  use RB,RB.SMV;
  package RBFlow is new RB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use PMore,RBFlow;

  Stem: constant String := "P13-R";

  procedure Make_Initial(FileName: in String; P: in out Point) is
    Decimal: constant Boolean := False;
    Eps: constant Radius := TwoPower(-2000);
    SP: Point;
    E,G1,G2: Scalar;
  begin
    Read(FileName & ".0006",SP,Decimal);
    Copy(SP.Gamma(1),G1);
    Copy(SP.Gamma(2),G2);
    Diskat0(Eps,E);
    if Run=1 then Add(E,G1); Add(E,G2); end if;  -- removed Enlarge when Run=0
    MakeZeroM(G1,G2,P);
    --if Run=0 then Enlarge(P.M); end if;        -- also seems useless
  end Make_Initial;

  procedure PFun(P0: in out Point) is
    TNeed: constant Flt := 113.0;
    F: constant Flt_Affine := (Zero,One,Zero,Zero);
    P1: Point;
    X,Y: SVector2;
  begin
    Verbosity := 1;
    Copy(P0.Gamma(1),X(1));
    Copy(P0.Gamma(2),X(2));
    MakeZeroM(X(1),X(2),P0);
    ShowApprox1("initial ",P0);
    --if Run=1 then Prepare_Plot(Zero,115.0,Pic_Dir & Stem & ".plot"); end if; -- plot if interval
    Neg2Pos_Poincare(TNeed,F,P0,P1);
    Write(Output_Dir & Stem & "." & Strng(Run) & ".final",P1);
    Copy(P1.M(2),Y(1));
    Copy(P1.M(3),Y(2));
    Show2("y=",Y(1),Y(2));
  end PFun;

  procedure Check_Final is
    R: constant Radius := TwoPower(-2000);
    K,Epsilon: Radius := Zero;
    SOne: Scalar;
    X,Y: SVector2;
    DPFun,MC,DC: SMatrix2;
    P1: Point;
  begin
    Read(Input_Dir & Stem & ".Der",DPFun);
    Inv2(DPFun,MC);
    Center2(MC);        --- quasi-Newton C(X) = X-MC*PFun(X)
    Prod2(MC,DPFun,DC);
    Neg2(DC);
    Assign(One,SOne);
    Add(SOne,DC(1,1));
    Add(SOne,DC(2,2));  --- DC = I-MC*DPFun
    K := Maxnorm(DC);

    Read(Input_Dir & Stem & ".0.final",P1); -- written by PFun with Run=0
    Copy(P1.M(2),X(1));
    Copy(P1.M(3),X(2));
    Prod2(MC,X,Y);
    Epsilon := MaxNorm(Y);

    Show2("K,eps",K,Epsilon);
    Show2("r,Kr+eps",R,K*R+Epsilon);
  end Check_Final;

  P0: Point;

begin
  NTasks := NT;
  Show_Rattle_Param;
  if Run=2 then
    Check_Final;
  else
    Make_Initial(Input_Dir & Stem,P0);
    PFun(P0);
  end if;

end R_Point;
