with Strings, Globals, Protected_Counters;
use Strings, Globals, Protected_Counters;

pragma Elaborate_All (Strings,Globals,Protected_Counters);

package body Rattleback.Flows is

  procedure Show_Rattle_Param is
  begin
    Show_Obo_Param;
    Show1("         Prec = ",Prec);
    Show1(" NegLogMinRad = ",NegLogMinRad);
    Show1("  ---> MinRad =",MinRad);
    Show1("    MaxRadFac = ",MaxRadFac);
    Show( "       ABFlag = ",ABFlag);
    Show0;
    Show(FRP);
    Show0;
  end Show_Rattle_Param;

  procedure Copy(C1: in Curve; C2: in out Curve) is
  begin
    if Compute_Alpha then
      Copy(C1.Alpha,C2.Alpha);
    end if;
    if Compute_Beta then
      Copy(C1.Beta,C2.Beta);
    end if;
    Copy(C1.Gamma,C2.Gamma);
    Copy(C1.M,C2.M);
  end Copy;

  procedure Val(R: in Radius; U: in TVector3; S: in Scalar; V: in out SVector3) is
  begin
    Val(R,U(1),S,V(1));
    Val(R,U(2),S,V(2));
    Val(R,U(3),S,V(3));
  end Val;

  procedure Val(R: in Radius; C: in Curve; S: in Scalar; P: in out Point) is
  begin
    if Compute_Alpha then
      Val(R,C.Alpha,S,P.Alpha);
    end if;
    if Compute_Beta then
      Val(R,C.Beta,S,P.Beta);
    end if;
    Val(R,C.Gamma,S,P.Gamma);
    Val(R,C.M,S,P.M);
  end Val;

  ---------------------------

  procedure Diag_Prod(Deg: in Power; B: in SVector3; U: in TVector3; V: in out TVector3) is
    --- V := diag(B)U
  begin
    Prod(B(1),U(1)(Deg),V(1)(Deg));
    Prod(B(2),U(2)(Deg),V(2)(Deg));
    Prod(B(3),U(3)(Deg),V(3)(Deg));
  end Diag_Prod;

  procedure TV_Prod(R: in Radius; Deg: in Power; P: in Taylor; U: in TVector3; V: in out TVector3) is
    --- V := P*U
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        exit when J>3;
        Prod(R,Deg,P,U(J),V(J));
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(2,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "TV_Prod Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
  end TV_Prod;

  procedure VV_Prod(R: in Radius; Deg: in Power; U,V: in TVector3; W: in out TVector3) is
    --- W(I) := U(I)*V(I)
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        exit when J>3;
        Prod(R,Deg,U(J),V(J),W(J));
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(2,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "VV_Prod Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
  end VV_Prod;

  procedure Dot(R: in Radius; Deg: in Power; O: in out Data_Dot; U,V: in TVector3; P: in out Taylor) is
    --- P := U*V
    S: Scalar renames P(Deg);
  begin
    VV_Prod(R,Deg,U,V,O);
    Sum(O(1)(Deg),O(2)(Deg),S); Add(O(3)(Deg),S);
  end Dot;

  procedure Cross1(R: in Radius; Deg: in Power; O: in out Data_Cross; U,V: in TVector3; W: in out TVector3) is
    --- W := U x V
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(R,Deg,U(2),V(3),O(1));
          when 2 => Prod(R,Deg,U(3),V(2),O(2));
          when 3 => Prod(R,Deg,U(3),V(1),O(3));
          when 4 => Prod(R,Deg,U(1),V(3),O(4));
          when 5 => Prod(R,Deg,U(1),V(2),O(5));
          when 6 => Prod(R,Deg,U(2),V(1),O(6));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(5,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ----
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Cross Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Diff(O(1)(Deg),O(2)(Deg),W(1)(Deg));
    Diff(O(3)(Deg),O(4)(Deg),W(2)(Deg));
    Diff(O(5)(Deg),O(6)(Deg),W(3)(Deg));
  end Cross1;

  procedure Cross2(R: in Radius; Deg: in Power; O1,O2: in out Data_Cross; U1,U2,V1,V2: in TVector3; W1,W2: in out TVector3) is
    --- Wj := Uj x Vj
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when  1 => Prod(R,Deg,U1(2),V1(3),O1(1));
          when  2 => Prod(R,Deg,U1(3),V1(2),O1(2));
          when  3 => Prod(R,Deg,U1(3),V1(1),O1(3));
          when  4 => Prod(R,Deg,U1(1),V1(3),O1(4));
          when  5 => Prod(R,Deg,U1(1),V1(2),O1(5));
          when  6 => Prod(R,Deg,U1(2),V1(1),O1(6));
          ---
          when  7 => Prod(R,Deg,U2(2),V2(3),O2(1));
          when  8 => Prod(R,Deg,U2(3),V2(2),O2(2));
          when  9 => Prod(R,Deg,U2(3),V2(1),O2(3));
          when 10 => Prod(R,Deg,U2(1),V2(3),O2(4));
          when 11 => Prod(R,Deg,U2(1),V2(2),O2(5));
          when 12 => Prod(R,Deg,U2(2),V2(1),O2(6));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(11,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ----
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Cross Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Diff(O1(1)(Deg),O1(2)(Deg),W1(1)(Deg));
    Diff(O1(3)(Deg),O1(4)(Deg),W1(2)(Deg));
    Diff(O1(5)(Deg),O1(6)(Deg),W1(3)(Deg));
    ---
    Diff(O2(1)(Deg),O2(2)(Deg),W2(1)(Deg));
    Diff(O2(3)(Deg),O2(4)(Deg),W2(2)(Deg));
    Diff(O2(5)(Deg),O2(6)(Deg),W2(3)(Deg));
  end Cross2;

  procedure Cross3(R: in Radius; Deg: in Power; O1,O2,O3: in out Data_Cross; U1,U2,U3,V1,V2,V3: in TVector3; W1,W2,W3: in out TVector3) is
    --- Wj := Uj x Vj
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when  1 => Prod(R,Deg,U1(2),V1(3),O1(1));
          when  2 => Prod(R,Deg,U1(3),V1(2),O1(2));
          when  3 => Prod(R,Deg,U1(3),V1(1),O1(3));
          when  4 => Prod(R,Deg,U1(1),V1(3),O1(4));
          when  5 => Prod(R,Deg,U1(1),V1(2),O1(5));
          when  6 => Prod(R,Deg,U1(2),V1(1),O1(6));
          ---
          when  7 => Prod(R,Deg,U2(2),V2(3),O2(1));
          when  8 => Prod(R,Deg,U2(3),V2(2),O2(2));
          when  9 => Prod(R,Deg,U2(3),V2(1),O2(3));
          when 10 => Prod(R,Deg,U2(1),V2(3),O2(4));
          when 11 => Prod(R,Deg,U2(1),V2(2),O2(5));
          when 12 => Prod(R,Deg,U2(2),V2(1),O2(6));
          ---
          when 13 => Prod(R,Deg,U3(2),V3(3),O3(1));
          when 14 => Prod(R,Deg,U3(3),V3(2),O3(2));
          when 15 => Prod(R,Deg,U3(3),V3(1),O3(3));
          when 16 => Prod(R,Deg,U3(1),V3(3),O3(4));
          when 17 => Prod(R,Deg,U3(1),V3(2),O3(5));
          when 18 => Prod(R,Deg,U3(2),V3(1),O3(6));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(17,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ----
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Cross Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Diff(O1(1)(Deg),O1(2)(Deg),W1(1)(Deg));
    Diff(O1(3)(Deg),O1(4)(Deg),W1(2)(Deg));
    Diff(O1(5)(Deg),O1(6)(Deg),W1(3)(Deg));
    ---
    Diff(O2(1)(Deg),O2(2)(Deg),W2(1)(Deg));
    Diff(O2(3)(Deg),O2(4)(Deg),W2(2)(Deg));
    Diff(O2(5)(Deg),O2(6)(Deg),W2(3)(Deg));
    ---
    Diff(O3(1)(Deg),O3(2)(Deg),W3(1)(Deg));
    Diff(O3(3)(Deg),O3(4)(Deg),W3(2)(Deg));
    Diff(O3(5)(Deg),O3(6)(Deg),W3(3)(Deg));
  end Cross3;

  procedure SM_AddProd(Deg: in Power; A: in SMatrix3; S: in Scalar; B: in TMatrix3; C: in out TMatrix3) is
    --- C := A+S*B
  begin
    for I in 1 .. 3 loop
      Prod(S,B(I,1)(Deg),C(I,1)(Deg));
      Prod(S,B(I,2)(Deg),C(I,2)(Deg));
      Prod(S,B(I,3)(Deg),C(I,3)(Deg));
    end loop;
    if Deg=0 then
      for I in 1 .. 3 loop
        Add(A(I,1),C(I,1)(0));
        Add(A(I,2),C(I,2)(0));
        Add(A(I,3),C(I,3)(0));
      end loop;
    end if;
  end SM_AddProd;

  procedure Symm_TM_Prod(R: in Radius; Deg: in Power; P: in Taylor; A: in TMatrix3; B: in out TMatrix3) is
    --- B := P*A
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(R,Deg,P,A(1,1),B(1,1));
          when 2 => Prod(R,Deg,P,A(1,2),B(1,2));
          when 3 => Prod(R,Deg,P,A(1,3),B(1,3));
          when 4 => Prod(R,Deg,P,A(2,2),B(2,2));
          when 5 => Prod(R,Deg,P,A(2,3),B(2,3));
          when 6 => Prod(R,Deg,P,A(3,3),B(3,3));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(5,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ----
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Symm_TM_Prod Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Copy(B(1,2)(Deg),B(2,1)(Deg));
    Copy(B(1,3)(Deg),B(3,1)(Deg));
    Copy(B(2,3)(Deg),B(3,2)(Deg));
  end Symm_TM_Prod;

  procedure MV_Prod(R: in Radius; Deg: in Power; O: in out Data_MV_Prod; A: in TMatrix3; U: in TVector3; V: in out TVector3) is
    --- V := AU
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when 1 => Prod(R,Deg,A(1,1),U(1),O(1));
          when 2 => Prod(R,Deg,A(1,2),U(2),O(2));
          when 3 => Prod(R,Deg,A(1,3),U(3),O(3));
          when 4 => Prod(R,Deg,A(2,1),U(1),O(4));
          when 5 => Prod(R,Deg,A(2,2),U(2),O(5));
          when 6 => Prod(R,Deg,A(2,3),U(3),O(6));
          when 7 => Prod(R,Deg,A(3,1),U(1),O(7));
          when 8 => Prod(R,Deg,A(3,2),U(2),O(8));
          when 9 => Prod(R,Deg,A(3,3),U(3),O(9));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(8,Allow_Parallel);
    S1: Scalar renames V(1)(Deg);
    S2: Scalar renames V(2)(Deg);
    S3: Scalar renames V(3)(Deg);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "MV_Prod Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Sum(O(1)(Deg),O(2)(Deg),S1); Add(O(3)(Deg),S1);
    Sum(O(4)(Deg),O(5)(Deg),S2); Add(O(6)(Deg),S2);
    Sum(O(7)(Deg),O(8)(Deg),S3); Add(O(9)(Deg),S3);
  end MV_Prod;

  procedure Sandwich(R: in Radius; Deg: in Power; O: in out Data_Sandwich; U: in TVector3; A: in out TMatrix3) is
    --- AX = U x X x U
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when 1 => Sqr(R,Deg,U(1),O(1));
          when 2 => Sqr(R,Deg,U(2),O(2));
          when 3 => Sqr(R,Deg,U(3),O(3));
          when 4 => Prod(R,Deg,U(1),U(2),O(4));
          when 5 => Prod(R,Deg,U(3),U(1),O(5));
          when 6 => Prod(R,Deg,U(2),U(3),O(6));
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(5,Allow_Parallel);
    A12: Scalar renames A(1,2)(Deg);
    A31: Scalar renames A(3,1)(Deg);
    A23: Scalar renames A(2,3)(Deg);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Sandwich Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    -- diag
    Sum(O(1)(Deg),O(2)(Deg),A(3,3)(Deg));
    Sum(O(2)(Deg),O(3)(Deg),A(1,1)(Deg));
    Sum(O(3)(Deg),O(1)(Deg),A(2,2)(Deg));
    -- off-diag
    Neg(O(4)(Deg),A12);
    Copy(A12,A(2,1)(Deg));
    Neg(O(5)(Deg),A31);
    Copy(A31,A(1,3)(Deg));
    Neg(O(6)(Deg),A23);
    Copy(A23,A(3,2)(Deg));
  end Sandwich;

  procedure Symm_Det(R: in Radius; Deg: in Power; O: in out Data_Det; A,B: in TMatrix3; P: in out Taylor) is
    --- for symmetric matrix A and B=adj(A)
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        exit when J>3;
        Prod(R,Deg,A(1,J),B(J,1),O(J));
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(2,Allow_Parallel);
    S: Scalar renames P(Deg);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Symm_Det Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Sum(O(1)(Deg),O(2)(Deg),S);
    Add(O(3)(Deg),S);
  end Symm_Det;

  procedure Symm_Adj(R: in Radius; Deg: in Power; O: in out Data_Adj; A: in TMatrix3; B: in out TMatrix3) is
    --- for symmetric matrices
    PC: Protected_Counter(1);
    procedure Prod_Loop is
      J: Positive;
    begin
      loop
        PC.Next(J);
        case J is
          when  1 => Prod(R,Deg,A(2,2),A(3,3),O( 1)); --  b11 = a22*a33 - a23*a32
          when  2 =>         Sqr(R,Deg,A(3,2),O( 2)); --  b11 = a22*a33 - a23*a32
          when  3 => Prod(R,Deg,A(3,2),A(1,3),O( 3)); --  b12 = a13*a32 - a12*a33
          when  4 => Prod(R,Deg,A(1,2),A(3,3),O( 4)); --  b12 = a13*a32 - a12*a33
          when  5 => Prod(R,Deg,A(1,2),A(2,3),O( 5)); --  b13 = a12*a23 - a13*a22
          when  6 => Prod(R,Deg,A(1,3),A(2,2),O( 6)); --  b13 = a12*a23 - a13*a22
          when  7 => Prod(R,Deg,A(1,1),A(3,3),O( 7)); --  b22 = a11*a33 - a13*a31
          when  8 =>         Sqr(R,Deg,A(3,1),O( 8)); --  b22 = a11*a33 - a13*a31
          when  9 => Prod(R,Deg,A(2,1),A(1,3),O( 9)); --  b23 = a13*a21 - a11*a23
          when 10 => Prod(R,Deg,A(1,1),A(2,3),O(10)); --  b23 = a13*a21 - a11*a23
          when 11 => Prod(R,Deg,A(1,1),A(2,2),O(11)); --  b33 = a11*a22 - a12*a21
          when 12 =>         Sqr(R,Deg,A(1,2),O(12)); --  b33 = a11*a22 - a12*a21
          when others => exit;
        end case;
      end loop;
    end Prod_Loop;
    NT: constant Natural := Reserve_Tasks(11,Allow_Parallel);
  begin
    if NT=0 then
      Prod_Loop;
    else
      declare
        Task_Error: Boolean := False;
        task type Prod_Task_Type is end Prod_Task_Type;
        task body Prod_Task_Type is
        begin
          Proper_Rounding;
          Prod_Loop;
        exception
          when others => Task_Error := True; raise;
        end Prod_Task_Type;
        ---
        Prod_Task: array(1 .. NT) of Prod_Task_Type;
        pragma Warnings (Off,Prod_Task);
      begin
        Prod_Loop;
        if Task_Error then raise Task_Exception with "Symm_Adj Task_Error"; end if;
      end;
      Free_Tasks(NT);
    end if;
    Diff(O( 1)(Deg),O( 2)(Deg),B(1,1)(Deg));  -- b11 = a22*a33 - a23*a32
    Diff(O( 3)(Deg),O( 4)(Deg),B(1,2)(Deg));  -- b12 = a13*a32 - a12*a33
    Diff(O( 5)(Deg),O( 6)(Deg),B(1,3)(Deg));  -- b13 = a12*a23 - a13*a22
    Diff(O( 7)(Deg),O( 8)(Deg),B(2,2)(Deg));  -- b22 = a11*a33 - a13*a31
    Diff(O( 9)(Deg),O(10)(Deg),B(2,3)(Deg));  -- b23 = a13*a21 - a11*a23
    Diff(O(11)(Deg),O(12)(Deg),B(3,3)(Deg));  -- b33 = a11*a22 - a12*a21
    Copy(B(1,2)(Deg),B(2,1)(Deg));
    Copy(B(1,3)(Deg),B(3,1)(Deg));
    Copy(B(2,3)(Deg),B(3,2)(Deg));
  end Symm_Adj;

  procedure Copy(O1: in Data_M_Inv; O2: in out Data_M_Inv) is
  begin
    Copy(O1.Det,O2.Det);
    Copy(O1.InvDet,O2.InvDet);
    Copy(O1.Adj,O2.Adj);
    Copy(O1.Det1,O2.Det1);
    Copy(O1.Inv1,O2.Inv1);
    Copy(O1.Adj1,O2.Adj1);
  end Copy;

  procedure Symm_M_Inv(R: in Radius; Deg: in Power; O: in out Data_M_Inv; A: in TMatrix3; B: in out TMatrix3) is
    DetA: Taylor renames O.Det;
    InvDetA: Taylor renames O.InvDet;
    AdjA: TMatrix3 renames O.Adj;
  begin
    Symm_Adj(R,Deg,O.Adj1, A,AdjA);         -- adj(A)
    Symm_Det(R,Deg,O.Det1, A,AdjA,DetA);    -- det(A)
    Inv(R,Deg,O.Inv1,      DetA,InvDetA);   -- 1/det(A)
    Symm_TM_Prod(R,Deg,    InvDetA,AdjA,B); -- B := adj(A)/det(A)
  end Symm_M_Inv;

  procedure Copy(O1: in Data_VecField; O2: in out Data_VecField) is
  begin
    Copy(O1.SS,O2.SS);
    Copy(O1.InvSS,O2.InvSS);
    Copy(O1.InvS,O2.InvS);
    Copy(O1.SDotS,O2.SDotS);
    Copy(O1.InvSDotS,O2.InvSDotS);
    Copy(O1.BBGamma,O2.BBGamma);
    Copy(O1.BBDotGamma,O2.BBDotGamma);
    Copy(O1.SDotR,O2.SDotR);
    Copy(O1.Omega,O2.Omega);
    Copy(O1.VTmp1,O2.VTmp1);
    Copy(O1.VTmp2,O2.VTmp2);
    Copy(O1.VTmp3,O2.VTmp3);
    Copy(O1.VTmp4,O2.VTmp4);
    Copy(O1.DotM1,O2.DotM1);
    Copy(O1.DotM2,O2.DotM2);
    Copy(O1.DotM3,O2.DotM3);
    Copy(O1.SSK,O2.SSK);
    Copy(O1.K,O2.K);
    Copy(O1.ImK,O2.ImK);
    Copy(O1.InvImK,O2.InvImK);
    Copy(O1.Inv1,O2.Inv1);
    Copy(O1.Sqrt1,O2.Sqrt1);
    Copy(O1.Dot1,O2.Dot1);
    Copy(O1.Dot2,O2.Dot2);
    Copy(O1.Cross1,O2.Cross1);
    Copy(O1.Cross2,O2.Cross2);
    Copy(O1.Cross3,O2.Cross3);
    Copy(O1.Cross4,O2.Cross4);
    Copy(O1.Cross5,O2.Cross5);
    Copy(O1.Cross6,O2.Cross6);
    Copy(O1.Cross7,O2.Cross7);
    Copy(O1.Inv2,O2.Inv2);
    Copy(O1.Prod1,O2.Prod1);
    Copy(O1.Sand1,O2.Sand1);
  end Copy;

  procedure VecField(R: in Radius; Deg: in Power; O: in out Data_VecField; X: in Curve; DotX: in out Curve) is
    Task_Error: Boolean := False;
    NT: Natural;
    ---
    Alpha: TVector3 renames X.Alpha;
    Beta: TVector3 renames X.Beta;
    Gamma: TVector3 renames X.Gamma;
    M: TVector3 renames X.M;
    DotAlpha: TVector3 renames DotX.Alpha;
    DotBeta: TVector3 renames DotX.Beta;
    DotGamma: TVector3 renames DotX.Gamma;
    DotM: TVector3 renames DotX.M;
    Omega: TVector3 renames O.Omega;
  begin
    Diag_Prod(Deg,BB,Gamma,O.BBGamma);               -- BB*Gamma
    Dot(R,Deg,O.Dot1,O.BBGamma,Gamma,O.SS);          -- S^2
    Inv(R,Deg,O.Inv1,O.SS,O.InvSS);                  -- S^{-2}
    ---------------------------------------------------
    Sandwich(R,Deg,O.Sand1,O.BBGamma,O.SSK);
    Symm_TM_Prod(R,Deg,O.InvSS,O.SSK,O.K);           -- K
    SM_AddProd(Deg,Inertia,Mass,O.K,O.ImK);          -- I+mK
    Symm_M_Inv(R,Deg,O.Inv2,O.ImK,O.InvImK);         -- 1/(I+mK)
                                                     ---
    MV_Prod(R,Deg,O.Prod1,O.InvImK,M,Omega);         -- Omega
    if Compute_Alpha_And_Beta then
      Cross3(R,Deg,O.Cross1,O.Cross2,O.Cross3,Alpha,Beta,Gamma,Omega,Omega,Omega,DotAlpha,DotBeta,DotGamma); -- DotAlpha, DotBeta, DotGamma
    elsif Compute_Alpha then
      Cross2(R,Deg,O.Cross1,O.Cross3,Alpha,Gamma,Omega,Omega,DotAlpha,DotGamma); -- DotAlpha, DotGamma
    elsif Compute_Beta then
      Cross2(R,Deg,O.Cross2,O.Cross3,Beta,Gamma,Omega,Omega,DotBeta,DotGamma);   -- DotBeta, DotGamma
    else
      Cross1(R,Deg,O.Cross3,Gamma,Omega,DotGamma);    -- DotGamma
    end if;
    ----------------------------------------------------
    Dot(R,Deg,O.Dot2,O.BBGamma,DotGamma,O.SDotS);     -- SDotS
    ---- begin parallel --------------------------------
    NT := Reserve_Tasks(1,Allow_Parallel);
    if NT=0 then
      Prod(R,Deg,O.InvSS,O.SDotS,O.InvSDotS);         -- DotS/S
      Diag_Prod(Deg,BB,DotGamma,O.BBDotGamma);        -- BB*DotGamma
    else
      declare
        task Task1;
        task body Task1 is begin
          Proper_Rounding;
          Prod(R,Deg,O.InvSS,O.SDotS,O.InvSDotS);
        exception
          when others => Task_Error := True; raise;
        end Task1;
      begin
        Diag_Prod(Deg,BB,DotGamma,O.BBDotGamma);
        if Task_Error then raise Task_Exception with "Task1 Error"; end if;
        Free_Tasks(NT);
      end;
    end if;
    ------ end parallel ---------------------------------
    TV_Prod(R,Deg,O.InvSDotS,O.BBGamma,O.VTmp1);       -- (DotS/S)BB*Gamma
    Diff(Deg,O.VTmp1,O.BBDotGamma,O.SDotR);            -- SDotR := (DotS/S)BB*Gamma-BB*DotGamma
    ---- begin parallel --------------------------------
    NT := Reserve_Tasks(1,Allow_Parallel);
    if NT=0 then
      Sqrt(R,Deg,O.Sqrt1,O.InvSS,O.InvS);              -- 1/S
      Cross1(R,Deg,O.Cross4,Gamma,O.BBGamma,O.VTmp2);  -- Gamma x BBGamma
    else
      declare
        task Task2;
        task body Task2 is begin
          Proper_Rounding;
          Sqrt(R,Deg,O.Sqrt1,O.InvSS,O.InvS);
        exception
          when others => Task_Error := True; raise;
        end Task2;
      begin
        Cross1(R,Deg,O.Cross4,Gamma,O.BBGamma,O.VTmp2);
        if Task_Error then raise Task_Exception with "Task2 Error"; end if;
        Free_Tasks(NT);
      end;
    end if;
    ------ end parallel -------------------------------
    TV_Prod(R,Deg,O.InvS,O.VTmp2,O.DotM1);           -- DotM1 := S^{-1}*(Gamma x BBGamma)
    ---------------------------------------------------
    Cross2(R,Deg,O.Cross5,O.Cross6,M,Omega,Omega,O.BBGamma,O.DotM2,O.VTmp3); -- DotM2 := M x Omega  VTmp3 := Omega x BBGamma
    Cross1(R,Deg,O.Cross7,O.VTmp3,O.SDotR,O.VTmp4); -- (Omega x BBGamma) x SDotR
    TV_Prod(R,Deg,O.InvSS,O.VTmp4,O.DotM3);         -- DotM3 := S^{-2}*[(Omega x BBGamma) x SDotR]
    --------------------------------------------------
    Copy(Deg,O.DotM2,DotM);
    AddProd(Deg,GForce,O.DotM1,DotM);
    AddProd(Deg,Mass,O.DotM3,DotM);                 -- DotM := GForce*DotM1+DotM2+Mass*DotM3
  end VecField;

  procedure Low_AntiDer(Deg: in Power; DotU: in TVector3; U0: in SVector3; U: in out TVector3) is
  begin
    Low_Antider(Deg,DotU(1),U0(1),U(1));
    Low_Antider(Deg,DotU(2),U0(2),U(2));
    Low_Antider(Deg,DotU(3),U0(3),U(3));
  end Low_Antider;

  procedure Low_AntiDer(Deg: in Power; DotX: in Curve; X0: in Point; X: in out Curve) is
  begin
    if Compute_Alpha then
      Low_Antider(Deg,DotX.Alpha,X0.Alpha,X.Alpha);
    end if;
    if Compute_Beta then
      Low_Antider(Deg,DotX.Beta,X0.Beta,X.Beta);
    end if;
    Low_Antider(Deg,DotX.Gamma,X0.Gamma,X.Gamma);
    Low_Antider(Deg,DotX.M,X0.M,X.M);
  end Low_Antider;

  procedure Hi_AntiDer(R: in Radius; DotU: in TVector3; U: in out TVector3; Tmp: in out SVector3; Contr: in out Boolean) is
    Tmp1: Scalar renames Tmp(1);
    Tmp2: Scalar renames Tmp(2);
    Tmp3: Scalar renames Tmp(3);
  begin
    Hi_Antider(R,DotU(1),U(1),Tmp1,Tmp2,Tmp3,Contr);
    Hi_Antider(R,DotU(2),U(2),Tmp1,Tmp2,Tmp3,Contr);
    Hi_Antider(R,DotU(3),U(3),Tmp1,Tmp2,Tmp3,Contr);
  end Hi_Antider;

  procedure Hi_AntiDer(R: in Radius; DotX: in Curve; X: in out Curve; Tmp: in out SVector3; Contr: in out Boolean) is
  begin
    if Rigorous_Alpha then
      Hi_Antider(R,DotX.Alpha,X.Alpha,Tmp,Contr);
    end if;
    if Rigorous_Beta then
      Hi_Antider(R,DotX.Beta,X.Beta,Tmp,Contr);
    end if;
    Hi_Antider(R,DotX.Gamma,X.Gamma,Tmp,Contr);
    Hi_Antider(R,DotX.M,X.M,Tmp,Contr);
  end Hi_Antider;

  procedure Hi_ErrMult(Fac: in Radius; U: in out TVector3) is
  begin
    ErrMult(Fac,U(1)(PDeg));
    ErrMult(Fac,U(2)(PDeg));
    ErrMult(Fac,U(3)(PDeg));
  end Hi_ErrMult;

  procedure Hi_ErrMult(Fac: in Radius; X: in out Curve) is
  begin
    if Rigorous_Alpha then
      Hi_ErrMult(Fac,X.Alpha);
    end if;
    if Rigorous_Beta then
      Hi_ErrMult(Fac,X.Beta);
    end if;
    Hi_ErrMult(Fac,X.Gamma);
    Hi_ErrMult(Fac,X.M);
  end Hi_ErrMult;

  function MaxCoeffPDegMinus1(C: Curve) return Radius is
    R: Flt := Zero;
  begin
    if Rigorous_Alpha then
      MaxCoeffPDegMinus1(C.Alpha,R);
    end if;
    if Rigorous_Beta then
      MaxCoeffPDegMinus1(C.Beta,R);
    end if;
    MaxCoeffPDegMinus1(C.Gamma,R);
    MaxCoeffPDegMinus1(C.M,R);
    return R;
  end MaxCoeffPDegMinus1;

  procedure Show_Stat(D: in RData) is
    Count: Counter renames D.Count;
  begin
    Show1("Prec = ",Prec);
    Show1("MinRad =",MinRad);
    for N in 1 .. MaxRadFac loop
      Show0(Strng(Count(N),6) & " uses of MinRad *" & Strng(N,3));
    end loop;
    Show0("possibly shorter:" & Strng(Count(0),6) & " times");
    Show0("possibly  longer:" & Strng(Count(MaxRadFacPlus1),6) & " times");
  end Show_Stat;

  procedure Update(C: in Curve; D: in out RData) is
    use Flt_EF;
    Log2: constant Flt := 0.69314718056;  --- just for guessing
    Adjust: constant Flt := One/Flt(8);   --- adjust Guess
    R: constant Radius := Quarter+MaxCoeffPDegMinus1(C);
    Rad: Radius renames D.Rad;
    Count: Counter renames D.Count;
    N: Integer;
  begin
    Rad := Adjust*Exp((Log(R)+Flt(Prec)*Log2)/Flt(-PDegMinus1)); -- first guess
    N := Integer(Flt'Floor(Rad/MinRad));
    if Verbosity>2 then
      Show1("NewRad guess factor N =",N);
    end if;
    if N>MaxRadFac then
      Rad := Flt(MaxRadFac)*MinRad;
      Count(MaxRadFacPlus1) := Count(MaxRadFacPlus1)+1;
    elsif N>0 then
      Rad := Flt(N)*MinRad;
      Count(N) := Count(N)+1;
    else
      Rad := MinRad;
      Count(1) := Count(1)+1;
    end if;
  end Update;

  procedure Integrate(X0: in Point; X,DotX: in out Curve; VTmp: in out SVector3; D: in out RData) is
    Iter: constant Positive := 5;
    Fac: constant Radius := Four; -- for contraction
    Rad: Radius renames D.Rad;
    Contr: Boolean;
    OTmp: XT_Pointer;
  begin
    Rad := MinRad;
    XPool.Allocate(OTmp);
    declare
      O: Data_VecField renames OTmp.all.Data;
    begin
      for Deg in 0 .. PDeg loop
        Low_AntiDer(Deg,DotX,X0,X);          -- computes X at Deg
        VecField(Rad,Deg,O,X,DotX);          -- computes DotX at Deg
        if Deg=PDegMinus1 then
          Update(DotX,D);                    -- determine new Rad
        end if;
      end loop;
      if Not_Strunc then
        for I in 1 .. Iter loop
          Contr := True;
          Hi_AntiDer(Rad,DotX,X,VTmp,Contr); -- adds h.o. errors
          if not Contr then
            Hi_ErrMult(Fac,X);
          end if;
          VecField(Rad,PDeg,O,X,DotX);       -- one more time after Contr
          exit when Contr;
        end loop;
        if not Contr then raise Sorry; end if;
      end if;
    end;
    XPool.Recycle(OTmp);
  end Integrate;

  ------------------- plotting

  procedure Prepare_Plot(T1,T2: in Flt; FileName: in String) is
    use ASU;
  begin
    if Plot_Filename=Null_Unbounded_String then
      Plot_TMin := RMin(T1,T2);
      PLot_Tmax := RMax(T1,T2);
      Plot_Filename := To_Unbounded_String(FileName);
    else
      raise Sorry with "non-empty Plot_Filenam: " & To_String(Plot_Filename);
    end if;
  end Prepare_Plot;

  procedure Init_Plot is
    use ASU;
  begin
    if Plot_Filename/=Null_Unbounded_String then
      declare
        FileName: constant String := To_String(Plot_Filename);
      begin
        Show0("plotting to " & FileName);
        Create(Plot_Filetype,Out_File,FileName);
      end;
    end if;
  end Init_Plot;

  procedure Plot(T: in Flt; P: in Point) is
    use ASU;
  begin
    if Plot_Filename=Null_Unbounded_String then
      null;
    elsif T>=Plot_TMin and then T<=Plot_TMax then
      Plot(Plot_Filetype,T,P);
    end if;
  end Plot;

  procedure Stop_Plot is
    use ASU;
  begin
    if Plot_Filename/=Null_Unbounded_String then
      Close(Plot_Filetype);
      Plot_Filename := Null_Unbounded_String;
    end if;
  end Stop_Plot;

  -----------------------

  procedure Show_Info(T: in Flt; P: in Point) is
  begin
    Show1("T =",T);
    if Verbosity>2 then
      Show_ABG_Errs(P);
      if Verbosity>3 then
        Show_Approx_Energies(P);
        if Verbosity>4 then
          ShowApprox1("Gamma",P.Gamma);
          ShowApprox1("    M",P.M);
        end if;
      end if;
    end if;
  end Show_Info;

  procedure YawNumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NY: in out Integer) is
    --- increments NY, uses YawSteps
    Alpha1: Taylor renames X.Alpha(1);
    Beta1:  Taylor renames X.Beta(1);
    DotAlpha1: Taylor renames DotX.Alpha(1);
    DotBeta1:   Taylor renames DotX.Beta(1);
    A1,B1,DotA1,DotB1: Scalar;
    ---
    procedure YawModPi(S: in Scalar; Psi: in out Scalar) is
    begin
      if IsZero(A1) and then IsZero(B1) then
        Val(R,DotAlpha1,S,DotA1);         -- DotAlpha1(S)
        Val(R,DotBeta1,S,DotB1);          -- DotBeta1(S)
        YawModPi(DotA1,DotB1,Psi);        -- Psi(S)
      else
        YawModPi(A1,B1,Psi);
      end if;
    end YawModPi;
    ---
    OK: Boolean;
    T0,T1,Psi0,Psi1,S: Scalar;
  begin
    SetZero(T1);
    Val(R,Alpha1,T1,A1);           -- Alpha1(0)
    Val(R,Beta1,T1,B1);            -- Beta1(0)
    YawModPi(T1,Psi1);             -- Psi(0)
    for K in 1 .. YawSteps loop
      Copy(T1,T0);
      Copy(Psi1,Psi0);             -- Psi(T0)
      Prod(Flt(K)/Flt(YawSteps),T,T1);
      Val(R,Alpha1,T1,A1);         -- Alpha1(T1)
      Val(R,Beta1,T1,B1);          -- Beta1(T1)
      YawModPi(T1,Psi1);           -- Psi(T1)
      Copy(T1,S);
      Union(T0,S);                 -- S=[T0,T1]
      Val(R,Alpha1,S,A1);          -- Alpha1(S)
      Val(R,Beta1,S,B1);           -- Beta1(S)
      Val(R,DotAlpha1,S,DotA1);    -- DotAlpha1(S)
      Val(R,DotBeta1,S,DotB1);     -- DotBeta1(S)
      Diff(T1,T0,S);
      YawNumPi(S,Psi0,Psi1,A1,B1,DotA1,DotB1,NY,OK);
      if not OK then
        Show0("problem in YawNumPi");
        raise Sorry;
      end if;
    end loop;
    ShowYaw(Psi1,NY); -- no point in computing but not showing
  end YawNumPi;

  procedure RollNumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NR: in out Integer) is
    --- increments NR, uses RollSteps
    Gamma2: Taylor renames X.Gamma(2);
    Gamma3: Taylor renames X.Gamma(3);
    DotGamma2: Taylor renames DotX.Gamma(2);
    DotGamma3: Taylor renames DotX.Gamma(3);
    G2,G3,DotG2,DotG3: Scalar;
    ---
    procedure RollModPi(S: in Scalar; Phi: in out Scalar) is
    begin
      if IsZero(G2) and then IsZero(G3) then
        Val(R,DotGamma2,S,DotG2);        -- DotGamma2(S)
        Val(R,DotGamma3,S,DotG3);        -- DotGamma3(S)
        RollModPi(DotG2,DotG3,Phi);      -- Phi(S)
      else
        RollModPi(G2,G3,Phi);
      end if;
    end RollModPi;
    ---
    OK: Boolean;
    T0,T1,Phi0,Phi1,S: Scalar;
  begin
    SetZero(T1);
    Val(R,Gamma2,T1,G2);           -- Gamma2(0)
    Val(R,Gamma3,T1,G3);           -- Gamma3(0)
    RollModPi(T1,Phi1);            -- Phi(0)
    for K in 1 .. RollSteps loop
      Copy(T1,T0);
      Copy(Phi1,Phi0);             -- Phi(T0)
      Prod(Flt(K)/Flt(RollSteps),T,T1);
      Val(R,Gamma2,T1,G2);         -- Gamma2(T1)
      Val(R,Gamma3,T1,G3);         -- Gamma3(T1)
      RollModPi(T1,Phi1);          -- Phi(T1)
      Copy(T1,S);
      Union(T0,S);                 -- S=[T0,T1]
      Val(R,Gamma2,S,G2);          -- Gamma2(S)
      Val(R,Gamma3,S,G3);          -- Gamma(S)
      Val(R,DotGamma2,S,DotG2);    -- DotGamma2(S)
      Val(R,DotGamma3,S,DotG3);    -- DotGamma3(S)
      Diff(T1,T0,S);
      RollNumPi(S,Phi0,Phi1,G2,G3,DotG2,DotG3,NR,OK);
      if not OK then
        Show0("Problem with RollNumPi");
        raise Sorry;
      end if;
    end loop;
    ShowRoll(Phi1,NR); -- no point in computing but not showing
  end RollNumPi;

  procedure NumPi(R: in Radius; T: in Scalar; X,DotX: in Curve; NY,NR: in out Integer) is
    NT: Natural := 0;
  begin
    if Compute_Yaw_And_Roll then
      NT := Reserve_Tasks(1,Allow_Parallel);
    end if;
    if NT=0 then
      if Compute_Yaw then YawNumPi(R,T,X,DotX,NY); end if;
      if Compute_Roll then RollNumPi(R,T,X,DotX,NR); end if;
    else
      declare
        Task_Error: Boolean := False;
        task Task3;
        task body Task3 is begin
          Proper_Rounding;
          YawNumPi(R,T,X,DotX,NY);
        exception
          when others => Task_Error := True; raise;
        end Task3;
      begin
        RollNumPi(R,T,X,DotX,NR);
        if Task_Error then raise Task_Exception with "Task3 Error"; end if;
        Free_Tasks(NT);
      end;
    end if;
  end NumPi;

  procedure Generic_Flow(T,DT: out Flt; XT: in out Point; X,DotX: in out Curve; Forward: in Boolean := True) is
    --- on input: XT is the initial condition
    --- on output: Done(T,DT,X)=True, X is centered at T, XT=X(T)
    NY: Integer renames XT.YawPi;
    NR: Integer renames XT.RollPi;
    D: RData;
    Rad: Radius renames D.Rad;
    DS: Scalar;
    VTmp: SVector3;
  begin
    Init_Plot;
    T := Zero;
    if Compute_Yaw then ShowYaw(XT); end if;
    if Compute_Roll then ShowRoll(XT); end if;
    for Steps in 1 .. Integer'Last loop
      Plot(T,XT);
      if Verbosity>0 then Show_Info(T,XT); end if;
      Integrate(XT,X,DotX,VTmp,D);
      if Forward then DT := Rad; else DT := -Rad; end if;
      if STrunc and then (Steps mod 32)=0 then NumFix(XT); end if;
      exit when Done(T,DT,X);
      T := T+DT;
      Assign(DT,DS);
      Val(Rad,X,DS,XT);  --- XT.Energy, XT.YawPi, XT.RollPi remain unchanged
      if Compute_Yaw_Or_Roll then
        NumPi(Rad,DS,X,DotX,NY,NR);
      end if;
    end loop;
    Stop_Plot;
    if Verbosity>0 then Show_Stat(D); end if;
  end Generic_Flow;

  procedure Plain_Flow(TNeed: in Flt; T,DT: out Flt; XT: in out Point; X,DotX: in out Curve) is
    Forward: constant Boolean := TNeed>Zero;
    function Done(Tau,DTau: in Flt; C: in Curve) return Boolean is
      pragma Unreferenced(DTau,C);
    begin
      return Tau=TNeed or else (Tau>TNeed)=Forward;
    end Done;
    procedure New_Flow is new Generic_Flow (Done => Done);
  begin
    New_Flow(T,DT,XT,X,DotX,Forward);
  end Plain_Flow;

  procedure TimeT_Map(TStop: in Scalar; X0: in Point; XT: in out Point) is
    TNeed: constant Flt := Approx(TStop);
    Forward: constant Boolean := TNeed>Zero;
    function Done(Tau,DTau: in Flt; C: in Curve) return Boolean is
      pragma Unreferenced(DTau,C);
    begin
      return Tau=TNeed or else (Tau>TNeed)=Forward;
    end Done;
    procedure New_Flow is new Generic_Flow (Done => Done);
    T,DT: Flt;
    Rad: Radius;
    S,Tmp: Scalar;
    X,DotX: Curve;
  begin
    Copy(X0,XT);
    New_Flow(T,DT,XT,X,DotX,Forward);
    Rad := abs(DT);
    Assign(T,Tmp);
    Diff(TStop,Tmp,S);
    Val(Rad,X,S,XT);  --- XT.Energy, XT.YawPi, XT.RollPi remain unchanged
    if Compute_Yaw_Or_Roll then
      NumPi(Rad,S,X,DotX,XT.YawPi,XT.RollPi);
    end if;
  end TimeT_Map;

  ----------------------------------

  procedure Val(F: in Flt_Affine; V: in TVector3; S: in out Taylor) is
    Tmp: Scalar;
  begin
    Assign(F(0),Tmp); Assign(Tmp,S);
    Assign(F(1),Tmp); AddProd(Tmp,V(1),S);
    Assign(F(2),Tmp); AddProd(Tmp,V(2),S);
    Assign(F(3),Tmp); AddProd(Tmp,V(3),S);
  end Val;

  procedure ShowTime(T0: in Flt; DT: in Scalar; Tmp: in out Scalar) is
  begin
    Assign(T0,Tmp);
    Add(DT,Tmp);
    Reduce_Precision(64,Tmp);
    Show1("Time ",Tmp);
  end ShowTime;

  procedure Plain_Poincare(T1,T2: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    --- F defines the Poincare section - find zero rougly between times T1 and T2
    TNeed,T,DT,Rad: Flt;
    S: Scalar;
    FM: Taylor;
    Tmp: Scalar renames FM(0);
    X,DotX: Curve;
  begin
    if abs(T1)<abs(T2) then TNeed := T1; else TNeed := T2; end if;
    Copy(X0,XT);
    Plain_Flow(TNeed,T,DT,XT,X,DotX);
    Rad := abs(DT);
    Write(Output_Dir & "temp",XT);
    Val(F,X.M,FM);
    FindZero(Rad,FM,T1-T,T2-T,S);
    Val(Rad,X,S,XT);  --- XT.Energy, XT.YawPi, XT.RollPi remain unchanged
    if Compute_Yaw_Or_Roll then
      NumPi(Rad,S,X,DotX,XT.YawPi,XT.RollPi);
    end if;
    if Verbosity>0 then
      ShowTime(T,S,Tmp);
    end if;
  end Plain_Poincare;

  function Sign_Change(DT: in Flt; F: in Flt_Affine; M: in TVector3; Tmp: in out Taylor) return Boolean is
    Rad: constant Radius := abs(DT);
    FM: Taylor renames Tmp;
    Sgn: Integer;
    DS,S: Scalar;
  begin
    Val(F,M,FM);
    SetZero(DS);
    Val(Rad,FM,DS,S);
    Sgn := Sign(S);
    Assign(DT,DS);
    Val(Rad,FM,DS,S);
    return Sign(S)/=Sgn;
  end Sign_Change;

  procedure Sign_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    Forward: constant Boolean := TNeed>Zero;
    Tmp: Taylor;
    function Done(Tau,DTau: in Flt; C: in Curve) return Boolean is
    begin
      if Tau=TNeed or else (Tau>TNeed)=Forward then
        return Sign_Change(DTau,F,C.M,Tmp);
      end if;
      return False;
    end Done;
    procedure New_Flow is new Generic_Flow (Done => Done);
    T,DT,Rad: Flt;
    FM: Taylor renames Tmp;
    S: Scalar;
    X,DotX: Curve;
  begin
    Copy(X0,XT);
    New_Flow(T,DT,XT,X,DotX,Forward);
    Rad := abs(DT);
    Val(F,X.M,FM);
    FindZero(Rad,FM,Zero,DT,S);
    Val(Rad,X,S,XT);  --- XT.Energy, XT.YawPi, XT.RollPi remain unchanged
    if Compute_Yaw_Or_Roll then
      NumPi(Rad,S,X,DotX,XT.YawPi,XT.RollPi);
    end if;
    if Verbosity>0 then
      ShowTime(T,S,Tmp(0));
    end if;
  end Sign_Poincare;

  function Neg2Pos(DT: in Flt; F: in Flt_Affine; M: in TVector3; Tmp: in out Taylor) return Boolean is
    Rad: constant Radius := abs(DT);
    FM: Taylor renames Tmp;
    Sgn: Integer;
    S: Scalar;
  begin
    Val(F,M,FM);
    Val(Rad,FM,SZero,S);
    Sgn := Sign(S);
    if Sgn>0 then return False; end if;
    declare
      DS: Scalar;
    begin
      Assign(DT,DS);
      Val(Rad,FM,DS,S);
      Sgn := Sign(S);
      return Sgn>0;
    end;
  end Neg2Pos;

  procedure Neg2Pos_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    Forward: constant Boolean := TNeed>Zero;
    Tmp: Taylor;
    function Done(Tau,DTau: in Flt; C: in Curve) return Boolean is
    begin
      if Tau=TNeed or else (Tau>TNeed)=Forward then
        return Neg2Pos(DTau,F,C.M,Tmp);
      end if;
      return False;
    end Done;
    procedure New_Flow is new Generic_Flow (Done => Done);
    T,DT,Rad: Flt;
    FM: Taylor renames Tmp;
    S: Scalar;
    X,DotX: Curve;
  begin
    Copy(X0,XT);
    New_Flow(T,DT,XT,X,DotX,Forward);
    Rad := abs(DT);
    Val(F,X.M,FM);
    FindZero(Rad,FM,Zero,DT,S);
    Val(Rad,X,S,XT);  --- XT.Energy, XT.YawPi, XT.RollPi remain unchanged
    if Compute_Yaw_Or_Roll then
      NumPi(Rad,S,X,DotX,XT.YawPi,XT.RollPi);
    end if;
    if Verbosity>0 then
      ShowTime(T,S,Tmp(0));
    end if;
  end Neg2Pos_Poincare;

  procedure Pos2Neg_Poincare(TNeed: in Flt; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    NegF: constant Flt_Affine := (-F(0),-F(1),-F(2),-F(3));
  begin
    Neg2Pos_Poincare(TNeed,NegF,X0,XT);
  end Pos2Neg_Poincare;

  procedure CountNeg2Pos_Poincare(Iter: in Integer; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    --- use Iter<0 for backward flow
    MaxCount: constant Positive := abs(Iter); -- so Iter/=0
    Forward: constant Boolean := Iter>0;
    Count: Integer := 0;
    Tmp: Taylor;
    function Done(Tau,DTau: in Flt; C: in Curve) return Boolean is
    begin
      if Tau=Zero then return False; end if;
      if Neg2Pos(DTau,F,C.M,Tmp) then
        Count := Count+1;
        if Verbosity>1 then
          Show1("Neg2Pos" & Strng(Count,3) & " at t =",Tau);
        end if;
        if Count=MaxCount then return True; end if;
      end if;
      return False;
    end Done;
    procedure New_Flow is new Generic_Flow (Done => Done);
    T,DT,Rad: Flt;
    FM: Taylor renames Tmp;
    E,S: Scalar;
    X,DotX: Curve;
  begin
    Copy(X0.Energy,E);
    Copy(X0,XT);
    New_Flow(T,DT,XT,X,DotX,Forward);
    Rad := abs(DT);
    Val(F,X.M,FM);
    FindZero(Rad,FM,Zero,DT,S);
    Val(Rad,X,S,XT);
    Copy(E,XT.Energy);
    if Verbosity>0 then
      ShowTime(T,S,Tmp(0));
    end if;
  end CountNeg2Pos_Poincare;

  procedure CountPos2Neg_Poincare(Iter: in Integer; F: in Flt_Affine; X0: in Point; XT: in out Point) is
    NegF: constant Flt_Affine := (-F(0),-F(1),-F(2),-F(3));
  begin
    CountNeg2Pos_Poincare(Iter,NegF,X0,XT);
  end CountPos2Neg_Poincare;

end Rattleback.Flows;
