with Ada.Numerics, Globals, Strings, ArcTangent;
use Globals, Strings;

pragma Elaborate_All (Ada.Numerics,Globals,Strings,ArcTangent);

package body Rattleback is

  package ATan is new ArcTangent(Scalar => Scalar);
  use ATan;
  InfPiHalf: constant Flt := Inf(PiHalf);
  PiFull: constant Scalar := Two*PiHalf;

  procedure SetZero(U: in out SVector3) is
  begin
    SetZero(U(1));
    SetZero(U(2));
    SetZero(U(3));
  end SetZero;

  procedure Reduce_Precision(Binary_Digits: in Positive; U: in out SVector3) is
  begin
    Reduce_Precision(Binary_Digits,U(1));
    Reduce_Precision(Binary_Digits,U(2));
    Reduce_Precision(Binary_Digits,U(3));
  end Reduce_Precision;

  procedure Assign(U1,U2,U3: in Flt; U: in out SVector3) is
  begin
    Assign(U1,U(1));
    Assign(U2,U(2));
    Assign(U3,U(3));
  end Assign;

  procedure ShowApprox1(N: in String; V: in SVector3) is
  begin
    Show1(N & "(1)",Approx(V(1)));
    Show1(N & "(2)",Approx(V(2)));
    Show1(N & "(3)",Approx(V(3)));
  end ShowApprox1;

  procedure ShowApprox1(N: in String; A: in SMatrix3) is
  begin
    Show1(N & "(1,1)",Approx(A(1,1)));
    Show1(N & "(1,2)",Approx(A(1,2)));
    Show1(N & "(1,3)",Approx(A(1,3)));
    Show1(N & "(2,1)",Approx(A(2,1)));
    Show1(N & "(2,2)",Approx(A(2,2)));
    Show1(N & "(2,3)",Approx(A(2,3)));
    Show1(N & "(3,1)",Approx(A(3,1)));
    Show1(N & "(3,2)",Approx(A(3,2)));
    Show1(N & "(3,3)",Approx(A(3,3)));
  end ShowApprox1;

  procedure ShowApprox1(N: in String; A: in SMatrix2) is
  begin
    Show1(N & "(1,1)",Approx(A(1,1)));
    Show1(N & "(1,2)",Approx(A(1,2)));
    Show1(N & "(2,1)",Approx(A(2,1)));
    Show1(N & "(2,2)",Approx(A(2,2)));
  end ShowApprox1;

  procedure Show1(N: in String; V: in SVector3) is
  begin
    Show1(N & "(1) ",V(1));
    Show1(N & "(2) ",V(2));
    Show1(N & "(3) ",V(3));
  end Show1;

  procedure Show1(N: in String; A: in SMatrix3) is
  begin
    Show1(N & "(1,1) ",A(1,1));
    Show1(N & "(1,2) ",A(1,2));
    Show1(N & "(1,3) ",A(1,3));
    Show1(N & "(2,1) ",A(2,1));
    Show1(N & "(2,2) ",A(2,2));
    Show1(N & "(2,3) ",A(2,3));
    Show1(N & "(3,1) ",A(3,1));
    Show1(N & "(3,2) ",A(3,2));
    Show1(N & "(3,3) ",A(3,3));
  end Show1;

  procedure Show1(N: in String; A: in SMatrix2) is
  begin
    Show1(N & "(1,1) ",A(1,1));
    Show1(N & "(1,2) ",A(1,2));
    Show1(N & "(2,1) ",A(2,1));
    Show1(N & "(2,2) ",A(2,2));
  end Show1;

  procedure Put(F: in File_Type; V: in SVector3; Decimal: in Boolean := False) is
  begin
    for I in OneTwoThree loop Put(F,V(I),Decimal); end loop;
  end Put;

  procedure Get(F: in File_Type; V: in out SVector3; Decimal: in Boolean := False) is
  begin
    for I in OneTwoThree loop Get(F,V(I),Decimal); end loop;
  end Get;

  procedure Put(F: in File_Type; A: in SMatrix2; Decimal: in Boolean := False) is
  begin
    for I in OneTwo loop
      for J in OneTwo loop Put(F,A(I,J),Decimal); end loop;
    end loop;
  end Put;

  procedure Get(F: in File_Type; A: in out SMatrix2; Decimal: in Boolean := False) is
  begin
    for I in OneTwo loop
      for J in OneTwo loop Get(F,A(I,J),Decimal); end loop;
    end loop;
  end Get;

  procedure Put(F: in File_Type; A: in SMatrix3; Decimal: in Boolean := False) is
  begin
    for I in OneTwoThree loop
      for J in OneTwoThree loop Put(F,A(I,J),Decimal); end loop;
    end loop;
  end Put;

  procedure Get(F: in File_Type; A: in out SMatrix3; Decimal: in Boolean := False) is
  begin
    for I in OneTwoThree loop
      for J in OneTwoThree loop Get(F,A(I,J),Decimal); end loop;
    end loop;
  end Get;

  procedure Write(FileName: in String; A: in SMatrix2; Decimal: in Boolean := False) is
    use Int_IO;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Int_IO.Put(F,2,2);
    New_Line(F);
    Int_IO.Put(F,2,2);
    New_Line(F);
    Put(F,A,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; A: in out SMatrix2; Decimal: in Boolean := False) is
    use Int_IO;
    N: Integer;
    pragma Warnings (Off,N);
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Int_IO.Get(F,N);
    Int_IO.Get(F,N);
    Get(F,A,Decimal);
    Close(F);
  end Read;

  procedure Write(FileName: in String; A: in SMatrix3; Decimal: in Boolean := False) is
    use Int_IO;
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Int_IO.Put(F,3,2);
    New_Line(F);
    Int_IO.Put(F,3,2);
    New_Line(F);
    Put(F,A,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; A: in out SMatrix3; Decimal: in Boolean := False) is
    use Int_IO;
    N: Integer;
    pragma Warnings (Off,N);
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Int_IO.Get(F,N);
    Int_IO.Get(F,N);
    Get(F,A,Decimal);
    Close(F);
  end Read;

  procedure Enlarge(S: in out Scalar) is
  begin
    if Not_STrunc then
      declare
        E: Scalar;
      begin
        DiskAt0(Flt'Model_small,E);
        Add(E,S);
      end;
    end if;
  end Enlarge;

  procedure Enlarge(V: in out SVector3) is
  begin
    if Not_STrunc then
      Enlarge(V(1));
      Enlarge(V(2));
      Enlarge(V(3));
    end if;
  end Enlarge;

  ------------

  procedure Put(F: in File_Type; P: in Point; Decimal: in Boolean := False) is
    use Mod_IO;
  begin
    Put(F,ABFlag);
    New_Line(F);
    if Compute_Alpha then
      Put(F,P.Alpha,Decimal);
    end if;
    if Compute_Beta then
      Put(F,P.Beta,Decimal);
    end if;
    Put(F,P.Gamma,Decimal);
    Put(F,P.M,Decimal);
    Put(F,P.Energy,Decimal);
  end Put;

  procedure Get(F: in File_Type; P: in out Point; Decimal: in Boolean := False) is
    use Mod_IO;
    A: SVector3 renames P.Alpha;
    B: SVector3 renames P.Beta;
    G: SVector3 renames P.Gamma;
    Got_A,Got_B: Boolean;
    ABF: Flag;
  begin
    Get(F,ABF);
    Got_A := (ABF and Alpha_Defined)=Alpha_Defined;
    Got_B := (ABF and Beta_Defined)=Beta_Defined;
    if Got_A then Get(F,A,Decimal); end if;
    if Got_B then Get(F,B,Decimal); end if;
    Get(F,G,Decimal);
    Get(F,P.M,Decimal);
    Get(F,P.Energy,Decimal);
    if Got_A then
      if not Got_B then MakeB(G,A,B); end if;
    elsif Got_B then
      MakeA(B,G,A);
    else
      Show0("Warning: making cheap Alpha and Beta");
      CheapAB(G,A,B);
    end if;
    P.YawPi := 0;
    P.RollPi := 0;
  end Get;

  procedure Write(FileName: in String; P: in Point; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Point; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  function Read_Flag(FileName: String) return Flag is
    use Mod_IO;
    F: File_Type;
    ABF: Flag;
  begin
    Open(F,In_File,FileName);
    Get(F,ABF);
    return ABF;
  end Read_Flag;

  ----------------------

  procedure Copy(P1: in Point; P2: in out Point) is
  begin
    Copy3(P1.Alpha,P2.Alpha);
    Copy3(P1.Beta,P2.Beta);
    Copy3(P1.Gamma,P2.Gamma);
    Copy3(P1.M,P2.M);
    Copy(P1.Energy,P2.Energy);
    P2.YawPi := P1.YawPi;
    P2.RollPi := P1.RollPi;
  end Copy;

  function MaxNormDiff(P,Q: Point) return Radius is
    PG: SVector3 renames P.Gamma;
    QG: SVector3 renames Q.Gamma;
    PM: SVector3 renames P.M;
    QM: SVector3 renames Q.M;
    R: Radius := Zero;
    Tmp: Scalar;
  begin
    for I in OneTwoThree loop
      Diff(PG(I),QG(I),Tmp);
      R := R+MaxNorm(Tmp);
      Diff(PM(I),QM(I),Tmp);
      R := R+MaxNorm(Tmp);
    end loop;
    return R;
  end MaxNormDiff;

  procedure ShowApprox1(N: in String; P: in Point) is
  begin
    if Compute_Alpha then
      ShowApprox1(N & " Alpha",P.Alpha);
    end if;
    if Compute_Beta then
      ShowApprox1(N & "  Beta",P.Beta);
    end if;
    ShowApprox1(N & " Gamma",P.Gamma);
    ShowApprox1(N & "     M",P.M);
    Show1(N & " Energy  ",Approx(P.Energy));
    if Compute_Yaw then
      Show1(N & " YawPi = ",P.YawPi);
    end if;
    if Compute_Roll then
      Show1(N & "RollPi = ",P.RollPi);
    end if;
  end ShowApprox1;

  procedure Show1(N: in String; P: in Point) is
  begin
    if Compute_Alpha then
      Show1(N & " Alpha",P.Alpha);
    end if;
    if Compute_Beta then
      Show1(N & "  Beta",P.Beta);
    end if;
    Show1(N & " Gamma",P.Gamma);
    Show1(N & "     M",P.M);
    Show1(N & "Energy  ",P.Energy);
    if Compute_Yaw then
      Show1(N & "YawPi = ",P.YawPi);
    end if;
    if Compute_Roll then
      Show1(N & "RollPi = ",P.RollPi);
    end if;
  end Show1;

  procedure Plot1(F: in File_Type; R: in Flt; Valid: in Boolean := True) is
    use Flt_IO;
    Plot_Fore: constant Field := 3;
    Plot_Aft:  constant Field := 5;
    Plot_Exp:  constant Field := 3;
  begin
    if Valid then
      Put(F,R,Plot_Fore,Plot_Aft,Plot_Exp);
    else
      Put(F," 0.0");
    end if;
  end Plot1;

  procedure Plot3(F: in File_Type; V: in SVector3; Valid: in Boolean := True) is
  begin
    Plot1(F,Approx(V(1)),Valid);
    Plot1(F,Approx(V(2)),Valid);
    Plot1(F,Approx(V(3)),Valid);
  end Plot3;

  procedure Plot(F: in File_Type; T: in Flt; P: in Point) is
  begin
    Plot1(F,T);
    Plot3(F,P.Alpha,Compute_Alpha);
    Plot3(F,P.Beta,Compute_Beta);
    Plot3(F,P.Gamma);
    Plot3(F,P.M);
    if Compute_Yaw_Or_Roll then
      Plot1(F,ApproxYaw(P),Compute_Yaw);
      Plot1(F,ApproxRoll(P),Compute_Roll);
    end if;
    New_Line(F);
  end Plot;

  ------------------------------------

  procedure Norm2(V: in SVector3; S: in out Scalar) is
    S2: Scalar;
  begin
    Dot3(V,V,S2);
    Sqrt(S2,S);
  end Norm2;

  function Norm2(V: SVector3) return Scalar is
    S: Scalar;
  begin
    Norm2(V,S);
    return S;
  end Norm2;

  procedure Normalize(U: in SVector3; V: in out SVector3) is
    V2: Scalar renames V(2);
    V3: Scalar renames V(3);
  begin
    Dot3(U,U,V3);
    Sqrt(V3,V2);
    Inv(V2,V3);          --- V3 = 1/|U|
    Prod(U(1),V3,V(1));
    Prod(U(2),V3,V2);
    Mult(U(3),V3);
  end Normalize;

  procedure MakeG1(G2,G3: in Scalar; G: in out SVector3) is
    --- G1=Sqrt(1-G2*G2-G3*G3)
    G1: Scalar renames G(1);
    S2: Scalar renames G(2);
    S3: Scalar renames G(3);
  begin
    Prod(G2,G2,S2);
    Prod(G3,G3,S3);
    Sum(S2,S3,G1);
    Diff(SOne,G1,S2);
    Sqrt(S2,G1);
    Copy(G2,S2);
    Copy(G3,S3);
  end MakeG1;

  procedure MakeG2(G3,G1: in Scalar; G: in out SVector3) is
    --- G2=Sqrt(1-G3*G3-G1*G1)
    G2: Scalar renames G(2);
    S3: Scalar renames G(3);
    S1: Scalar renames G(1);
  begin
    Prod(G3,G3,S3);
    Prod(G1,G1,S1);
    Sum(S3,S1,G2);
    Diff(SOne,G2,S3);
    Sqrt(S3,G2);
    Copy(G3,S3);
    Copy(G1,S1);
  end MakeG2;

  procedure MakeG3(G1,G2: in Scalar; G: in out SVector3) is
    --- G3=Sqrt(1-G1*G1-G2*G2)
    G3: Scalar renames G(3);
    S1: Scalar renames G(1);
    S2: Scalar renames G(2);
  begin
    Prod(G1,G1,S1);
    Prod(G2,G2,S2);
    Sum(S1,S2,G3);
    Diff(SOne,G3,S1);
    Sqrt(S1,G3);
    Copy(G1,S1);
    Copy(G2,S2);
  end MakeG3;

  procedure MakeA(B,G: in SVector3; A: in out SVector3) is
    --- A inherits properties from B,G
    BG: SVector3;
  begin
    Cross3(B,G,BG);
    Normalize(BG,A);
  end MakeA;

  procedure MakeB(G,A: in SVector3; B: in out SVector3) is
    --- B inherits properties from G,A
    GA: Svector3;
  begin
    Cross3(G,A,GA);
    Normalize(GA,B);
  end MakeB;

  procedure CheapAB(G: in SVector3; A,B: in out SVector3) is
    --- notice: A and B inherit properties from G
    R1: constant Flt := abs(Approx(G(1)));
    R2: constant Flt := abs(Approx(G(2)));
    R3: constant Flt := abs(Approx(G(3)));
    R_Max: constant Flt := RMax(RMax(R1,R2),R3);
    JMax,JMin: Integer;
    V: SVector3;
  begin
    if R_Max=R3 then
      JMax := 3;
      if R2<R1 then JMin := 2; else JMin := 1; end if;
    elsif R_Max=R2 then
      JMax := 2;
      if R1<R3 then JMin := 1; else JMin := 3; end if;
    else
      JMax := 1;
      if R3<R2 then JMin := 3; else JMin := 2; end if;
    end if;
    SetZero(V);
    Copy(G(JMax),V(JMin));
    Neg(G(JMin),V(JMax));
    Normalize(V,A);
    Cross3(G,A,V);
    Normalize(V,B);
  end CheapAB;

  procedure MakeABG1(G2,G3: in Scalar; A,B,G: in out SVector3) is
    --- best if G2 and G3 are small
    --- notice: A and B inherit properties from G
  begin
    MakeG1(G2,G3,G);
    CheapAB(G,A,B);
  end MakeABG1;

  procedure MakeABG1(G2,G3: in Flt; A,B,G: in out SVector3) is
    --- best if G2 and G3 are small
    S2,S3: Scalar;
  begin
    Assign(G2,S2);
    Assign(G3,S3);
    MakeABG1(S2,S3,A,B,G);
  end MakeABG1;

  procedure MakeABG2(G3,G1: in Scalar; A,B,G: in out SVector3) is
    --- best if G3 and G1 are small
    --- notice: A and B inherit properties from G
  begin
    MakeG2(G3,G1,G);
    CheapAB(G,A,B);
  end MakeABG2;

  procedure MakeABG2(G3,G1: in Flt; A,B,G: in out SVector3) is
    --- best if G3 and G1 are small
    S3,S1: Scalar;
  begin
    Assign(G3,S3);
    Assign(G1,S1);
    MakeABG2(S3,S1,A,B,G);
  end MakeABG2;

  procedure MakeABG3(G1,G2: in Scalar; A,B,G: in out SVector3) is
    --- best if G1 and G2 are small
    --- notice: A and B inherit properties from G
  begin
    MakeG3(G1,G2,G);
    CheapAB(G,A,B);
  end MakeABG3;

  procedure MakeABG3(G1,G2: in Flt; A,B,G: in out SVector3) is
    --- best if G1 and G2 are small
    S1,S2: Scalar;
  begin
    Assign(G1,S1);
    Assign(G2,S2);
    MakeABG3(S1,S2,A,B,G);
  end MakeABG3;

  --------------------------------------------

  procedure Sandwich(U: in SVector3; A: in out SMatrix3) is
    --- AX = U x X x U
    A12: Scalar renames A(1,2);
    A23: Scalar renames A(2,3);
    A31: Scalar renames A(3,1);
  begin
    --- diag
    Prod(U(1),U(1),A23);
    Prod(U(2),U(2),A31);
    Prod(U(3),U(3),A12);
    Sum(A31,A12,A(1,1));
    Sum(A12,A23,A(2,2));
    Sum(A23,A31,A(3,3));
    --- off-diag
    Prod(U(1),U(2),A12);
    Neg(A12);
    Copy(A12,A(2,1));
    Prod(U(2),U(3),A23);
    Neg(A23);
    Copy(A23,A(3,2));
    Prod(U(3),U(1),A31);
    Neg(A31);
    Copy(A31,A(1,3));
  end Sandwich;

  procedure SS_ImK(Gamma: in SVector3; SS: in out Scalar; ImK: in out SMatrix3; Tmp: in out SVector3) is
    Tmp1: Scalar renames Tmp(1);
    BBGamma: SVector3 renames Tmp;
  begin
    DiagProd3(BB,Gamma,BBGamma);     -- BB*Gamma
    Dot3(BBGamma,Gamma,SS);          -- S*S
    Sandwich(BBGamma,ImK);           -- SS*K
    Inv(SS,Tmp1);                    -- 1/SS
    Mult(Mass,Tmp1);                 -- m/SS
    Mult3(Tmp1,ImK);                 -- m*K
    Add3(Inertia,ImK);               -- Inertia+m*K
  end SS_ImK;

  procedure ImK0(ImKo: in out SMatrix3) is
    --- ImK for r=(0,0,-b_3)
  begin
    Copy3(Inertia,ImKo);
    AddProd(Mass,BB(3),ImKo(1,1));
    AddProd(Mass,BB(3),ImKo(2,2));
  end ImK0;

  function ImK0 return SMatrix3 is
    ImKo: SMatrix3;
  begin
    ImK0(ImKo);
    return ImKo;
  end ImK0;

  procedure Energies(P: in Point; EKin,EPot: in out Scalar) is
    SS: Scalar;
    Omega: SVector3;
    Tmp: SVector3 renames Omega;
    ImK,InvImK: SMatrix3;
  begin
    SS_ImK(P.Gamma,SS,ImK,Tmp); -- I+mK
    Inv3(ImK,InvImK);           -- (I+mK)^{-1}
    Prod3(InvImK,P.M,Omega);    -- Omega
    Dot3(P.M,Omega,EKin);
    Mult(SHalf,EKin);           -- EKin=(1/2)M*Omega
                                ---
    Sqrt(SS,EPot);
    Mult(GForce,EPot);          -- EPot=m*g*S
  end Energies;

  procedure Ham(P: in Point; E: in out Scalar) is
    EPot: Scalar;
  begin
    Energies(P,E,EPot);
    Add(Epot,E);
  end Ham;

  procedure NumFix(P: in out Point) is
    --- first normalize (Alpha, Beta, and) Gamma
    --- then scale M to get energy P.E
    Tmp: SVector3;
    Tmp1: Scalar renames Tmp(1);
    Tmp2: Scalar renames Tmp(2);
    Tmp3: Scalar renames Tmp(3);
  begin
    if Not_Strunc then raise Not_Implemented; end if;
    if Compute_Alpha then
      Copy3(P.Alpha,Tmp);
      Normalize(Tmp,P.Alpha);
    end if;
    if Compute_Beta then
      Copy3(P.Beta,Tmp);
      Normalize(Tmp,P.Beta);
    end if;
    Copy3(P.Gamma,Tmp);
    Normalize(Tmp,P.Gamma);
    ---
    Energies(P,Tmp1,Tmp2);    --- EKin, EPot
    Diff(P.Energy,Tmp2,Tmp3); --- EKin_new
    Quot(Tmp3,Tmp1,Tmp2);     --- EKin_new/EKin
    Sqrt(Tmp2,Tmp1);
    Mult3(Tmp1,P.M);
    --- keeps P.YawPi and P.RollPi
  end NumFix;

  procedure Show_ABG_Errs(P: in Point) is
    S: Scalar;
  begin
    if Compute_Alpha then
      Norm2(P.Alpha,S);
      Sub(SOne,S);
      Show1("|A| err",MaxNorm(S));
    end if;
    if Compute_Beta then
      Norm2(P.Beta,S);
      Sub(SOne,S);
      Show1("|B| err",MaxNorm(S));
    end if;
    Norm2(P.Gamma,S);
    Sub(SOne,S);
    Show1("|G| err",MaxNorm(S));
  end Show_ABG_Errs;

  procedure Show_Approx_Energies(P: in Point) is
    EKin,EPot: Scalar;
  begin
    Energies(P,EKin,EPot);
    Show2("EKin,EPot",Approx(EKin),Approx(EPot));
    Add(EKin,EPot);
    Sub(P.Energy,EPot);
    Show1(" E  err  ",MaxNorm(EPot));
  end Show_Approx_Energies;

  procedure Show_Approx_Constraints(P: in Point) is
  begin
    Show_ABG_Errs(P);
    Show_Approx_Energies(P);
  end Show_Approx_Constraints;

  function Cyclic(Index: OneTwoThree; Shift: Natural) return OneTwoThree is
  begin
    return ((Index-1+Shift) mod 3)+1;
  end Cyclic;

  procedure MakeM(Index: in OneTwoThree; E: in Scalar; P: in out Point) is
    --- set P.E := E
    --- remaining comments are for Index=1
    --- determines M1 via the energy E
    J1: OneTwoThree renames Index;
    J2: constant OneTwoThree := Cyclic(Index,1);
    J3: constant OneTwoThree := Cyclic(Index,2);
    M: SVector3 renames P.M;
    Sgn: Integer;
    SS,S,Sigma,Tau2,STmp: Scalar;
    VTmp: SVector3;
    ImK,C: SMatrix3;
  begin
    SS_ImK(P.Gamma,SS,ImK,VTmp);   -- I+mK
    Inv3(ImK,C);                   -- C := (I+mK)^{-1}
                                   ---
    Prod(C(J1,J2),M(J2),Sigma);
    AddProd(C(J1,J3),M(J3),Sigma); -- Sigma = C12*M2+C13*M3
                                   ---
    Prod(M(J2),M(J2),STmp);
    Prod(C(J2,J2),STmp,Tau2);
    Prod(M(J3),M(J3),STmp);
    AddProd(C(J3,J3),STmp,Tau2);
    Mult(SHalf,Tau2);            -- (C22*M2*M2+C33*M3*M3)/2
    Prod(M(J2),M(J3),STmp);
    AddProd(C(J2,J3),STmp,Tau2); -- (C22*M2*M2+C33*M3*M3)/2 + C23*M2*M3
                                 ---
    Sqrt(SS,S);
    Mult(GForce,S);              -- mgS
    Add(S,Tau2);
    Sub(E,Tau2);                 -- Tau = (C22*M2*M2+C33*M3*M3)/2 + C23*M2*M3 + mgS-E
    Mult(STwo,Tau2);             -- 2*Tau
                                 ---
    Prod(Tau2,C(J1,J1),STmp);    -- 2*Tau*C(1,1)
    Prod(Sigma,Sigma,SS);
    Sub(STmp,SS);                -- Sigma*Sigma-2*Tau*C(1,1)
    Sqrt(SS,S);                  -- Sqrt(Sigma*Sigma-2*Tau*C(1,1))
                                 ---
    Prod(C(J1,J1),M(J1),STmp);
    Add(Sigma,STmp);
    Sgn := Sign(STmp);
    if Sgn>0 then
      Sum(Sigma,S,STmp);         -- Sigma+Sqrt(Sigma*Sigma-2*Tau*C(1,1))
    elsif Sgn<0 then
      Diff(Sigma,S,STmp);        -- Sigma-Sqrt(Sigma*Sigma-2*Tau*C(1,1))
    else
      raise Sorry;
    end if;
    Quot(Tau2,STmp,M(J1));
    Neg(M(J1));
    Copy(E,P.Energy);
    P.YawPi := 0;
    P.RollPi := 0;
  end MakeM;

  procedure MakeM1(E: in Scalar; P: in out Point) is
    --- set P.E := E
    --- determines M1 via the energy E
  begin
    MakeM(1,E,P);
  end MakeM1;

  procedure MakeM2(E: in Scalar; P: in out Point) is
    --- set P.E := E
    --- determines M2 via the energy E
  begin
    MakeM(2,E,P);
  end MakeM2;

  procedure MakeM3(E: in Scalar; P: in out Point) is
    --- set P.E := E
    --- determines M3 via the energy E
  begin
    MakeM(3,E,P);
  end MakeM3;

  procedure MakeZeroM(G1,G2: in Scalar; P: in out Point) is
    Gamma: SVector3 renames P.Gamma;
    BBGamma: SVector3 renames P.Alpha;
    SS: Scalar renames P.Beta(1);
  begin
    MakeG3(G1,G2,Gamma);
    DiagProd3(BB,Gamma,BBGamma);   -- BB*Gamma
    Dot3(BBGamma,Gamma,SS);        -- S*S
    Sqrt(SS,P.Energy);             -- S
    Mult(GForce,P.Energy);         -- E=m*g*S
                                   ---
    CheapAB(Gamma,P.Alpha,P.Beta);
    SetZero(P.M);
    P.YawPi := 0;
    P.RollPi := 0;
  end MakeZeroM;

  procedure MakePot1(EPot,BB1,BB2,BB3,G1: in Scalar; G2,G3: in out Scalar) is
    --- make G with given G1 and potential energy EPot
    --- best if G1 is not close to 1
    Tmp2: Scalar renames G2;
    Tmp3: Scalar renames G3;
    GG1,GG2,S: Scalar;
  begin
    Quot(EPot,GForce,S);
    Prod(S,S,Tmp2);       -- (EPot/mg)^2
    Sub(BB3,Tmp2);        -- Tmp2 = (EPot/mg)^2 -BB3
    Prod(G1,G1,GG1);      -- GG1 = G1^2
    Diff(BB1,BB3,Tmp3);
    Prod(Tmp3,GG1,S);     -- (BB1-BB3)*G1^2
    Sub(S,Tmp2);          -- Tmp2 = (EPot/mg)^2 -BB3 -(BB1-BB3)*G1^2
    Diff(BB2,BB3,Tmp3);   -- Tmp3 = BB2-BB3
    Quot(Tmp2,Tmp3,GG2);  -- GG2 = [(EPot/mg)^2 -BB3 -(BB1-BB3)*G1^2]/[BB2-BB3]
    Sqrt(GG2,G2);         -- G2 = Sqrt(GG2)
    Diff(SOne,GG2,S);
    Sub(GG1,S);           -- S = 1-G2^2-G1^2
    Sqrt(S,G3);           -- G3 = Sqrt(S)
  end MakePot1;

  procedure MakePot1(EPot,G1: in Scalar; G2,G3: in out Scalar) is
    --- make G with given G1 and potential energy EPot
    --- best if G1 is not close to 1
  begin
    MakePot1(EPot,BB(1),BB(2),BB(3),G1,G2,G3);
  end MakePot1;

  procedure MakePot2(EPot,G2: in Scalar; G3,G1: in out Scalar) is
    --- make G with given G2 and potential energy EPot
    --- best if G2 is not close to 1
  begin
    MakePot1(EPot,BB(2),BB(3),BB(1),G2,G3,G1);
  end MakePot2;

  procedure MakePot3(EPot,G3: in Scalar; G1,G2: in out Scalar) is
    --- make G with given G3 and potential energy EPot
    --- best if G3 is not close to 1
  begin
    MakePot1(EPot,BB(3),BB(1),BB(2),G3,G1,G2);
  end MakePot3;

  procedure MakePurePot1(E,G1: in Scalar; P: in out Point) is
    --- make P with given G1, energy E, and P.M=0
  begin
    Copy(G1,P.Gamma(1));
    MakePot1(E,G1,P.Gamma(2),P.Gamma(3));
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    SetZero(P.M);
    Copy(E,P.Energy);
    P.YawPi := 0;
    P.RollPi := 0;
  end MakePurePot1;

  procedure MakePurePot2(E,G2: in Scalar; P: in out Point) is
    --- make P with given G2, energy E, and P.M=0
  begin
    Copy(G2,P.Gamma(2));
    MakePot2(E,G2,P.Gamma(3),P.Gamma(1));
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    SetZero(P.M);
    Copy(E,P.Energy);
    P.YawPi := 0;
    P.RollPi := 0;
  end MakePurePot2;

  procedure MakePurePot3(E,G3: in Scalar; P: in out Point) is
    --- make P with given G3, energy E, and P.M=0
  begin
    Copy(G3,P.Gamma(3));
    MakePot2(E,G3,P.Gamma(1),P.Gamma(2));
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    SetZero(P.M);
    Copy(E,P.Energy);
    P.YawPi := 0;
    P.RollPi := 0;
  end MakePurePot3;

  procedure RSpSymmFun(Gamma: in SVector3; M: in out SVector3; Sg,Sm: in out Scalar) is
    --- assuming Gamma=(G1,G2,G3) with G1*G1+G2*G2=1
    --- use Omega=Gamma=(G1,G2,0)
    --- Sg: 3rd component of GFoorce*S^{-1}*(Gamma x BBGamma)
    --- Sm: 3rd component of M x Omega
    SS,InvSS,InvS,STmp: Scalar;
    Omega: SVector3 renames Gamma;
    BBGamma,DotM,VTmp: SVector3;
    ImK: SMatrix3;
  begin
    DiagProd3(BB,Gamma,BBGamma);   -- BB*Gamma
    Dot3(BBGamma,Gamma,SS);        -- S*S
    Inv(SS,InvSS);                 -- S^{-2}
    Sqrt(InvSS,InvS);              -- S^{-1}
                                   ---
    Cross3(Gamma,BBGamma,VTmp);    -- Gamma x BBGamma
    Prod3(InvS,VTmp,DotM);         -- S^{-1}*(Gamma x BBGamma)
    Mult3(GForce,DotM);            -- GFoorce*S^{-1}*(Gamma x BBGamma)
    Copy(DotM(3),Sg);
    ---
    Sandwich(BBGamma,ImK);         -- S^2*K
    Prod(Mass,InvSS,STmp);
    Mult3(STmp,ImK);               -- mK
    Add3(Inertia,ImK);             -- I+mK
    Prod3(ImK,Omega,M);            -- M
    Cross3(M,Omega,DotM);          -- M x Omega
    Copy(DotM(3),Sm);
  end RSpSymmFun;

  procedure RSpSymmFun(Angle: in Scalar; Sg,Sm: in out Scalar) is
    --- mostly for info and testing
    Gamma,M: SVector3;
  begin
    Cos(Angle,Gamma(1));
    Sin(Angle,Gamma(2));
    SetZero(Gamma(3));
    RSpSymmFun(Gamma,M,Sg,Sm);
  end RSpSymmFun;

  procedure RSpSymm(P: in out Point) is
    --- RS'-symmetric equilibrium with Omega=c*Gamma
    --- assuming P.Gamma=(G1,G2,0) with G1*G1+G2*G2=1
    --- leaves P.Alpha and P.Beta undefined
    Sg,Sm,C,EKin,EPot: Scalar;
    Tmp: Scalar renames EKin;
    Gamma: SVector3 renames P.Gamma;
    M: SVector3 renames P.M;
  begin
    if IsZero(Gamma(1)) or else IsZero(Gamma(2)) then
      SetZero(M);
    else
      RSpSymmFun(Gamma,M,Sg,Sm);
      if Sign(Sg)*Sign(Sm) >= 0 then raise Undefined; end if;
      Quot(Sg,Sm,Tmp);
      Neg(Tmp);
      Sqrt(Tmp,C);
      Mult3(C,M);
    end if;
    Energies(P,EKin,EPot);
    Sum(EKin,EPot,P.Energy);
    P.YawPi := 0;
    P.RollPi := 0;
  end RSpSymm;

  procedure RSpSymmG1(G1: in Scalar; P: in out Point) is
    --- find RS'-symmetric equilibrium, given G1
  begin
    MakeG2(SZero,G1,P.Gamma);
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    RSpSymm(P);
  end RSpSymmG1;

  procedure RSpSymmG2(G2: in Scalar; P: in out Point) is
    --- find RS'-symmetric equilibrium, given G2
  begin
    MakeG1(G2,SZero,P.Gamma);
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    RSpSymm(P);
  end RSpSymmG2;

  procedure RSpSymm(Angle: in Scalar; P: in out Point; OK: out Boolean) is
    --- find RS'-symmetric equilibrium with G1=Cos(Theta) and G2=Sin(Theta)
    --- meant mostly for numerical computations
  begin
    Cos(Angle,P.Gamma(1));
    Sin(Angle,P.Gamma(2));
    SetZero(P.Gamma(3));
    CheapAB(P.Gamma,P.Alpha,P.Beta);
    OK := True;
    RSpSymm(P);
  exception
    when others => OK := False;
  end RSpSymm;

  procedure VecField(Gamma,M: in SVector3; DotGamma,DotM: in out SVector3) is
    --- wasting storage
    SS,InvSS,InvS,InvSDotS,STmp: Scalar;
    BBGamma,BBDotGamma,Omega,SDotR,DotMt,OBBG,VTmp: SVector3;
    ImK,InvImK: SMatrix3;
  begin
    DiagProd3(BB,Gamma,BBGamma);         -- BB*Gamma
    Dot3(BBGamma,Gamma,SS);              -- S^2
    Inv(SS,InvSS);                       -- S^{-2}
    Sqrt(InvSS,InvS);                    -- S^{-1}
                                         ---
    Sandwich(BBGamma,ImK);               -- S^2*K
    Prod(Mass,InvSS,STmp);
    Mult3(STmp,ImK);                     -- mK
    Add3(Inertia,ImK);                   -- I+mK
    Inv3(ImK,InvImK);                    -- 1/(I+mK)
                                         ---
    Prod3(InvImK,M,Omega);               -- Omega
    Cross3(Gamma,Omega,DotGamma);        -- DotGamma := Gamma x Omega
    Cross3(M,Omega,DotM);                -- DotM := M x Omega
                                         ---
    Dot3(BBGamma,DotGamma,STmp);         -- SDotS
    Prod(InvSS,STmp,InvSDotS);
    DiagProd3(BB,DotGamma,BBDotGamma);   -- BB*DotGamma
    Prod3(InvSDotS,BBGamma,VTmp);        -- (DotS/S)BB*Gamma
    Diff3(VTmp,BBDotGamma,SDotR);        -- SDotR := (DotS/S)BB*Gamma-BB*DotGamma
                                         ---
    Cross3(Gamma,BBGamma,VTmp);          -- Gamma x BBGamma
    Prod3(InvS,VTmp,DotMt);              -- S^{-1}*(Gamma x BBGamma)
    Mult3(GForce,DotMt);                 -- DotMt := GFoorce*S^{-1}*(Gamma x BBGamma)
    Add3(DotMt,DotM);                    -- DotM := DotM+DotMt
                                         ---
    Cross3(Omega,BBGamma,OBBG);          -- Omega x BBGamma
    Cross3(OBBG,SDotR,VTmp);             -- (Omega x BBGamma) x SDotR
    Prod3(InvSS,VTmp,DotMt);             -- S^{-2}*[(Omega x BBGamma) x SDotR]
    Mult3(Mass,DotMt);                   -- DotMt := Mass*S^{-2}*[(Omega x BBGamma) x SDotR]
    Add3(DotMt,DotM);                    -- DotM := DotM+DotMt
  end VecField;

  procedure TrivSqrEigen0(LL1,LL2: in out Scalar) is
    S1,S2,T1,T2,D,B3,DetA,TrA,Tmp: Scalar;
  begin
    Prod(Mass,BB(3),Tmp);
    Sum(Inertia(1,1),Tmp,S1);            -- S1 = I11+m*BB3
    Sum(Inertia(2,2),Tmp,S2);            -- S2 = I22+m*BB3
    Prod(S1,S2,D);
    Prod(Inertia(1,2),Inertia(2,1),Tmp);
    Sub(Tmp,D);                          -- D = S1*S2-I12*I21
                                         ---
    Quot(BB(1),BB(3),Tmp);
    Diff(SOne,Tmp,T1);                   -- T1 = 1-BB1/BB3
    Quot(BB(2),BB(3),Tmp);
    Diff(SOne,Tmp,T2);                   -- T2 = 1-BB2/BB3
                                         ---
    Prod(T1,T2,DetA);
    Quot(BB(3),D,Tmp);
    Mult(Tmp,DetA);                      -- DetA = (BB3/D)*T1*T2
                                         ---
    Prod(T1,S1,TrA);
    Prod(T2,S2,Tmp);
    Add(Tmp,TrA);                        -- TrA = T1*S1+T2*S2
    Sqrt(BB(3),B3);
    Quot(B3,D,Tmp);
    Mult(Tmp,TrA);                       -- TrA = (B3/D)*[T1*S1+T2*S2]
                                         ---
    Neg(TrA);
    Roots2(TrA,DetA,LL1,LL2,Tmp);
    if not IsZero(Tmp) then raise Inconsistent_Data; end if;
    Mult(GForce,LL1);
    Mult(GForce,LL2);
  end TrivSqrEigen0;

  procedure TrivImEigen0(L1,L2: in out Scalar) is
    LL1,LL2: Scalar;
  begin
    TrivSqrEigen0(LL1,LL2);
    Neg(LL1);
    Neg(LL2);
    Sqrt(LL1,L1);
    Sqrt(LL2,L2);
  end TrivImEigen0;

  procedure YawModPi(A1,B1: in Scalar; Psi: in out Scalar) is
  begin
    ArcTan(A1,B1,Psi);
  end YawModPi;

  procedure RollModPi(G2,G3: in Scalar; Phi: in out Scalar) is
  begin
    ArcTan(G2,G3,Phi);
  exception
    when Range_Error =>
      Show0("RollModPi Warning: range may include Pi/2"); -- no need to abort
      ArcTan(G2,G3,Phi,False);
  end RollModPi;

  procedure DotYaw(A1,B1,DotA1,DotB1: in Scalar; DotPsi: in out Scalar) is
    Num,Den,Tmp: Scalar;
  begin
    Prod(DotA1,B1,Num);
    Prod(A1,DotB1,Tmp);
    Sub(Tmp,Num);         -- Num = dotalpha1*beta1-alpha1*dotbeta1
    Prod(A1,A1,Den);
    Prod(B1,B1,Tmp);
    Add(Tmp,Den);         -- Den = alpha1*alpha1+beta1*beta
    Quot(Num,Den,DotPsi); -- DotPsi = Num/Den
  end DotYaw;

  procedure DotRoll(G2,G3,DotG2,DotG3: in Scalar; DotPhi: in out Scalar) is
    Num,Den,Tmp: Scalar;
  begin
    Prod(DotG2,G3,Num);
    Prod(G2,DotG3,Tmp);
    Sub(Tmp,Num);         -- Num = dotgamma2*gamma3-gamma2*dotgamma3
    Prod(G2,G2,Den);
    Prod(G3,G3,Tmp);
    Add(Tmp,Den);         -- Den = gamma2*gamma2+gamma3*gamma3
    Quot(Num,Den,DotPhi); -- DotPhi = Num/Den
  end DotRoll;

  procedure NumPi(X: in out Scalar; N: out Integer; OK: out Boolean) is
    --- subtract N*Pi from X to get enclosure in (-Pi/2,Pi/2) of 0
  begin
    N := Integer(Flt'Rounding(Approx(X)/Ada.Numerics.Pi)); -- guess
    if N/=0 then
      AddProd(Flt(-N),PiFull,X);
    end if;
    --- check guess
    OK := MaxNorm(X)<InfPiHalf;
    if Not_Strunc then
      OK := OK and then Contains0(X)=True;
    end if;
  end NumPi;

  procedure YawNumPi(T,Psi0,PsiT: in Scalar; A1,B1,DotA1,DotB1: in Scalar; N: in out Integer; OK: out Boolean) is
    --- input  N: lifted Psi(0)=Psi0+N*Pi
    --- output N: lifted Psi(T)=PsiT+N*Pi
    --- Alpha and DotAlpha are enclosures on [0,T]
    K: Integer;
    PsiK,KPi: Scalar;
    DPsi: Scalar renames KPi;
  begin
    DotYaw(A1,B1,DotA1,DotB1,DPsi);
    Mult(T,DPsi);                 -- increment
    Diff(PsiT,DPsi,PsiK);         -- PsiK encloses Psi0+K*Pi for some K
    Diff(PsiK,Psi0,KPi);          -- KPi encloses K*Pi
    NumPi(KPi,K,OK);              -- K
    N := N-K;
  end YawNumPi;

  procedure RollNumPi(T,Phi0,PhiT: in Scalar; G2,G3,DotG2,DotG3: in Scalar; N: in out Integer; OK: out Boolean) is
    --- input  N: lifted Phi(0)=Phi0+N*Pi
    --- output N: lifted Phi(T)=PhiT+N*Pi
    --- Alpha and DotAlpha are enclosures on [0,T]
    K: Integer;
    PhiK,KPi: Scalar;
    DPhi: Scalar renames KPi;
  begin
    DotRoll(G2,G3,DotG2,DotG3,DPhi);
    Mult(T,DPhi);                 -- increment
    Diff(PhiT,DPhi,PhiK);         -- PhiK encloses Phi0+K*Pi for some K
    Diff(PhiK,Phi0,KPi);          -- KPi encloses K*Pi
    NumPi(KPi,K,OK);              -- K
    N := N-K;
  end RollNumPi;

  procedure ShowYaw(Psi: in Scalar; NPi: in Integer) is
    Yaw: Scalar;
  begin
    Prod(Flt(NPi),PiFull,Yaw);
    Add(Psi,Yaw);
    Reduce_Precision(64,Yaw);
    Show1(" Yaw = ",Yaw);
  end ShowYaw;

  procedure ShowRoll(Phi: in Scalar; NPi: in Integer) is
    Roll: Scalar;
  begin
    Prod(Flt(NPi),PiFull,Roll);
    Add(Phi,Roll);
    Reduce_Precision(64,Roll);
    Show1("Roll = ",Roll);
  end ShowRoll;

  procedure ShowYaw(P: in Point) is
    A1: Scalar renames P.Alpha(1);
    B1: Scalar renames P.Beta(1);
    Psi: Scalar;
  begin
    if Compute_Yaw then
      if IsZero(A1) and then IsZero(B1) then
        Show0(" Tan Yaw = 0/0");
      else
        YawModPi(A1,B1,Psi);
        ShowYaw(Psi,P.YawPi);
      end if;
    end if;
  end ShowYaw;

  procedure ShowRoll(P: in Point) is
    G2: Scalar renames P.Gamma(2);
    G3: Scalar renames P.Gamma(3);
    Phi: Scalar;
  begin
    if Compute_Roll then
      if IsZero(G2) and then IsZero(G3) then
        Show0("Tan Roll = 0/0");
      else
        RollModPi(G2,G3,Phi);
        ShowRoll(Phi,P.YawPi);
      end if;
    end if;
  end ShowRoll;

  function ApproxYaw(P: Point) return Flt is
  begin
    if not Compute_Yaw then return Zero; end if;
    declare
      use Ada.Numerics, Flt_EF;
      Eps: constant Radius := 1.0E-512;
      B1: constant Flt := Approx(P.Beta(1));
    begin
      if abs(B1)<Eps then return Half*Pi; end if;
      return Flt(P.YawPi)*Pi+ArcTan(Approx(P.Alpha(1))/B1);
    end;
  end ApproxYaw;

  function ApproxRoll(P: Point) return Flt is
  begin
    if not Compute_Roll then return Zero; end if;
    declare
      use Ada.Numerics, Flt_EF;
      Eps: constant Radius := 1.0E-512;
      G3: constant Flt := Approx(P.Gamma(3));
    begin
      if abs(G3)<Eps then return Half*Pi; end if;
      return Flt(P.RollPi)*Pi+ArcTan(Approx(P.Gamma(2))/G3);
    end;
  end ApproxRoll;

  procedure QFun(Theta: in Scalar; Q: in out Scalar) is
    DQ: Scalar;
  begin
    TQFun(Theta,Q,DQ,False);
  end QFun;

  procedure TQFun(Theta: in Scalar; Q,DQ: in out Scalar; NeedDQ: in Boolean := True) is
    BB1: Scalar renames BB(1);
    BB2: Scalar renames BB(2);
    I11: Scalar renames Inertia(1,1);
    I12: Scalar renames Inertia(1,2);
    I22: Scalar renames Inertia(2,2);
    TwoInvMass: Scalar renames DQ;
    SumBB,SumInvBB,B12,InvB12,SinTheta,CosTheta,U,V,W,X,Tmp: Scalar;
    InvBB1: Scalar renames W;
    InvBB2: Scalar renames X;
  begin
    Sum(BB2,BB1,SumBB);             --- SumBB = B2^2+B1^2
    Inv(BB2,InvBB2);
    Inv(BB1,InvBB1);
    Sum(InvBB2,InvBB1,SumInvBB);    --- SumInvBB = B2^{-2}+B1^{-2}
    Prod(BB1,BB2,Tmp);
    Sqrt(Tmp,B12);                  --- B12 = B1*B2
    Inv(B12,InvB12);                --- InvB12 = 1/(B1*B2)
    Quot(STwo,Mass,TwoInvMass);     --- TwoInvMass = 2/m
                                    ---
    Prod(TwoInvMass,I12,U);         --- U = (2/m)*I12
    Sin(Theta,SinTheta);
    Prod(B12,SinTheta,Tmp);
    Sub(Tmp,U);                     --- U = (2/m)*I12 - B1*B2*SinTheta
                                    ---
    Cos(Theta,CosTheta);
    Prod(SumInvBB,CosTheta,V);      --- V = SumInvBB*CosTheta
    Diff(InvBB2,InvBB1,Tmp);
    Add(Tmp,V);                     --- V = (B2^{-2}-B1^{-2}) + SumInvBB*CosTheta
                                    ---
    Prod(SumBB,CosTheta,W);         --- W = SumBB*CosTheta
    Diff(BB2,BB1,Tmp);
    Add(Tmp,W);                     --- W = (B2^2-B1^2) + SumBB*CosTheta
    Diff(I11,I22,X);
    Prod(TwoInvMass,X,Tmp);
    Add(Tmp,W);                     --- W = (2/m)*(I11-I22) + (B2^2-B1^2) + SumBB*CosTheta
                                    ---
    Prod(InvB12,SinTheta,X);        --- X = SinTheta/(B1*B2)
    Prod(U,V,Q);
    Prod(W,X,Tmp);
    Add(Tmp,Q);                     --- Q = U*V + W*X
                                    ---
    if NeedDQ then
      declare
        DU,DV,DW,DX: Scalar;
      begin
        Prod(B12,CosTheta,DU);
        Neg(DU);                    --- DU = -B1*B2*CosTheta
        Prod(SumInvBB,SinTheta,DV);
        Neg(DV);                    --- DV = -SumInvBB*SinTheta
        Prod(SumBB,SinTheta,DW);
        Neg(DW);                    --- DW = -SumBB*SinTheta
        Prod(InvB12,CosTheta,DX);   --- DX = CosTheta/(B1*B2)
        Prod(DU,V,DQ);              --- DU*V
        Prod(U,DV,Tmp);             --- U*DV
        Add(Tmp,DQ);
        Prod(DW,X,Tmp);             --- DW*X
        Add(Tmp,DQ);
        Prod(W,DX,Tmp);             --- W*DX
        Add(Tmp,DQ);                --- DQ = DU*V + U*DV + DW*X + W*DX
      end;
    end if;
  end TQFun;

  ----------------------------------------

  procedure Det4(A: in SMatrix4; S: in out Scalar) is
    use MS;
    I1,I2,I3: Positive;
    D3,Tmp: Scalar;
    B: SMatrix3;
  begin
    SetZero(S);
    for I in 1..4 loop
      I1 := (I mod 4)+1;
      I2 := ((I+1) mod 4)+1;
      I3 := ((I+2) mod 4)+1;
      for J in 1..3 loop
        Copy(A(I1,J+1),B(1,J));
        Copy(A(I2,J+1),B(2,J));
        Copy(A(I3,J+1),B(3,J));
      end loop;
      Det3(B,D3);
      Prod(A(I,1),D3,Tmp);
      if (I mod 2)=1 then Add(Tmp,S); else Sub(Tmp,S); end if;
    end loop;
  end Det4;

  procedure HatDVecField_Stat(Omega3: in Scalar; DX: in out SMatrix4) is
    --- sharp if BB(3)=1
    I33: Scalar renames Inertia(3,3);
    D,S,Tmp: Scalar;
    HatJ,HatB: SMatrix2;
  begin
    --- HatJ and D
    Prod(Mass,BB(3),Tmp);            --- Tmp = m*BB(3)
    Sum(Inertia(2,2),Tmp,HatJ(1,1));
    Neg(Inertia(1,2),HatJ(1,2));
    Neg(Inertia(2,1),HatJ(2,1));
    Sum(Inertia(1,1),Tmp,HatJ(2,2));
    Det2(HatJ,D);                    --- D=det(HatJ)
    --- matrix HatB
    for K in 1..2 loop
      Prod(Mass,BB(K),Tmp);
      Prod(HatJ(1,K),Tmp,HatB(1,K)); --- HatB(1,K) = HatJ(1,K)*m*BB(K)
      Prod(HatJ(2,K),Tmp,HatB(2,K)); --- HatB(2,K) = HatJ(2,K)*m*BB(K)
    end loop;
    --- DX: order Omega3**2
    for I in 1..4 loop
      for K in 1..4 loop SetZero(DX(I,K)); end loop;
    end loop;
    Prod(I33,HatB(2,1),DX(3,1));
    Neg(DX(3,1));                    --- (3,1):  -I33*HatB21
    Prod(I33,HatB(2,2),DX(3,2));
    Neg(DX(3,2));                    --- (3,2):  -I33*HatB22
    Prod(I33,HatB(1,1),DX(4,1));     --- (4,1):   I33*HatB11
    Prod(I33,HatB(1,2),DX(4,2));     --- (4,2):   I33*HatB12
    --- multiply DX by Omega3
    for I in 1..4 loop
      for K in 1..4 loop Mult(Omega3,DX(I,K)); end loop;
    end loop;
    --- DX: add order Omega3**1
    Sub(HatB(2,1),DX(1,1));          --- (1,1):  -HatB21
    Diff(D,HatB(2,2),Tmp);
    Add(Tmp,DX(1,2));                --- (1,2):  D-HatB22
    Diff(HatB(1,1),D,Tmp);
    Add(Tmp,DX(2,1));                --- (2,1):  -D+HatB11
    Add(HatB(1,2),DX(2,2));          --- (2,2):  HatB12
    Prod(I33,HatJ(2,1),Tmp);
    Sub(Tmp,DX(3,3));                --- (3,3):  -I33*HatJ21
    Prod(I33,HatJ(2,2),Tmp);
    Sub(D,Tmp);
    Sub(Tmp,DX(3,4));                --- (3,4):  D-I33*HatJ22
    Prod(I33,HatJ(1,1),Tmp);
    Sub(D,Tmp);
    Add(Tmp,DX(4,3));                --- (4,3):  -D+I33*HatJ11
    Prod(I33,HatJ(1,2),Tmp);
    Add(Tmp,DX(4,4));                --- (4,4):  I33*HatJ12
    --- multiply DX by Omega3
    for I in 1..4 loop
      for K in 1..4 loop Mult(Omega3,DX(I,K)); end loop;
    end loop;
    --- DX: add order Omega**0
    Sub(HatJ(2,1),DX(1,3));          --- (1,3):  -HatJ21
    Sub(HatJ(2,2),DX(1,4));          --- (1,4):  -HatJ22
    Add(HatJ(1,1),DX(2,3));          --- (2,3):  HatJ11
    Add(HatJ(1,2),DX(2,4));          --- (2,4):  HatJ12
    Sqrt(BB(3),S);
    Quot(GForce,S,Tmp);
    Prod(D,Tmp,S);                   --- S = D*mg/B3
    Diff(BB(3),BB(2),Tmp);
    Mult(S,Tmp);
    Add(Tmp,DX(3,2));                --- (3,2):  (BB3-BB2)*S
    Diff(BB(1),BB(3),Tmp);
    Mult(S,Tmp);
    Add(Tmp,DX(4,1));                --- (4,1):  (BB1-BB3)*S
  end HatDVecField_Stat;

  procedure HatCharPoly(Omega3,Lambda: in Scalar; P: in out Scalar) is
    DX: SMatrix4;
  begin
    HatDVecField_Stat(Omega3,DX);
    for I in 1..4 loop Sub(Lambda,DX(I,I)); end loop;
    Det4(DX,P);
  end HatCharPoly;

  procedure DetImK0(D: in out Scalar) is
    S1,S2,Tmp: Scalar;
  begin
    Prod(Mass,BB(3),Tmp);
    Sum(Inertia(1,1),Tmp,S1);            --- S1 = I11+m*BB(3)
    Sum(Inertia(2,2),Tmp,S2);            --- S2 = I22+m*BB(3)
    Prod(S1,S2,D);
    Prod(Inertia(1,2),Inertia(2,1),Tmp);
    Sub(Tmp,D);                          --- D=S1*S2-I12*I21
  end DetImK0;

  procedure DVecField_Stat(Omega3: in Scalar; DX: in out SMatrix4) is
    --- divide result from HatDVecField_Stat by DetImK0
    D,InvD: Scalar;
  begin
    DetImK0(D);
    Inv(D,InvD);
    HatDVecField_Stat(Omega3,DX);
    for I in 1..4 loop
      for J in 1..4 loop Mult(InvD,DX(I,J)); end loop;
    end loop;
  end DVecField_Stat;

  ----------------------------------------

  procedure Val(F: in Flt_Affine; V: in SVector3; S: in out Scalar) is
  begin
    Assign(F(0),S);
    AddProd(F(1),V(1),S);
    AddProd(F(2),V(2),S);
    AddProd(F(3),V(3),S);
  end Val;

end Rattleback;
