with Ada.Text_IO, Interfaces, Ints, Logicals, Rationals, Flts, ScalInfo, Params;
use Ada.Text_IO, Interfaces, Ints, Logicals, Rationals, Flts, ScalInfo, Params;
with Small_Vectors, Params.More;

pragma Elaborate_All (Ada.Text_IO,Interfaces,Ints,Logicals,Rationals,Flts,ScalInfo,Params);
pragma Elaborate_All (Small_Vectors,Params.More);

generic

  FRP: in FRParam;
  ABFlag: in Flag := 0; --- nothing set

  -------------------- begin_include scalars Scalar

  type Scalar is private;
  with procedure Proper_Rounding is <>;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure FltBall(S: in Scalar; C,R: out Flt) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Reduce_Precision(Binary_Digits: in Positive; S: in out Scalar) is <>;

  -------------- end_include scalars

package Rattleback is

  Compute_Alpha:  constant Boolean := (ABFlag and Alpha_Defined)=Alpha_Defined;
  Compute_Beta:   constant Boolean := (ABFlag and Beta_Defined)=Beta_Defined;
  Compute_Yaw:    constant Boolean := (ABFlag and Yaw_Defined)=Yaw_Defined;   -- implies Compute_Alpha and Compute_Beta
  Compute_Roll:   constant Boolean := (ABFlag and Roll_Defined)=Roll_Defined;

  Compute_Alpha_And_Beta: constant Boolean := Compute_Alpha and Compute_Beta;
  Compute_Yaw_And_Roll:   constant Boolean := Compute_Yaw and Compute_Roll;
  Compute_Yaw_Or_Roll:    constant Boolean := Compute_Yaw or Compute_Roll;

  package PMore is new Params.More (Scalar => Scalar);
  package MS renames PMore.MS;
  subtype SVector2 is MS.Vector2;
  subtype SVector3 is MS.Vector3;
  subtype SVector4 is PMore.Vector4;
  subtype SMatrix2 is MS.Matrix2;
  subtype SMatrix3 is MS.Matrix3;
  subtype SMatrix4 is PMore.Matrix4;
  use PMore;

  procedure SetZero(U: in out SVector3);
  procedure Reduce_Precision(Binary_Digits: in Positive; U: in out SVector3);
  procedure Assign(U1,U2,U3: in Flt; U: in out SVector3);
  procedure ShowApprox1(N: in String; V: in SVector3);
  procedure ShowApprox1(N: in String; A: in SMatrix3);
  procedure ShowApprox1(N: in String; A: in SMatrix2);
  procedure Show1(N: in String; V: in SVector3);
  procedure Show1(N: in String; A: in SMatrix2);
  procedure Show1(N: in String; A: in SMatrix3);
  procedure Put(F: in File_Type; V: in SVector3; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; V: in out SVector3; Decimal: in Boolean := False);
  procedure Put(F: in File_Type; A: in SMatrix2; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; A: in out SMatrix2; Decimal: in Boolean := False);
  procedure Put(F: in File_Type; A: in SMatrix3; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; A: in out SMatrix3; Decimal: in Boolean := False);
  --- compatible with Matrices
  procedure Write(FileName: in String; A: in SMatrix2; Decimal: in Boolean := False);
  procedure Read(FileName: in String; A: in out SMatrix2; Decimal: in Boolean := False);
  procedure Write(FileName: in String; A: in SMatrix3; Decimal: in Boolean := False);
  procedure Read(FileName: in String; A: in out SMatrix3; Decimal: in Boolean := False);

  procedure Enlarge(S: in out Scalar);
  procedure Enlarge(V: in out SVector3);

  SRP: constant RParam := Scal(FRP);
  Mass: Scalar renames SRP.Mass;
  GForce: Scalar renames SRP.GForce;
  BB: SVector3 renames SRP.BB;
  Inertia: SMatrix3 renames SRP.Inertia;

  ----------------------------------

  type Point is
    record
      Alpha,Beta,Gamma,M: SVector3;
      Energy: Scalar;
      YawPi: Integer := 0;
      RollPi: Integer := 0;
    end record;
  type Points is array(Integer range <>) of Point;

  procedure Copy(P1: in Point; P2: in out Point);
  procedure ShowApprox1(N: in String; P: in Point);
  procedure Show1(N: in String; P: in Point);
  function MaxNormDiff(P,Q: Point) return Radius; --- only includes components Gamma and M
  procedure Plot(F: in File_Type; T: in Flt; P: in Point);
  procedure Put(F: in File_Type; P: in Point; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Point; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Point; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Point; Decimal: in Boolean := False);
  function Read_Flag(FileName: String) return Flag;

  ------------------------------

  package SMV is new Small_Vectors (Scalar => Scalar, MS => MS);
  use SMV;

  procedure Norm2(V: in SVector3; S: in out Scalar);
  function Norm2(V: SVector3) return Scalar;
  procedure Normalize(U: in SVector3; V: in out SVector3);
  procedure MakeG1(G2,G3: in Scalar; G: in out SVector3);
  procedure MakeG2(G3,G1: in Scalar; G: in out SVector3);
  procedure MakeG3(G1,G2: in Scalar; G: in out SVector3);
  procedure MakeA(B,G: in SVector3; A: in out SVector3);
  procedure MakeB(G,A: in SVector3; B: in out SVector3);
  procedure CheapAB(G: in SVector3; A,B: in out SVector3);
  procedure MakeABG1(G2,G3: in Scalar; A,B,G: in out SVector3);
  procedure MakeABG1(G2,G3: in Flt; A,B,G: in out SVector3);
  procedure MakeABG2(G3,G1: in Scalar; A,B,G: in out SVector3);
  procedure MakeABG2(G3,G1: in Flt; A,B,G: in out SVector3);
  procedure MakeABG3(G1,G2: in Scalar; A,B,G: in out SVector3);
  procedure MakeABG3(G1,G2: in Flt; A,B,G: in out SVector3);

  procedure Sandwich(U: in SVector3; A: in out SMatrix3);
  procedure SS_ImK(Gamma: in SVector3; SS: in out Scalar; ImK: in out SMatrix3; Tmp: in out SVector3);
  procedure ImK0(ImKo: in out SMatrix3); --- for r=(0,0,-b_3)
  function ImK0 return SMatrix3;

  procedure Energies(P: in Point; EKin,EPot: in out Scalar);
  procedure Ham(P: in Point; E: in out Scalar);
  procedure NumFix(P: in out Point); -- restore constraints, using Normalize and scaling of M
  procedure Show_ABG_Errs(P: in Point);
  procedure Show_Approx_Energies(P: in Point);
  procedure Show_Approx_Constraints(P: in Point);

  function Cyclic(Index: OneTwoThree; Shift: Natural) return OneTwoThree; -- cyclic shift in {1,2,3}
  procedure MakeM(Index: in OneTwoThree; E: in Scalar; P: in out Point); -- make M(Index), need rough value
  procedure MakeM1(E: in Scalar; P: in out Point);                   -- make M(1), need rough value
  procedure MakeM2(E: in Scalar; P: in out Point);               -- make M(2), need rough value
  procedure MakeM3(E: in Scalar; P: in out Point);           -- make M(3), need rough value
  procedure MakeZeroM(G1,G2: in Scalar; P: in out Point); -- make P with given P.Gamma and P.M=0

  procedure MakePot1(EPot,G1: in Scalar; G2,G3: in out Scalar); -- make G with given G1 and pot energy EPot
  procedure MakePot2(EPot,G2: in Scalar; G3,G1: in out Scalar); -- make G with given G2 and pot energy EPot
  procedure MakePot3(EPot,G3: in Scalar; G1,G2: in out Scalar); -- make G with given G3 and pot energy EPot
  procedure MakePurePot1(E,G1: in Scalar; P: in out Point);     -- make P with given G1, energy E, and M=0
  procedure MakePurePot2(E,G2: in Scalar; P: in out Point);     -- make P with given G2, energy E, and M=0
  procedure MakePurePot3(E,G3: in Scalar; P: in out Point);     -- make P with given G3, energy E, and M=0

  procedure RSpSymmFun(Angle: in Scalar; Sg,Sm: in out Scalar);
  procedure RSpSymmG1(G1: in Scalar; P: in out Point);  -- find RS'-symmetric equilibrium, given G1 (preferably<Half)
  procedure RSpSymmG2(G2: in Scalar; P: in out Point);  -- find RS'-symmetric equilibrium, given G2 (preferably<Half)
  procedure RSpSymm(Angle: in Scalar; P: in out Point; OK: out Boolean); -- find RS'-symmetric equilibrium with ...

  procedure VecField(Gamma,M: in SVector3; DotGamma,DotM: in out SVector3);
  procedure TrivSqrEigen0(LL1,LL2: in out Scalar);
  procedure TrivImEigen0(L1,L2: in out Scalar);

  --- Yaw and Roll
  procedure YawModPi(A1,B1: in Scalar; Psi: in out Scalar);
  procedure RollModPi(G2,G3: in Scalar; Phi: in out Scalar);
  procedure DotYaw(A1,B1,DotA1,DotB1: in Scalar; DotPsi: in out Scalar);
  procedure DotRoll(G2,G3,DotG2,DotG3: in Scalar; DotPhi: in out Scalar);
  procedure YawNumPi(T,Psi0,PsiT: in Scalar; A1,B1,DotA1,DotB1: in Scalar; N: in out Integer; OK: out Boolean);
  procedure RollNumPi(T,Phi0,PhiT: in Scalar; G2,G3,DotG2,DotG3: in Scalar; N: in out Integer; OK: out Boolean);
  procedure ShowYaw(Psi: in Scalar; NPi: in Integer);
  procedure ShowRoll(Phi: in Scalar; NPi: in Integer);
  procedure ShowYaw(P: in Point);
  procedure ShowRoll(P: in Point);
  function ApproxYaw(P: Point) return Flt;
  function ApproxRoll(P: Point) return Flt;

  --- RS'-symmetric stationary solutions
  procedure QFun(Theta: in Scalar; Q: in out Scalar);
  procedure TQFun(Theta: in Scalar; Q,DQ: in out Scalar; NeedDQ: in Boolean := True);

  --------------------------

  procedure Det4(A: in SMatrix4; S: in out Scalar);
  procedure HatDVecField_Stat(Omega3: in Scalar; DX: in out SMatrix4);
  procedure HatCharPoly(Omega3,Lambda: in Scalar; P: in out Scalar);
  procedure DetImK0(D: in out Scalar);
  procedure DVecField_Stat(Omega3: in Scalar; DX: in out SMatrix4);

  --------------------------

  type Flt_Affine is array(0 .. 3) of Flt;
  procedure Val(F: in Flt_Affine; V: in SVector3; S: in out Scalar); -- F0+F1*V1+F2*V2+F3*V3

private

  procedure Plot1(F: in File_Type; R: in Flt; Valid: in Boolean := True);
  procedure Plot3(F: in File_Type; V: in SVector3; Valid: in Boolean := True);

  procedure MakePot1(EPot,BB1,BB2,BB3,G1: in Scalar; G2,G3: in out Scalar);
  procedure RSpSymmFun(Gamma: in SVector3; M: in out SVector3; Sg,Sm: in out Scalar);
  procedure RSpSymm(P: in out Point);
  procedure NumPi(X: in out Scalar; N: out Integer; OK: out Boolean);

  SZero:      constant Scalar  := Scal(Zero);
  SOne:       constant Scalar  := Scal(One);
  STwo:       constant Scalar  := Scal(Two);
  SHalf:      constant Scalar  := Scal(Half);
  SInfo:  constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean := SInfo.IsNumeric;
  Not_STrunc: constant Boolean := not STrunc;

end Rattleback;
