with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
with Tangents, Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps.Balls);
pragma Elaborate_All (Tangents,Rattleback.Flows);

procedure Roll_Der is --- derivative of G1_initial --> G3_final for P13-roll

  FRP:     constant FRParam  := Param13;    -- rattleback parameters
  Rho:     constant Radius   := One;        -- for Tangents
  ABFlag:  constant Flag     := 0;          -- no need to compute Alpha and Beta
  NDigits: constant Positive := 3072;       -- number of binary digits
  Dummy:   constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package Tangs is new Tangents (Dim => 1, Rho => Rho, Scalar => Scalar);
  package SRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);
  use Tangs;
  package TRB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Tangent);

  PDeg:         constant Positive := 511;   -- Taylor degree
  Prec:         constant Positive := 900;   -- target precision for Integrate
  NegLogMinRad: constant Positive := 9;     -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;    -- MaxRad := MaxRadFac*MinRad

  use TRB;
  package TRBFlows is new TRB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use TRBFlows;

  NT:     constant Positive := 7;                     -- max number of concurrent tasks
  FileName: constant String := "P13-roll.0102";       -- determines starting point
  TNeed:       constant Flt := 8.5;                   -- true T is 9.03...
  F: constant    Flt_Affine := (Zero,Zero,Zero,One);  -- Poincare section at M3=0

  LogFac: constant Positive := 256;  -- Half**LogFac is the size of the initial tangent vectors

  procedure Adjust_G1(G1: in out Scalar) is
    R_DG1: constant Flt := TwoPower(-1536); -- approx 4E-463
    DG1: Scalar;
  begin
    Show0("adding [-DG1,DG1] to G1");
    DiskAt0(R_DG1,DG1);
    Add(DG1,G1);
  end Adjust_G1;

  procedure Make_Initial(P: in out Point) is
    --- RS'-symmetry has G3=M3=0
    TSize: constant Radius := TwoPower(-LogFac);
    M1: constant Flt := Flt(   76555)/Flt(262144); ---  0.292034149169921875
    M2: constant Flt := Flt(-1908461)/Flt(262144); --- -7.280200958251953125
    SP: SRB.Point;
    G1,G3,E: Tangent;
  begin
    SRB.Read(Input_Dir & FileName,SP);
    for I in 1..3 loop
      Assign(SP.Alpha(I),P.Alpha(I)); -- not really needed
      Assign(SP.Beta(I),P.Beta(I));   -- not really needed
    end loop;
    Copy(SP.Gamma(1),G1(0));
    Adjust_G1(G1(0));
    Assign(TSize,G1(1));
    SetZero(G3);
    MakeG2(G3,G1,P.Gamma);
    Assign(M1,M2,Zero,P.M);
    Ham(P,E);                --- E is determined by (G,M), just for info
    Copy(E,P.Energy);
  end Make_Initial;

  procedure Save_Der(P: in Point) is
    InvTSize: constant Radius := TwoPower(LogFac);
    SDer: Scalar;
  begin
    Prod(InvTSize,P.Gamma(3)(1),SDer);
    Write(Output_Dir & FileName & ".Der",SDer);
    Reduce_Precision(64,SDer);
    Show1("Der =",SDer);
  end Save_Der;

  P0,P1: Point;

begin
  NTasks := NT;
  Verbosity := 3;

  Show0("running Roll_Der with");
  Show_Rattle_Param;
  Make_Initial(P0);
  ShowApprox1("P0",P0);
  Neg2Pos_Poincare(TNeed,F,P0,P1);
  ShowApprox1("P1",P1);
  Save_Der(P1);

end Roll_Der;
