with Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
use Ada.Text_IO, Strings, Ints, Flts, Globals, Params, Reps.Balls;
with Rattleback, Rattleback.Flows;

pragma Elaborate_All (Ada.Text_IO,Strings,Ints,Flts,Globals,Params,Reps.Balls);
pragma Elaborate_All (Rattleback.Flows);

procedure Roll_Point is --- RS'-symmetric roll-over orbit P13-43-pot1

  Sign_DG1:  constant Integer := GetArg;
  --- Sign_DG1=-1  -->  use G1-DG1            want G3>0 @end
  --- Sign_DG1= 1  -->  use G1+DG1            want G3<0 @end
  --- Sign_DG1= 0  -->  use G1+[-DG1,DG1]     want existence, data

  FRP:     constant FRParam  := Param13;    -- rattleback parameters
  ABFlag:  constant Flag     := Choose(Sign_DG1=0,YawRoll_Defined,0);
  NDigits: constant Positive := 4096;       -- number of binary digits
  Dummy:   constant Positive := Reps.Set_Precision(NDigits);

  subtype Scalar is Ball;
  package RB is new Rattleback(FRP => FRP, ABFlag => ABFlag, Scalar => Scalar);

  PDeg:         constant Positive := 511;   -- Taylor degree
  Prec:         constant Positive := 900;   -- target precision for Integrate
  NegLogMinRad: constant Positive := 9;     -- MinRad := Half**NegLogMinRad
  MaxRadFac:    constant Positive := 64;    -- MaxRad := MaxRadFac*MinRad

  use RB;
  package RBFlow is new RB.Flows(PDeg => PDeg, Prec => Prec, NegLogMinRad => NegLogMinRad, MaxRadFac => MaxRadFac);
  use RBFlow;

  FileName: constant String := "P13-roll.0102";      -- determines the starting point
  F:    constant Flt_Affine := (Zero,Zero,Zero,One); -- Poincare section at M3=0
  TNeed:       constant Flt := 8.5;                  -- Poincare section tau'
  NT:     constant Positive := 7;                    -- max number of concurrent tasks

  procedure Adjust_G1(G1: in out Scalar) is
    R_DG1: constant Flt := TwoPower(-1536);          -- approx 4E-463
    DG1: Scalar;
  begin
    if Sign_DG1=0 then
      Show0("adding [-DG1,DG1] to G1");
      DiskAt0(R_DG1,DG1);
      Prepare_Plot(Zero,99.9,Pic_Dir & Filename & ".plot");  -- plot if Sign_DG1=0
    elsif Sign_DG1>0 then
      Show0("adding DG1 to G1");
      Assign(R_DG1,DG1);
    else
      Show0("subtracting DG1 from G1");
      Assign(-R_DG1,DG1);
    end if;
    Add(DG1,G1);
  end Adjust_G1;

  procedure Make_Initial(P: in out Point) is
    --- RS'-symmetry has G3=M3=0
    M1: constant Flt := Flt(   76555)/Flt(262144); --  0.292034149169921875
    M2: constant Flt := Flt(-1908461)/Flt(262144); -- -7.280200958251953125
    G1,G3,E: Scalar;
  begin
    Read(Input_Dir & FileName,P);
    Copy(P.Gamma(1),G1);
    Adjust_G1(G1);
    SetZero(G3);
    if Sign_DG1=0 then Enlarge(G1); Enlarge(G3); end if; -- for open domain
    MakeG2(G3,G1,P.Gamma);
    Assign(M1,M2,Zero,P.M);
    if Sign_DG1=0 then Enlarge(P.M); end if;             -- for open domain
    Ham(P,E);                                            -- just for info
    Copy(E,P.Energy);
  end Make_Initial;

  G3: Scalar;
  P0,P1: Point;

begin
  NTasks := NT;
  Verbosity := 3;

  Show0("running Roll_Point with");
  Show_Rattle_Param;
  Make_Initial(P0);
  ShowApprox1("P0",P0);
  Neg2Pos_Poincare(TNeed,F,P0,P1);
  ShowApprox1("P1",P1);
  Copy(P1.Gamma(3),G3);
  Reduce_Precision(64,G3);
  Show1("G3 =",G3);

end Roll_Point;
