with System.Machine_Code;
use System.Machine_Code;

pragma Optimize (Off);

package body Rounding.FPU is

  function Union(A,B: R_Set) return R_Set renames "or";

  procedure Set_Rounding_Mode(M: in Rounding_Mode) is
    W: R_Set;
  begin
    if (M.Mask = Strict) then
      case M.Direction is
        when Up        => W := Union(R_Strict,R_Up);
        when Nearest   => W := Union(R_Strict,R_Nearest);
        when Down      => W := Union(R_Strict,R_Down);
        when Truncate  => W := Union(R_Strict,R_Truncate);
      end case;
    else
      case M.Direction is
        when Nearest   => W := Union(R_Relaxed,R_Nearest);
        when Up        => W := Union(R_Relaxed,R_Up);
        when Down      => W := Union(R_Relaxed,R_Down);
        when Truncate  => W := Union(R_Relaxed,R_Truncate);
      end case;
    end if;

    if Current_Rounding_Mode.Mask=Relaxed then
      --- clear exception flags
      Asm("fnclex",Volatile => True);
      --- load control word into FPU
      Asm("fldcw %0",Inputs => R_Set'Asm_Input("g",W), Volatile => True);
    else
      --- clear exception flags after handling pending unmasked exceptions
      Asm("fclex",Volatile => True);
      --- load control word into FPU
      Asm("fldcw %0",Inputs => R_Set'Asm_Input("g",W), Volatile => True);
    end if;

    if Current_Rounding_Mode /= M then Current_Rounding_Mode := M; end if;
  end Set_Rounding_Mode;

  function Get_Rounding_Mode return Rounding_Mode is
  begin
    return Current_Rounding_Mode;
  end Get_Rounding_Mode;

  procedure Clear_Exception_Flags is
  begin
    Asm("fnclex",Volatile => True);
  end Clear_Exception_Flags;

  procedure Set_Default_Up_Mask(M: in Rounding_Mask) is
  begin
    Default_Up_Mask := M;
    Set_Rounding_Mode((Up,M));
  end Set_Default_Up_Mask;

  procedure Rounding_Up is
  begin
    Set_Rounding_Mode((Up,Default_Up_Mask));
  end Rounding_Up;

  procedure Rounding_Up_Strict is
  begin
    Set_Rounding_Mode((Up,Strict));
  end Rounding_Up_Strict;

  procedure Rounding_Up_Relaxed is
  begin
    Set_Rounding_Mode((Up,Relaxed));
  end Rounding_Up_Relaxed;

  procedure Rounding_Nearest_Relaxed is
  begin
    Set_Rounding_Mode((Nearest,Relaxed));
  end Rounding_Nearest_Relaxed;

end Rounding.FPU;
