
generic

  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>;
  with function Sup(S: Scalar) return Flt is <>;
  with function Scal(K: Integer) return Scalar is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with function Get_Precision(R: Scalar) return Positive is <>;

package ScalVectors.Eigenv.RNum is  --- purely numerical, for REAL matrices

  function CountRange(B: Vector; Left,Right: Flt) return Natural;             -- Number of elements in [Left,Right]
  procedure Eigen(A: in Matrix; W: in out Vector);                            -- eigenvalues of symmetric A
  procedure Eigen(A: in Matrix; W: in out Vector; V: in out Matrix);          -- eigenvalues/eigenvectors of symmetric A
  procedure SetEigen(V: in Matrix; W: in Vector; A: in out Matrix);           -- A := V*diag(W)*V^{-1}
  procedure Eigen(Ar,Ai: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix); -- eigenvals/eigenvecs of general Ar+i*Ai
  procedure Eigen(A: in Matrix; Wr,Wi: in out Vector; Vr,Vi: in out Matrix);     -- eigenvalues/eigenvectors of general A
  procedure QuasiEigen(A: in Matrix; Wr,Wi: in out Vector; Vr: in out Matrix);   -- for cc eigenvectors return real and imag part
  procedure ClosestEigen(A: in Matrix; Er,Ei: in out Scalar; Ur,Ui: in out Vector);
  function SpecRadius(A: Matrix; Eps: Radius) return Scalar;                           -- spectral radius (precision Eps)
  procedure Sign(X: in out Matrix; Eps: in Radius);           -- sign funcion via iterated cubic rational functions
  procedure SpecProjPos3(A: in out Matrix; Eps: in Radius);   -- spectral projection for Re(z)>0 (iterating cubic)
  procedure SpecProjPos2(A: in out Matrix; Eps: in Radius);   -- spectral projection for Re(z)>0 (iterting quadratic)
  procedure SpecProjPos(A: in out Matrix; Eps: in Radius) renames SpecProjPos3;
  procedure SpecProjOutsideInterval(L,R: in Scalar; P: in out Matrix; Eps: in Radius); -- spectral projection for not L<Re(z)<R
  procedure SpecProjInterval(L,R: in Scalar; P: in out Matrix; Eps: in Radius);        -- spectral projection for L<Re(z)<R

private

  function Epsilon(R: Flt) return Flt;
  procedure FixEval(Wr,Wi: in out Vector; E: in out Scalar);
  procedure FixEvec(Wi: in Vector; Vr,Vi: in out Matrix);
  procedure NormalizeEvec(Vr,Vi: in out Matrix);
  function HFun(L: Scalar) return Scalar;
  function NormSmallerOne(A: in Matrix) return Boolean;
  procedure SpecHalf(A,B,C,D: in out Matrix; Eps: in Radius);
  procedure SpecProjPos(A,B,C,D: in out Matrix; Eps: in Radius);

end ScalVectors.Eigenv.RNum;
