with Flts, ScalInfo, MultiScalars;
use Flts, ScalInfo;

pragma Elaborate_All (Flts,ScalInfo);

generic

  type Scalar is private;

  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;

  with function Scal(R: Flt) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;

  with package MS is new MultiScalars (Scalar => Scalar);

package Small_Vectors is

  use MS;

  -------------------- Vector2

  procedure Center2(V: in out Vector2);
  procedure Neg2(V: in out Vector2);
  procedure Copy2(U: in Vector2; V: in out Vector2);
  procedure Add2(U: in Vector2; V: in out Vector2);
  procedure Sum2(U,V: in Vector2; W: in out Vector2);
  procedure Sub2(U: in Vector2; V: in out Vector2);
  procedure Diff2(U,V: in Vector2; W: in out Vector2);
  procedure Mult2(S: in Scalar; U: in out Vector2);
  procedure Prod2(S: in Scalar; U: in Vector2; V: in out Vector2);
  procedure AddProd2(S: in Scalar; U: in Vector2; V: in out Vector2);
  procedure Dot2(U,V: in Vector2; S: in out Scalar);
  function Dot2(U,V: Vector2) return Scalar;
  procedure Vol2(U1,U2: in Vector2; S: in out Scalar);
  function Vol2(U1,U2: Vector2) return Scalar;
  function MaxNorm(V: Vector2) return Radius;

  -------------------- Matrix2

  procedure Center2(A: in out Matrix2);
  procedure Neg2(A: in out Matrix2);
  procedure Copy2(A: in Matrix2; B: in out Matrix2);
  procedure Transpose2(A: in out Matrix2);
  procedure Transpose2(A: in Matrix2; B: in out Matrix2);
  procedure SetRows2(U1,U2: in Vector2; A: in out Matrix2);
  procedure GetRows2(A: in Matrix2; U1,U2: in out Vector2);
  procedure SetColumns2(U1,U2: in Vector2; A: in out Matrix2);
  procedure GetColumns2(A: in Matrix2; U1,U2: in out Vector2);
  procedure Add2(A: in Matrix2; B: in out Matrix2);
  procedure Sum2(A,B: in Matrix2; C: in out Matrix2);
  procedure Sub2(A: in Matrix2; B: in out Matrix2);
  procedure Diff2(A,B: in Matrix2; C: in out Matrix2);
  procedure Mult2(S: in Scalar; A: in out Matrix2);
  function MaxNorm(A: Matrix2) return Radius;
  procedure Prod2(A: in Matrix2; U: in Vector2; V: in out Vector2);
  procedure Prod2(A,B: in Matrix2; C: in out Matrix2);

  procedure Tr2(A: in Matrix2; S: in out Scalar);
  function Tr2(A: Matrix2) return Scalar;
  procedure Det2(A: in Matrix2; S: in out Scalar);
  function Det2(A: Matrix2) return Scalar;
  procedure Adj2(A: in Matrix2; B: in out Matrix2);
  procedure Inv2(A: in Matrix2; B: in out Matrix2);
  procedure Quot2(B: in Vector2; A: in Matrix2; X: in out Vector2); -- solve A*X=B
  procedure Quot2(B,A: in Matrix2; X: in out Matrix2);              -- solve A*X=B

  -------------------- Vector3

  procedure Center3(V: in out Vector3);
  procedure Neg3(V: in out Vector3);
  procedure Copy3(U: in Vector3; V: in out Vector3);
  procedure Add3(U: in Vector3; V: in out Vector3);
  procedure Sum3(U,V: in Vector3; W: in out Vector3);
  procedure Sub3(U: in Vector3; V: in out Vector3);
  procedure Diff3(U,V: in Vector3; W: in out Vector3);
  procedure Mult3(S: in Scalar; U: in out Vector3);
  procedure Prod3(S: in Scalar; U: in Vector3; V: in out Vector3);
  procedure AddProd3(S: in Scalar; U: in Vector3; V: in out Vector3);
  procedure Dot3(U,V: in Vector3; S: in out Scalar);
  function Dot3(U,V: Vector3) return Scalar;
  procedure DiagProd3(B,U: in Vector3; BU: in out Vector3);
  procedure Vol3(U1,U2,U3: in Vector3; S: in out Scalar); -- S = U*(V x W)
  function Vol3(U1,U2,U3: Vector3) return Scalar;
  procedure Cross3(U,V: in Vector3; W: in out Vector3);   -- W = U x V
  procedure Cross3(U,V,W: in Vector3; X: in out Vector3); -- X = U x (V x W)
  function MaxNorm(V: Vector3) return Radius;

  -------------------- Matrix3

  procedure Center3(A: in out Matrix3);
  procedure Neg3(A: in out Matrix3);
  procedure Copy3(A: in Matrix3; B: in out Matrix3);
  procedure Transpose3(A: in out Matrix3);
  procedure Transpose3(A: in Matrix3; B: in out Matrix3);
  procedure SetRows3(U1,U2,U3: in Vector3; A: in out Matrix3);
  procedure GetRows3(A: in Matrix3; U1,U2,U3: in out Vector3);
  procedure SetColumns3(U1,U2,U3: in Vector3; A: in out Matrix3);
  procedure GetColumns3(A: in Matrix3; U1,U2,U3: in out Vector3);
  procedure Add3(A: in Matrix3; B: in out Matrix3);
  procedure Sum3(A,B: in Matrix3; C: in out Matrix3);
  procedure Sub3(A: in Matrix3; B: in out Matrix3);
  procedure Diff3(A,B: in Matrix3; C: in out Matrix3);
  procedure Mult3(S: in Scalar; A: in out Matrix3);
  function MaxNorm(A: Matrix3) return Radius;
  procedure Prod3(A: in Matrix3; U: in Vector3; V: in out Vector3);
  procedure Prod3(A,B: in Matrix3; C: in out Matrix3);
  procedure ColCross3(A: in Matrix3; U: in Vector3; B: in out Matrix3);
  procedure CrossSandwich3(U: in Vector3; A: in out Matrix3);  -- AX = U x X x U

  procedure Tr3(A: in Matrix3; S: in out Scalar);
  function Tr3(A: Matrix3) return Scalar;
  procedure Det3(A: in Matrix3; S: in out Scalar);
  function Det3(A: Matrix3) return Scalar;
  procedure Adj3(A: in Matrix3; B: in out Matrix3);
  procedure Inv3(A: in Matrix3; B: in out Matrix3);
  function Inv3(A: Matrix3) return Matrix3;

  -------------------- some Newton stuff
  -------------------- use Steps=0 for error estimate only

  generic
    with procedure TF(S: in Scalar; T,A: in out Scalar);
  procedure Find01(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0);

  generic
    with procedure F(S: in Scalar; T: in out Scalar);
    with procedure DF(S: in Scalar; A: in out Scalar);
  procedure FindZero1(S: in out Scalar; Steps: in Natural := 64; Verb: in Natural := 0);

  generic
    with procedure TF(V: in Vector2; W: in out Vector2; A: in out Matrix2);
  procedure Find02(V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0);

  generic
    with procedure F(V: in Vector2; W: in out Vector2);
    with procedure DF(V: in Vector2; A: in out Matrix2);
  procedure FindZero2(V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0);

  generic --- Quasi_Newton, rigorous if DerF is a bound on DF on a ball of radius R
    with procedure F(V: in Vector2; W: in out Vector2);
  procedure QFindZero2(DerF: in Matrix2; R: in Radius; V: in out Vector2; Steps: in Natural := 64; Verb: in Natural := 0);

private

  procedure Add(I: in Integer; S: in out Scalar);
  procedure Add(I: in Integer; A: in out Matrix2);
  procedure BallAt0(R: in Flt; V: in out Vector2);
  procedure BallAt0(R: in Flt; A: in out Matrix2);

  Scal_Zero: constant Scalar      := Scal(Zero);
  SInfo:     constant Scalar_Info := Info(Scal_Zero);
  STrunc:    constant Boolean     := SInfo.IsNumeric;

end Small_Vectors;
