with Ints, Globals, Rounding, Rounding.FPU, Strings, IPowers, QPowers;
use Ints, Globals, Rounding, Strings;

pragma Elaborate_All (Ints,Globals,Rounding,Strings,IPowers,QPowers);

package body Tangents is

  package SIP is new IPowers (Scalar => Tangent);
  procedure IPower(I: in Integer; P1: in Tangent; P2: in out Tangent) renames SIP.IPower;
  package SQP is new QPowers (Scalar => Tangent);
  procedure QPower(Q: in Rational; P1: in Tangent; P2: in out Tangent) renames SQP.QPower;

  procedure Show_Tangent_Param is
    use Rounding.FPU;
    M: constant Rounding_Mode := Get_Rounding_Mode;
    Prec: constant Positive := Get_Precision(SOne/Three);
  begin
    Show1("          Dim = ",Dim);
    Show0("      RepType =  " & SInfo.RepType);
    Show1("    Precision = ",Prec);   -- set precision before elaboration
    Show1("       STrunc =  ",STrunc);
    Show0("Rounding_Mode =  " & Strng(M));
    Show1("       NTasks = ",NTasks);
    Show1("          Rho = ",Rho);
  end Show_Tangent_Param;

  --- basic

  function Info(Dummy: Tangent) return Scalar_Info is
    pragma Unreferenced(Dummy);
    TInfo: constant Scalar_Info := (RepType    => SInfo.RepType,
                                    IsStandard => SInfo.IsStandard,
                                    IsNumeric  => SInfo.IsNumeric,
                                    IsComplex  => SInfo.IsComplex,
                                    NReps      => SInfo.NReps*DimPlus1,
                                    RMode      => SInfo.RMode);
  begin
    return TInfo;
  end Info;

  function IsSharp(P: Tangent) return Boolean is
  begin
    for I in 0 .. DimMinus1 loop
      if not IsSharp(P(I)) then return False; end if;
    end loop;
    return IsZero(P(Dim));
  end IsSharp;

  function IsZero(P: Tangent) return Boolean is
  begin
    for I in 0 .. DimMinus1 loop
      if not IsZero(P(I)) then return False; end if;
    end loop;
    return IsZero(P(Dim));
  end IsZero;

  function "="(P1,P2: Tangent) return Boolean is
  begin
    for I in 0 .. DimMinus1 loop
      if P1(I)/=P2(I) then return False; end if;
    end loop;
    return P1(Dim)=P2(Dim);
  end "=";

  procedure SetZero(P: in out Tangent) is
  begin
    for I in ZeroToDim loop SetZero(P(I)); end loop;
  end SetZero;

  procedure Copy(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Copy(P1(I),P2(I)); end loop;
  end Copy;

  procedure Swap(P1,P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Swap(P1(I),P2(I)); end loop;
  end Swap;

  --- sets

  function Center0(P: Tangent) return Boolean is
  begin
    for I in 0 .. DimMinus1 loop
      if not Center0(P(I)) then return False; end if;
    end loop;
    return Center0(P(Dim));
  end Center0;

  function Contains0(P: Tangent) return Logical is
    C: Logical := True;
  begin
    for I in ZeroToDim loop
      C := C and Contains0(P(I));
      if C=False then return False; end if;
    end loop;
    return C;
  end Contains0;

  function Contains(P1,P2: Tangent) return Logical is
    C: Logical := True;
  begin
    for I in ZeroToDim loop
      C := C and Contains(P1(I),P2(I));
      if C=False then return False; end if;
    end loop;
    return C;
  end Contains;

  procedure BallAt0(R: in Flt; P: in out Tangent) is
    E: constant Radius := R/Rho;
  begin
    BallAt0(R,P(0));
    for I in OneToDim loop BallAt0(E,P(I)); end loop;
  end BallAt0;

  function BallAt0(R: Flt) return Tangent is
    P: Tangent;
  begin
    BallAt0(R,P);
    return P;
  end BallAt0;

  procedure DiskAt0(R: in Flt; P: in out Tangent) is
  begin
    DiskAt0(R,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end DiskAt0;

  procedure ToErr(P: in out Tangent) is
  begin
    for I in ZeroToDim loop ToErr(P(I)); end loop;
  end ToErr;

  procedure ToErr(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop ToErr(P1(I),P2(I)); end loop;
  end ToErr;

  function ToErr(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    ToErr(P,Q);
    return Q;
  end ToErr;

  procedure FltBall(P: in Tangent; C,R: out Flt) is
    B: Flt := MaxNorm(P(1));
  begin
    FltBall(P(0),C,R);
    for I in 2 .. Dim loop B := B+MaxNorm(P(I)); end loop;
    R := R+Rho*B;
  end FltBall;

  procedure Center(P: in out Tangent) is
  begin
    for I in ZeroToDim loop Center(P(I)); end loop;
  end Center;

  procedure Center(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Center(P1(I),P2(I)); end loop;
  end Center;

  function Center(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Center(P,Q);
    return Q;
  end Center;

  procedure ModCenter(P: in out Tangent) is
  begin
    for I in ZeroToDim loop ModCenter(P(I)); end loop;
  end ModCenter;

  procedure ModCenter(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop ModCenter(P1(I),P2(I)); end loop;
  end ModCenter;

  function ModCenter(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    ModCenter(P,Q);
    return Q;
  end ModCenter;

  procedure ErrMult(R: in Radius; P: in out Tangent) is
  begin
    for I in ZeroToDim loop ErrMult(R,P(I)); end loop;
  end ErrMult;

  procedure Union(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Union(P1(I),P2(I)); end loop;
  end Union;

  function Union(P1,P2: Tangent) return Tangent is
    Q: Tangent;
  begin
    Copy(P2,Q);
    Union(P1,Q);
    return Q;
  end Union;

  procedure Intersection(P1: in Tangent; P2: in out Tangent; Empty: out Logical) is
    E: Logical;
  begin
    Empty := False;
    for I in ZeroToDim loop
      Intersection(P1(I),P2(I),E);
      Empty := Empty or E;
    end loop;
  end Intersection;

  --- order

  function Sign(P: Tangent) return Integer is
  begin
    return Sign(P(0));
  end Sign;

  function Compare(P1,P2: Tangent) return Integer is
  begin
    return Compare(P1(0),P2(0));
  end Compare;

  function "<"(P1,P2: Tangent) return Boolean is
  begin
    return (Compare(P1,P2) < 0);
  end "<";

  function "<="(P1,P2: Tangent) return Boolean is
  begin
    return (Compare(P1,P2) <= 0);
  end "<=";

  function ">="(P1,P2: Tangent) return Boolean is
  begin
    return (Compare(P1,P2) >= 0);
  end ">=";

  function ">"(P1,P2: Tangent) return Boolean is
  begin
    return (Compare(P1,P2) > 0);
  end ">";

  procedure Min(P1: in Tangent; P2: in out Tangent) is
  begin
    if P1<P2 then Copy(P1,P2); end if;
  end Min;

  procedure Min(P1,P2: in Tangent; P3: in out Tangent) is
  begin
    if P1<P2 then Copy(P1,P3); else Copy(P2,P3); end if;
  end Min;

  function Min(P1,P2: Tangent) return Tangent is
  begin
    if P1<P2 then return P1; else return P2; end if;
  end Min;

  procedure Max(P1: in Tangent; P2: in out Tangent) is
  begin
    if P1>P2 then Copy(P1,P2); end if;
  end Max;

  procedure Max(P1,P2: in Tangent; P3: in out Tangent) is
  begin
    if P1>P2 then Copy(P1,P3); else Copy(P2,P3); end if;
  end Max;

  function Max(P1,P2: Tangent) return Tangent is
  begin
    if P1>P2 then return P1; else return P2; end if;
  end Max;

  function Sup(P: Tangent) return Flt is
    R: Flt := abs(Sup(P(1)));
  begin
    for I in 2 .. Dim loop
      R := RMax(abs(Sup(P(I))),R);
    end loop;
    return Sup(P(0))+Rho*R;
  end Sup;

  function Inf(P: Tangent) return Flt is
    R: Flt := abs(Sup(P(1)));
  begin
    for I in 2 .. Dim loop
      R := RMax(abs(Sup(P(I))),R);
    end loop;
    return -(Rho*R-Inf(P(0)));
  end Inf;

  --- addition and multiplication etc.

  procedure Neg(P: in out Tangent) is
  begin
    for I in ZeroToDim loop Neg(P(I)); end loop;
  end Neg;

  procedure Neg(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Neg(P1(I),P2(I)); end loop;
  end Neg;

  function "-"(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Neg(P,Q);
    return Q;
  end "-";

  procedure Add(I: in Integer; P: in out Tangent) is
  begin
    Add(I,P(0));
  end Add;

  procedure Add(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Add(P1(I),P2(I)); end loop;
  end Add;

  procedure Sum(P1,P2: in Tangent; P3: in out Tangent) is
  begin
    for I in ZeroToDim loop Sum(P1(I),P2(I),P3(I)); end loop;
  end Sum;

  function "+"(P1,P2: Tangent) return Tangent is
    Q: Tangent;
  begin
    Sum(P1,P2,Q);
    return Q;
  end "+";

  procedure Sub(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Sub(P1(I),P2(I)); end loop;
  end Sub;

  procedure Diff(P1,P2: in Tangent; P3: in out Tangent) is
  begin
    for I in ZeroToDim loop Diff(P1(I),P2(I),P3(I)); end loop;
  end Diff;

  function "-"(P1,P2: Tangent) return Tangent is
    Q: Tangent;
  begin
    Diff(P1,P2,Q);
    return Q;
  end "-";

  procedure Mult(R: in Flt; P,Tmp: in out Tangent) is
    STmp: Scalar renames Tmp(0);
  begin
    Assign(R,STmp);
    for I in ZeroToDim loop Mult(STmp,P(I)); end loop;
  end Mult;

  procedure Mult(R: in Flt; P: in out Tangent; Tmp: in out Scalar) is
  begin
    Assign(R,Tmp);
    for I in ZeroToDim loop Mult(Tmp,P(I)); end loop;
  end Mult;

  procedure Mult(R: in Flt; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Mult(R,P(I)); end loop;
  end Mult;

  procedure Prod(R: in Flt; P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Prod(R,P1(I),P2(I)); end loop;
  end Prod;

  function "*"(R: Flt; P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Prod(R,P,Q);
    return Q;
  end "*";

  procedure AddProd(R: in Flt; P1: in Tangent; P2,Tmp: in out Tangent) is
    STmp: Scalar renames Tmp(0);
  begin
    for I in ZeroToDim loop AddProd(R,P1(I),P2(I),STmp); end loop;
  end AddProd;

  procedure AddProd(R: in Flt; P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop AddProd(R,P1(I),P2(I)); end loop;
  end AddProd;

  procedure Mult(Q: in Rational; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Mult(Q,P(I)); end loop;
  end Mult;

  procedure Mult(P1: in Tangent; P2,Tmp: in out Tangent) is
  begin
    Prod(P1,P2,Tmp);
    Copy(Tmp,P2);
  end Mult;

  procedure Mult(P1: in Tangent; P2: in out Tangent) is
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    declare
      Q: Tangent renames Tmp.all.Data;
    begin
      Prod(P1,P2,Q);
      Copy(Q,P2);
    end;
    TPool.Recycle(Tmp);
  end Mult;

  procedure Prod(P1,P2: in Tangent; P3: in out Tangent) is
    P10: Scalar renames P1(0);
    P20: Scalar renames P2(0);
  begin
    Prod(P10,P20,P3(0));
    for I in OneToDim loop
      Prod(P10,P2(I),P3(I));
      AddProd(P20,P1(I),P3(I));
    end loop;
  end Prod;

  function "*"(P1,P2: Tangent) return Tangent is
    Q: Tangent;
  begin
    Prod(P1,P2,Q);
    return Q;
  end "*";

  procedure AddProd(P1,P2: in Tangent; P3: in out Tangent) is
    P10: Scalar renames P1(0);
    P20: Scalar renames P2(0);
  begin
    AddProd(P10,P20,P3(0));
    for I in OneToDim loop
      AddProd(P10,P2(I),P3(I));
      AddProd(P20,P1(I),P3(I));
    end loop;
  end AddProd;

  procedure AddProd(P1,P2: in Tangent; P3,Tmp: in out Tangent) is
    STmp: Scalar renames Tmp(0);
    P10: Scalar renames P1(0);
    P20: Scalar renames P2(0);
  begin
    AddProd(P10,P20,P3(0),STmp);
    for I in OneToDim loop
      AddProd(P10,P2(I),P3(I),STmp);
      AddProd(P20,P1(I),P3(I),STmp);
    end loop;
  end AddProd;

  procedure SubProd(P1,P2: in Tangent; P3,Tmp: in out Tangent) is
    STmp: Scalar renames Tmp(0);
    P10: Scalar renames P1(0);
    P20: Scalar renames P2(0);
  begin
    SubProd(P10,P20,P3(0),STmp);
    for I in OneToDim loop
      SubProd(P10,P2(I),P3(I),STmp);
      SubProd(P20,P1(I),P3(I),STmp);
    end loop;
  end SubProd;

  procedure Div(R: in Flt; P,Tmp: in out Tangent) is
    STmp: Scalar renames Tmp(0);
  begin
    Quot(SOne,R,STmp);
    for I in ZeroToDim loop Mult(STmp,P(I)); end loop;
  end Div;

  procedure Div(R: in Flt; P: in out Tangent; Tmp: in out Scalar) is
  begin
    Quot(SOne,R,Tmp);
    for I in ZeroToDim loop Mult(Tmp,P(I)); end loop;
  end Div;

  procedure Div(R: in Flt; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Div(R,P(I)); end loop;
  end Div;

  procedure Quot(P1: in Tangent; R: in Flt; P2: in out Tangent) is
    STmp: Scalar;
  begin
    Quot(SOne,R,STmp);
    for I in ZeroToDim loop Prod(STmp,P1(I),P2(I)); end loop;
  end Quot;

  function "/"(P: Tangent; R: Flt) return Tangent is
    Q: Tangent;
  begin
    Quot(P,R,Q);
    return Q;
  end "/";

  procedure Inv(P: in out Tangent) is
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    declare
      Q: Tangent renames Tmp.all.Data;
    begin
      Inv(P,Q);
      Copy(Q,P);
    end;
    TPool.Recycle(Tmp);
  end Inv;

  function Inv(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Inv(P,Q);
    return Q;
  end Inv;

  procedure Inv(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
    P21: Scalar renames P2(1);
  begin
    Inv(P1(0),P20);
    Prod(P20,P20,P21);
    Neg(P21);               --- -1/Sqr(P1(0))
    for I in 2 .. Dim loop
      Prod(P21,P1(I),P2(I));
    end loop;
    Mult(P1(1),P21);
  end Inv;

  procedure Div(P1: in Tangent; P2: in out Tangent) is
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    declare
      Q: Tangent renames Tmp.all.Data;
    begin
      Inv(P2,Q);
      Prod(P1,Q,P2);
    end;
    TPool.Recycle(Tmp);
  end Div;

  procedure Quot(P1,P2: in Tangent; P3: in out Tangent) is
  begin
    Inv(P2,P3);
    Mult(P1,P3);
  end Quot;

  function "/"(P1,P2: Tangent) return Tangent is
    Q: Tangent;
  begin
    Inv(P2,Q);
    Mult(P1,Q);
    return Q;
  end "/";

  function "**"(P: Tangent; I: Integer) return Tangent is
    Q: Tangent;
  begin
    IPower(I,P,Q);
    return Q;
  end "**";

  --- functions

  function IsReal(P: Tangent) return Boolean is
  begin
    if SComplex then
      for I in ZeroToDim loop
        if not IsReal(P(I)) then return False; end if;
      end loop;
    end if;
    return True;
  end IsReal;

  procedure Adjoint(P: in out Tangent) is
  begin
    if SComplex then
      for I in ZeroToDim loop Adjoint(P(I)); end loop;
    end if;
  end Adjoint;

  procedure Adjoint(P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Adjoint(P1(I),P2(I)); end loop;
  end Adjoint;

  function Adjoint(P: Tangent) return Tangent is
  begin
    if SComplex then
      declare
        Q: Tangent;
      begin
        Adjoint(P,Q);
        return Q;
      end;
    end if;
    return P;
  end Adjoint;

  procedure Real_Part(P: in out Tangent) is
  begin
    if SComplex then
      for I in ZeroToDim loop Real_Part(P(I)); end loop;
    end if;
  end Real_Part;

  procedure Imag_Part(P: in out Tangent) is
  begin
    for I in ZeroToDim loop Imag_Part(P(I)); end loop;
  end Imag_Part;

  procedure Eval0(P: in out Tangent) is
  begin
    Eval0(P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Eval0;

  procedure Norm(P: in Tangent; Q: in out Tangent) is
  begin
    Norm(P,Q(0));
    for I in OneToDim loop SetZero(Q(I)); end loop;
  end Norm;

  function Norm(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Norm(P,Q);
    return Q;
  end Norm;

  function MaxNorm(P: Tangent) return Radius is
    R: Flt := MaxNorm(P(1));
  begin
    for I in 2 .. Dim loop
      R := RMax(MaxNorm(P(I)),R);
    end loop;
    return MaxNorm(P(0))+Rho*R;
  end MaxNorm;

  procedure Sqr(P: in out Tangent) is
    Tmp: LT_Pointer;
  begin
    TPool.Allocate(Tmp);
    declare
      Q: Tangent renames Tmp.all.Data;
    begin
      Prod(P,P,Q);
      Copy(Q,P);
    end;
    TPool.Recycle(Tmp);
  end Sqr;

  function Sqr(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Prod(P,P,Q);
    return Q;
  end Sqr;

  procedure Sqrt(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
    P21: Scalar renames P2(1);
  begin
    Sqrt(P1(0),P20);
    Inv(P20,P21);
    Mult(SHalf,P21);         --- 1/(2*Sqrt(P(0)))
    for I in 2 .. Dim loop
      Prod(P21,P1(I),P2(I));
    end loop;
    Mult(P1(1),P21);
  end Sqrt;

  function Sqrt(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Sqrt(P,Q);
    return Q;
  end Sqrt;

  procedure Root(K: in Positive; P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
    P21: Scalar renames P2(1);
  begin
    if K=1 then
      Copy(P1,P2);
    else
      Root(K,P1(0),P20);
      Quot(P20,P1(0),P21);
      Div(Flt(K),P21);          -- (1/K)*P1(0)**(1/K-1)
      for I in 2 .. Dim loop
        Prod(P21,P1(I),P2(I));
      end loop;
      Mult(P1(1),P21);
    end if;
  end Root;

  function Root(K: Positive; P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Root(K,P,Q);
    return Q;
  end Root;

  procedure Exp(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
  begin
    Exp(P1(0),P20);
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
  end Exp;

  function Exp(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Exp(P,Q);
    return Q;
  end Exp;

  procedure ArcCos(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
    P21: Scalar renames P2(1);
  begin
    Prod(P1(0),P1(0),P21);
    Diff(SOne,P21,P20);
    Sqrt(P20,P21);
    Inv(P21,P20);
    Neg(P20);              --- -1/Sqrt(1-P1(0)**2)
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
    ArcCos(P1(0),P20);
  end ArcCos;

  function ArcCos(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    ArcCos(P,Q);
    return Q;
  end ArcCos;

  procedure ArcSin(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
    P21: Scalar renames P2(1);
  begin
    Prod(P1(0),P1(0),P21);
    Diff(SOne,P21,P20);
    Sqrt(P20,P21);
    Inv(P21,P20);          --- 1/Sqrt(1-P1(0)**2)
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
    ArcSin(P1(0),P20);
  end ArcSin;

  function ArcSin(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    ArcSin(P,Q);
    return Q;
  end ArcSin;

  procedure Cos(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
  begin
    Sin(P1(0),P20);
    Neg(P20);               --- -sin(P1(0))
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
    Cos(P1(0),P20);
  end Cos;

  function Cos(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Cos(P,Q);
    return Q;
  end Cos;

  procedure Sin(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
  begin
    Cos(P1(0),P20);         --- cos(P1(0))
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
    Sin(P1(0),P20);
  end Sin;

  function Sin(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Sin(P,Q);
    return Q;
  end Sin;

  procedure Log(P1: in Tangent; P2: in out Tangent) is
    P20: Scalar renames P2(0);
  begin
    Inv(P1(0),P20);         --- 1/P1(0)
    for I in OneToDim loop
      Prod(P20,P1(I),P2(I));
    end loop;
    Log(P1(0),P20);
  end Log;

  function Log(P: Tangent) return Tangent is
    Q: Tangent;
  begin
    Log(P,Q);
    return Q;
  end Log;

  procedure Simple_Random(P: in out Tangent) is
    Fac: constant Radius := Half/Rho; -- ad hoc
  begin
    Simple_Random(P(0));
    for I in OneToDim loop
      Simple_Random(P(I));
      Mult(Fac,P(I));
    end loop;
  end Simple_Random;

  procedure Roots2(B,C: in Tangent; U1,U2,V: in out Tangent) is
  begin
    if not (IsConst(B) and IsConst(C)) then raise Not_Implemented; end if;
    SetZero(U1);
    SetZero(U2);
    SetZero(V);
    Roots2(B(0),C(0),U1(0),U2(0),V(0));
  end Roots2;

  procedure Roots3(B,C,D: in Tangent; U0,U1,U2,V: in out Tangent) is
  begin
    if not (IsConst(B) and IsConst(C) and IsZero(D)) then raise Not_Implemented; end if;
    SetZero(U0);
    SetZero(U1);
    SetZero(U2);
    SetZero(V);
    Roots3(B(0),C(0),D(0),U0(0),U1(0),U2(0),V(0));
  end Roots3;

  --- conversion and i/o

  function Approx(P: Tangent) return Flt is
  begin
    return Approx(P(0));
  end Approx;

  procedure Assign(I: in Integer; P: in out Tangent) is
  begin
    Assign(I,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Assign;

  procedure Assign(Q: in Rational; P: in out Tangent) is
  begin
    Assign(Q,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Assign;

  procedure Assign(R: in Flt; P: in out Tangent) is
  begin
    Assign(R,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Assign;

  function Scal(I: Integer) return Tangent is
    P: Tangent;
  begin
    Assign(I,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
    return P;
  end Scal;

  function Scal(Q: Rational) return Tangent is
    P: Tangent;
  begin
    Assign(Q,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
    return P;
  end Scal;

  function Scal(R: Flt) return Tangent is
    P: Tangent;
  begin
    Assign(R,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
    return P;
  end Scal;

  procedure Enclose(R1,R2: in Flt; P: in out Tangent) is
  begin
    Enclose(R1,R2,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Enclose;

  function Enclose(R1,R2: Flt) return Tangent is
    P: Tangent;
  begin
    Enclose(R1,R2,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
    return P;
  end Enclose;

  procedure Show1(N: in String; P: in Tangent; Hide0: in Boolean := True) is
    W: constant Positive := 5;
    AllZero: Boolean := True;
  begin
    for I in ZeroToDim loop
      if not IsZero(P(I)) then
        Show1(N & Strng(I,W) & " ",P(I));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0"); end if;
      Show0;
    end if;
  end Show1;

  procedure Show2(N: in String; P1,P2: in Tangent; Hide0: in Boolean := True) is
    W: constant Positive := 5;
    AllZero: Boolean := True;
  begin
    for I in ZeroToDim loop
      if not (IsZero(P1(I)) and then IsZero(P2(I))) then
        Show2(N & Strng(I,W) & " ",P1(I),P2(I));
        AllZero := False;
      end if;
    end loop;
    if not Hide0 then
      if AllZero then Show0(N & "0 0"); end if;
      Show0;
    end if;
  end Show2;

  procedure Put(F: in File_Type; P: in Tangent; Decimal: in Boolean := False) is
  begin
    Put(F,DimPlus1);
    for I in ZeroToDim loop
      Put(F,P(I),Decimal);
    end loop;
  end Put;

  procedure Get(F: in File_Type; P: in out Tangent; Decimal: in Boolean := False) is
    Trunc: Boolean := False;
    D: Natural := 0;
    S: Scalar;
  begin
    SetZero(P);
    Get(F,D);
    Get(F,P(0),Decimal);
    if Scalar_Get then D := 0; end if; -- read just P(0)
    for I in 1 .. D-1 loop
      if D>Dim then
        Get(F,S,Decimal);
        Trunc := True;
      else
        Get(F,P(I),Decimal);
      end if;
    end loop;
    if Trunc then Show0("Warning: input truncated"); end if;
  end Get;

  procedure Write(FileName: in String; P: in Tangent; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>0 then Show0("Writing " & FileName); end if;
    Create(F,Out_File,FileName);
    Put(F,P,Decimal);
    Close(F);
  end Write;

  procedure Read(FileName: in String; P: in out Tangent; Decimal: in Boolean := False) is
    F: File_Type;
  begin
    if Verbosity>1 then Show0("Reading " & FileName); end if;
    Open(F,In_File,FileName);
    Get(F,P,Decimal);
    Close(F);
  end Read;

  function Read(FileName: String; Decimal: Boolean := False) return Tangent is
    P: Tangent;
  begin
    Read(FileName,P,Decimal);
    return P;
  end Read;

  --- misc

  function Get_Precision(P: Tangent) return Positive is
  begin
    return Get_Precision(P(0));
  end Get_Precision;

  procedure Reduce_Precision(Binary_Digits: in Positive; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Reduce_Precision(Binary_Digits,P(I)); end loop;
  end Reduce_Precision;

  --------------- other standard procedures

  function IsConst(P: Tangent) return Boolean is
  begin
    for I in 1 .. DimMinus1 loop
      if not IsZero(P(I)) then return False; end if;
    end loop;
    return IsZero(P(Dim));
  end IsConst;

  procedure Assign(S: in Scalar; P: in out Tangent) is
  begin
    Copy(S,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
  end Assign;

  function Scal(S: Scalar) return Tangent is
    P: Tangent;
  begin
    Copy(S,P(0));
    for I in OneToDim loop SetZero(P(I)); end loop;
    return P;
  end Scal;

  procedure Add(S: in Scalar; P: in out Tangent) is
  begin
    Add(S,P(0));
  end Add;

  procedure Mult(S: in Scalar; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Mult(S,P(I)); end loop;
  end Mult;

  procedure Prod(S: in Scalar; P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop Prod(S,P1(I),P2(I)); end loop;
  end Prod;

  procedure AddProd(S: in Scalar; P1: in Tangent; P2: in out Tangent) is
  begin
    for I in ZeroToDim loop AddProd(S,P1(I),P2(I)); end loop;
  end AddProd;

  procedure SubProd(S: in Scalar; P1: in Tangent; P2: in out Tangent) is
    Tmp: Scalar;
  begin
    for I in ZeroToDim loop SubProd(S,P1(I),P2(I),Tmp); end loop;
  end SubProd;

  procedure Div(S: in Scalar; P: in out Tangent) is
  begin
    for I in ZeroToDim loop Div(S,P(I)); end loop;
  end Div;

  procedure Norm(P: in Tangent; S: in out Scalar) is
    Tmp: Scalar;
  begin
    Norm(P(1),S);
    for I in 2 .. Dim loop
      Norm(P(I),Tmp);
      Max(Tmp,S);
    end loop;
    Mult(Rho,S,Tmp);
    Norm(P(0),Tmp);
    Add(Tmp,S);
  end Norm;

end Tangents;
