with Ada.Text_IO, Logicals, Rationals, Flts, ScalInfo;
use Ada.Text_IO, Logicals, Rationals, Flts, ScalInfo;
with Pooling;

pragma Elaborate_All (Ada.Text_IO,Logicals,Rationals,Flts,ScalInfo);
pragma Elaborate_All (Pooling);

generic

  Dim: in Positive; -- number of dimensions
  Rho: in Radius;   -- necessary, see below

  -------------------- begin_include scalars Scalar

  type Scalar is private;
  with procedure Proper_Rounding is <>;

  --- basic
  with function Info(Dummy: Scalar) return Scalar_Info is <>;
  with function IsSharp(S: Scalar) return Boolean is <>;
  with function IsZero(S: Scalar) return Boolean is <>;
  with function "="(S1,S2: Scalar) return Boolean is <>;
  with procedure SetZero(S: in out Scalar) is <>;
  with procedure Copy(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Swap(S1,S2: in out Scalar) is <>;

  --- sets
  with function Center0(S: Scalar) return Boolean is <>;
  with function Contains0(S: Scalar) return Logical is <>;
  with function Contains(S1,S2: Scalar) return Logical is <>;
  with procedure BallAt0(R: in Flt; S: in out Scalar) is <>;
  with function BallAt0(R: Flt) return Scalar is <>;
  with procedure DiskAt0(R: in Flt; S: in out Scalar) is <>;
  with procedure ToErr(S: in out Scalar) is <>;
  with procedure ToErr(S1: in Scalar; S2: in out Scalar) is <>;
  with function ToErr(S: Scalar) return Scalar is <>;
  with procedure FltBall(S: in Scalar; C,R: out Flt) is <>;
  with procedure Center(S: in out Scalar) is <>;
  with procedure Center(S1: in Scalar; S2: in out Scalar) is <>;
  with function Center(S: Scalar) return Scalar is <>;
  with procedure ModCenter(S: in out Scalar) is <>;
  with procedure ModCenter(S1: in Scalar; S2: in out Scalar) is <>;
  with function ModCenter(S: Scalar) return Scalar is <>;
  with procedure ErrMult(R: in Radius; S: in out Scalar) is <>;
  with procedure Union(S1: in Scalar; S2: in out Scalar) is <>;
  with function Union(S1,S2: Scalar) return Scalar is <>;
  with procedure Intersection(S1: in Scalar; S2: in out Scalar; Empty: out Logical) is <>;

  --- order
  with function Sign(S: Scalar) return Integer is <>;
  with function Compare(S1,S2: Scalar) return Integer is <>;
  with function "<"(S1,S2: Scalar) return Boolean is <>;
  with function "<="(S1,S2: Scalar) return Boolean is <>;
  with function ">"(S1,S2: Scalar) return Boolean is <>;
  with function ">="(S1,S2: Scalar) return Boolean is <>;
  with procedure Min(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Min(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Min(S1,S2: Scalar) return Scalar is <>;
  with procedure Max(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Max(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function Max(S1,S2: Scalar) return Scalar is <>;
  with function Inf(S: Scalar) return Flt is <>; -- inf of real part
  with function Sup(S: Scalar) return Flt is <>; -- sup of real part

  --- arithmetic
  with procedure Neg(S: in out Scalar) is <>;
  with procedure Neg(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(I: in Integer; S: in out Scalar) is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Sum(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "+"(S1,S2: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Diff(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "-"(S1,S2: Scalar) return Scalar is <>;
  with procedure Mult(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Mult(R: in Flt; S: in out Scalar) is <>;
  with procedure Prod(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(R: Flt; S: Scalar) return Scalar is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2,Tmp: in out Scalar) is <>;
  with procedure AddProd(R: in Flt; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Mult(Q: in Rational; S: in out Scalar) is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Prod(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "*"(S1,S2: Scalar) return Scalar is <>;
  with procedure AddProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure AddProd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with procedure SubProd(S1,S2: in Scalar; S3,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S,Tmp: in out Scalar) is <>;
  with procedure Div(R: in Flt; S: in out Scalar) is <>;
  with procedure Quot(S1: in Scalar; R: in Flt; S2: in out Scalar) is <>;
  with function "/"(S: Scalar; R: Flt) return Scalar is <>;
  with procedure Div(S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Quot(S1,S2: in Scalar; S3: in out Scalar) is <>;
  with function "/"(S1,S2: Scalar) return Scalar is <>;
  with procedure Inv(S1: in Scalar; S2: in out Scalar) is <>;
  with function Inv(S: Scalar) return Scalar is <>;
  with function "**"(S: Scalar; I: Integer) return Scalar is <>;

  --- fun
  with function IsReal(S: Scalar) return Boolean is <>;
  with procedure Adjoint(S: in out Scalar) is <>;
  with procedure Adjoint(S1: in Scalar; S2: in out Scalar) is <>;
  with function Adjoint(S: Scalar) return Scalar is <>;
  with procedure Real_Part(S: in out Scalar) is <>;
  with procedure Imag_Part(S: in out Scalar) is <>;
  with procedure Eval0(S: in out Scalar) is <>;
  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure Norm(S1: in Scalar; S2: in out Scalar) is <>;
  with function Norm(S: Scalar) return Scalar is <>;
  with function MaxNorm(S: Scalar) return Radius is <>;
  with procedure Sqr(S: in out Scalar) is <>;
  with function Sqr(S: Scalar) return Scalar is <>;
  with procedure Sqrt(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sqrt(S: Scalar) return Scalar is <>;
  with procedure Root(K: in Positive; S1: in Scalar; S2: in out Scalar) is <>;
  with function Root(K: Positive; S: Scalar) return Scalar is <>;
  with procedure Exp(S1: in Scalar; S2: in out Scalar) is <>;
  with function Exp(S: Scalar) return Scalar is <>;
  with procedure ArcCos(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcCos(S: Scalar) return Scalar is <>;
  with procedure ArcSin(S1: in Scalar; S2: in out Scalar) is <>;
  with function ArcSin(S: Scalar) return Scalar is <>;
  with procedure Cos(S1: in Scalar; S2: in out Scalar) is <>;
  with function Cos(S: Scalar) return Scalar is <>;
  with procedure Sin(S1: in Scalar; S2: in out Scalar) is <>;
  with function Sin(S: Scalar) return Scalar is <>;
  with procedure Log(S1: in Scalar; S2: in out Scalar) is <>;
  with function Log(S: Scalar) return Scalar is <>;
  with procedure Simple_Random(S: in out Scalar) is <>;
  with procedure QPower(Q: in Rational; S1: in Scalar; S2: in out Scalar) is <>;
  with procedure Roots2(B,C: in Scalar; U1,U2,V: in out Scalar) is <>;
  with procedure Roots3(B,C,D: in Scalar; U0,U1,U2,V: in out Scalar) is <>;

  --- conversion and i/o
  with function Approx(S: Scalar) return Flt is <>;
  with procedure Assign(K: in Integer; S: in out Scalar) is <>;
  with procedure Assign(Q: in Rational; S: in out Scalar) is <>;
  with procedure Assign(R: in Flt; S: in out Scalar) is <>;
  with function Scal(I: Integer) return Scalar is <>;
  with function Scal(Q: Rational) return Scalar is <>;
  with function Scal(R: Flt) return Scalar is <>;
  with procedure Enclose(R1,R2: in Flt; S: in out Scalar) is <>;
  with function Enclose(R1,R2: Flt) return Scalar is <>;
  with procedure Show1(N: in String; S: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Show2(N: in String; S1,S2: in Scalar; Whatever: in Boolean := True) is <>;
  with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := False) is <>;
  with procedure Get(F: in File_Type; S: in out Scalar; Decimal: in Boolean := False) is <>;

  --- misc
  with function Get_Precision(R: Scalar) return Positive is <>;
  with procedure Reduce_Precision(Binary_Digits: in Positive; S: in out Scalar) is <>;

  -------------------- end_include scalars

package Tangents is  --- points and tangent vectors

  --- Rho, and including tangent vectors in Norm and BallAt0
  --- is necessary for estimating higher order errors in series
  --- a large Rho inceases Norm, a small Rho increases BallAt0

  subtype ZeroToDim is Integer range 0 .. Dim;
  subtype OneToDim is Integer range 1 .. Dim;

  type Tangent is array(ZeroToDim) of Scalar;

  procedure Show_Tangent_Param;

  --------------- standard Scalar procedures

  --- basic
  function Info(Dummy: Tangent) return Scalar_Info;
  function IsSharp(P: Tangent) return Boolean;         --- polynomial with sharp coeffs
  function IsZero(P: Tangent) return Boolean;
  function "="(P1,P2: Tangent) return Boolean;
  procedure SetZero(P: in out Tangent);
  procedure Copy(P1: in Tangent; P2: in out Tangent);
  procedure Swap(P1,P2: in out Tangent);

  --- sets
  function Center0(P: Tangent) return Boolean;
  function Contains0(P: Tangent) return Logical;
  function Contains(P1,P2: Tangent) return Logical;
  procedure BallAt0(R: in Flt; P: in out Tangent);
  function BallAt0(R: Flt) return Tangent;
  procedure DiskAt0(R: in Flt; P: in out Tangent);
  procedure ToErr(P: in out Tangent);
  procedure ToErr(P1: in Tangent; P2: in out Tangent);
  function ToErr(P: Tangent) return Tangent;
  procedure FltBall(P: in Tangent; C,R: out Flt);
  procedure Center(P: in out Tangent);
  procedure Center(P1: in Tangent; P2: in out Tangent);
  function Center(P: Tangent) return Tangent;
  procedure ModCenter(P: in out Tangent);
  procedure ModCenter(P1: in Tangent; P2: in out Tangent);
  function ModCenter(P: Tangent) return Tangent;
  procedure ErrMult(R: in Radius; P: in out Tangent);
  procedure Union(P1: in Tangent; P2: in out Tangent);
  function Union(P1,P2: Tangent) return Tangent;
  procedure Intersection(P1: in Tangent; P2: in out Tangent; Empty: out Logical);

  --- order, ignores tangent vector (except sup and inf)
  function Sign(P: Tangent) return Integer;
  function Compare(P1,P2: Tangent) return Integer;
  function "<"(P1,P2: Tangent) return Boolean;
  function "<="(P1,P2: Tangent) return Boolean;
  function ">="(P1,P2: Tangent) return Boolean;
  function ">"(P1,P2: Tangent) return Boolean;
  procedure Min(P1: in Tangent; P2: in out Tangent);
  procedure Min(P1,P2: in Tangent; P3: in out Tangent);
  function Min(P1,P2: Tangent) return Tangent;
  procedure Max(P1: in Tangent; P2: in out Tangent);
  procedure Max(P1,P2: in Tangent; P3: in out Tangent);
  function Max(P1,P2: Tangent) return Tangent;
  function Inf(P: Tangent) return Flt; -- on ball defined by l^1 norm
  function Sup(P: Tangent) return Flt; -- on ball defined by l^1 norm

  --- addition and multiplication etc.
  procedure Neg(P: in out Tangent);
  procedure Neg(P1: in Tangent; P2: in out Tangent);
  function "-"(P: Tangent) return Tangent;
  procedure Add(I: in Integer; P: in out Tangent);
  procedure Add(P1: in Tangent; P2: in out Tangent);
  procedure Sum(P1,P2: in Tangent; P3: in out Tangent);
  function "+"(P1,P2: Tangent) return Tangent;
  procedure Sub(P1: in Tangent; P2: in out Tangent);
  procedure Diff(P1,P2: in Tangent; P3: in out Tangent);
  function "-"(P1,P2: Tangent) return Tangent;
  procedure Mult(R: in Flt; P,Tmp: in out Tangent);
  procedure Mult(R: in Flt; P: in out Tangent; Tmp: in out Scalar);
  procedure Mult(R: in Flt; P: in out Tangent);
  procedure Prod(R: in Flt; P1: in Tangent; P2: in out Tangent);
  function "*"(R: Flt; P: Tangent) return Tangent;
  procedure AddProd(R: in Flt; P1: in Tangent; P2,Tmp: in out Tangent);
  procedure AddProd(R: in Flt; P1: in Tangent; P2: in out Tangent);
  procedure Mult(Q: in Rational; P: in out Tangent);
  procedure Mult(P1: in Tangent; P2,Tmp: in out Tangent);
  procedure Mult(P1: in Tangent; P2: in out Tangent);
  procedure Prod(P1,P2: in Tangent; P3: in out Tangent);
  function "*"(P1,P2: Tangent) return Tangent;
  procedure AddProd(P1,P2: in Tangent; P3: in out Tangent);
  procedure AddProd(P1,P2: in Tangent; P3,Tmp: in out Tangent);
  procedure SubProd(P1,P2: in Tangent; P3,Tmp: in out Tangent);
  procedure Div(R: in Flt; P,Tmp: in out Tangent);
  procedure Div(R: in Flt; P: in out Tangent; Tmp: in out Scalar);
  procedure Div(R: in Flt; P: in out Tangent);
  procedure Quot(P1: in Tangent; R: in Flt; P2: in out Tangent);
  function "/"(P: Tangent; R: Flt) return Tangent;
  procedure Inv(P: in out Tangent);
  procedure Inv(P1: in Tangent; P2: in out Tangent);
  function Inv(P: Tangent) return Tangent;
  procedure Div(P1: in Tangent; P2: in out Tangent);
  function "/"(P1,P2: Tangent) return Tangent;
  procedure Quot(P1,P2: in Tangent; P3: in out Tangent);
  function "**"(P: Tangent; I: Integer) return Tangent;

  --- functions
  function IsReal(P: Tangent) return Boolean;
  procedure Adjoint(P: in out Tangent);
  procedure Adjoint(P1: in Tangent; P2: in out Tangent);
  function Adjoint(P: Tangent) return Tangent;
  procedure Real_Part(P: in out Tangent);
  procedure Imag_Part(P: in out Tangent);
  procedure Eval0(P: in out Tangent);
  procedure Norm(P: in Tangent; Q: in out Tangent);
  function Norm(P: Tangent) return Tangent;
  function "abs"(P: Tangent) return Tangent renames Norm; -- avoid
  function MaxNorm(P: Tangent) return Radius;
  procedure Sqr(P: in out Tangent);
  function Sqr(P: Tangent) return Tangent;
  procedure Sqrt(P1: in Tangent; P2: in out Tangent);
  function Sqrt(P: Tangent) return Tangent;
  procedure Root(K: in Positive; P1: in Tangent; P2: in out Tangent);
  function Root(K: Positive; P: Tangent) return Tangent;
  procedure Exp(P1: in Tangent; P2: in out Tangent);
  function Exp(P: Tangent) return Tangent;
  procedure ArcCos(P1: in Tangent; P2: in out Tangent);
  function ArcCos(P: Tangent) return Tangent;
  procedure ArcSin(P1: in Tangent; P2: in out Tangent);
  function ArcSin(P: Tangent) return Tangent;
  procedure Cos(P1: in Tangent; P2: in out Tangent);
  function Cos(P: Tangent) return Tangent;
  procedure Sin(P1: in Tangent; P2: in out Tangent);
  function Sin(P: Tangent) return Tangent;
  procedure Log(P1: in Tangent; P2: in out Tangent);
  function Log(P: Tangent) return Tangent;
  procedure Simple_Random(P: in out Tangent);
  procedure IPower(I: in Integer; P1: in Tangent; P2: in out Tangent);
  procedure QPower(Q: in Rational; P1: in Tangent; P2: in out Tangent);
  procedure Roots2(B,C: in Tangent; U1,U2,V: in out Tangent);
  procedure Roots3(B,C,D: in Tangent; U0,U1,U2,V: in out Tangent);

  --- conversion and i/o
  function Approx(P: Tangent) return Flt;
  procedure Assign(I: in Integer; P: in out Tangent);
  procedure Assign(Q: in Rational; P: in out Tangent);
  procedure Assign(R: in Flt; P: in out Tangent);
  function Scal(I: Integer) return Tangent;
  function Scal(Q: Rational) return Tangent;
  function Scal(R: Flt) return Tangent;
  procedure Enclose(R1,R2: in Flt; P: in out Tangent);
  function Enclose(R1,R2: Flt) return Tangent;
  procedure Show1(N: in String; P: in Tangent; Hide0: in Boolean := True);
  procedure Show2(N: in String; P1,P2: in Tangent; Hide0: in Boolean := True);
  procedure Put(F: in File_Type; P: in Tangent; Decimal: in Boolean := False);
  procedure Get(F: in File_Type; P: in out Tangent; Decimal: in Boolean := False);
  procedure Write(FileName: in String; P: in Tangent; Decimal: in Boolean := False);
  procedure Read(FileName: in String; P: in out Tangent; Decimal: in Boolean := False);
  function Read(FileName: String; Decimal: Boolean := False) return Tangent;

  --- misc
  function Get_Precision(P: Tangent) return Positive;
  procedure Reduce_Precision(Binary_Digits: in Positive; P: in out Tangent);

  --------------- other standard procedures
  function IsConst(P: Tangent) return Boolean;
  procedure Assign(S: in Scalar; P: in out Tangent);
  function Scal(S: Scalar) return Tangent;
  procedure Add(S: in Scalar; P: in out Tangent);
  procedure Mult(S: in Scalar; P: in out Tangent);
  procedure Prod(S: in Scalar; P1: in Tangent; P2: in out Tangent);
  procedure AddProd(S: in Scalar; P1: in Tangent; P2: in out Tangent);
  procedure SubProd(S: in Scalar; P1: in Tangent; P2: in out Tangent);
  procedure Div(S: in Scalar; P: in out Tangent);
  procedure Norm(P: in Tangent; S: in out Scalar);

  procedure CP(P1: in Tangent; P2: in out Tangent) renames Copy;
  package TP is new Pooling (Item => Tangent);
  subtype Pooled_Tangent is TP.Controlled_LD_Pointer;

  package TAlloc renames TP.Alloc;
  subtype LT_Pointer is TAlloc.LD_Pointer;
  TAC: TAlloc.Controlled_Protected_Data;
  TPool: TAlloc.Protected_Data renames TAC.P;

  Scalar_Get: Boolean := False; --- reads Scalar instead of Tangent (use with caution)

private

  DimMinus1 : constant Integer := Dim-1;
  DimPlus1:   constant Integer := Dim+1;
  SZero:      constant Scalar  := Scal(Zero);
  SOne:       constant Scalar  := Scal(One);
  SHalf:      constant Scalar  := Scal(Half);
  SInfo:  constant Scalar_Info := Info(SZero);
  STrunc:     constant Boolean := SInfo.IsNumeric;
  Not_STrunc: constant Boolean := not STrunc;
  SComplex:   constant Boolean := SInfo.IsComplex;

end Tangents;
