with Ints, Quick_Sort;
use Ints;

package body Vectors.Sorting is

  function GT_MaxNorm(A,B: Component) return Boolean is
  begin
    return MaxNorm(A)>MaxNorm(B);
  end GT_MaxNorm;

  procedure Dec_MaxNorm_Sort(W: in out Vector) is
    procedure QSort is new Quick_Sort (Item => Component, Item_Vector => Vector, "<" => GT_MaxNorm);
  begin
    QSort(W'First,W'Last,W);
  end Dec_MaxNorm_Sort;

  function GT_MaxNorm(Eps: Radius; Wr,Wi: Vector; I,J: Integer) return Boolean is
    OneMinusEps: constant Radius := One-Eps;
    NrI: constant Flt := MaxNorm(Wr(I));
    NiI: constant Flt := MaxNorm(Wi(I));
    NI: constant Flt := NrI*NrI+NiI*NiI;
    NrJ: constant Flt := MaxNorm(Wr(J));
    NiJ: constant Flt := MaxNorm(Wi(J));
    NJ: constant Flt := NrJ*NrJ+NiJ*NiJ;
  begin
    if OneMinusEps*NI > NJ then return True; end if;
    if OneMinusEps*NJ > NI then return False; end if;
    if OneMinusEps*NrI > NrJ then return True; end if;
    if OneMinusEps*NrJ > NrI then return False; end if;
    declare
      ArI: constant Flt := Approx(Wr(I));
      AiI: constant Flt := Approx(Wi(I));
      ArJ: constant Flt := Approx(Wr(J));
      AiJ: constant Flt := Approx(Wi(J));
    begin
      if ArI > ArJ then return True; end if;
      if ArJ > ArI then return False; end if;
      if AiI > AiJ then return True; end if;
      if AiJ > AiI then return False; end if;
    end;
    return False;
  end GT_MaxNorm;

  procedure Dec_MaxNorm_Sort(Eps: in Radius; Wr,Wi: in out Vector) is
    K: Integer;
    procedure Swap(I,J: in out Integer) is begin K := I; I := J; J := K; end Swap;
    function GT(I,J: Integer) return Boolean is begin return GT_MaxNorm(Eps,Wr,Wi,I,J); end GT;
    procedure QSort is new Quick_Sort (Item => Integer, Item_Vector => Int_Vec, "<" => GT);
    P: Int_Vec(Wr'Range); -- Pivot
    Tmp: Vector(Wr'Range);
  begin
    for I in P'Range loop P(I) := I; end loop;
    QSort(P'First,P'Last,P);
    for I in Wr'Range loop Copy(Wr(I),Tmp(I)); end loop;
    for I in Wr'Range loop Copy(Tmp(P(I)),Wr(I)); end loop;
    for I in Wi'Range loop Copy(Wi(I),Tmp(I)); end loop;
    for I in Wi'Range loop Copy(Tmp(P(I)),Wi(I)); end loop;
  end Dec_MaxNorm_Sort;

end Vectors.Sorting;
